// Generated by the protocol buffer compiler.  DO NOT EDIT!
// NO CHECKED-IN PROTOBUF GENCODE
// source: Flight.proto
// Protobuf C++ Version: 7.34.1

#include "Flight.pb.h"

#include <algorithm>
#include <type_traits>
#include "google/protobuf/io/coded_stream.h"
#include "google/protobuf/generated_message_tctable_impl.h"
#include "google/protobuf/internal_visibility.h"
#include "google/protobuf/extension_set.h"
#include "google/protobuf/generated_message_util.h"
#include "google/protobuf/wire_format_lite.h"
#include "google/protobuf/descriptor.h"
#include "google/protobuf/generated_message_reflection.h"
#include "google/protobuf/reflection_ops.h"
#include "google/protobuf/wire_format.h"
// @@protoc_insertion_point(includes)

// Must be included last.
#include "google/protobuf/port_def.inc"
PROTOBUF_PRAGMA_INIT_SEG
namespace _pb = ::google::protobuf;
namespace _pbi = ::google::protobuf::internal;
namespace _fl = ::google::protobuf::internal::field_layout;
namespace arrow {
namespace flight {
namespace protocol {

inline constexpr Ticket::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        ticket_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()) {}

template <typename>
constexpr Ticket::Ticket(::_pbi::ConstantInitialized)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(Ticket_class_data_.base()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
struct TicketDefaultTypeInternal {
  constexpr TicketDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~TicketDefaultTypeInternal() {}
  union {
    Ticket _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 TicketDefaultTypeInternal _Ticket_default_instance_;

inline constexpr SetSessionOptionsResult_Error::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        value_{static_cast< ::arrow::flight::protocol::SetSessionOptionsResult_ErrorValue >(0)} {}

template <typename>
constexpr SetSessionOptionsResult_Error::SetSessionOptionsResult_Error(::_pbi::ConstantInitialized)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(SetSessionOptionsResult_Error_class_data_.base()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
struct SetSessionOptionsResult_ErrorDefaultTypeInternal {
  constexpr SetSessionOptionsResult_ErrorDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~SetSessionOptionsResult_ErrorDefaultTypeInternal() {}
  union {
    SetSessionOptionsResult_Error _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 SetSessionOptionsResult_ErrorDefaultTypeInternal _SetSessionOptionsResult_Error_default_instance_;

inline constexpr SessionOptionValue_StringListValue::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        #ifdef PROTOBUF_INTERNAL_REMOVE_ARENA_PTRS_REPEATED_PTR_FIELD
        values_{visibility, ::_pbi::InternalMetadataOffset::Build<
            ::arrow::flight::protocol::SessionOptionValue_StringListValue,
            PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::SessionOptionValue_StringListValue, _impl_.values_)>()
        }
        #else  // !PROTOBUF_INTERNAL_REMOVE_ARENA_PTRS_REPEATED_PTR_FIELD
        values_ {}
        #endif
     {}

template <typename>
constexpr SessionOptionValue_StringListValue::SessionOptionValue_StringListValue(::_pbi::ConstantInitialized)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(SessionOptionValue_StringListValue_class_data_.base()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
struct SessionOptionValue_StringListValueDefaultTypeInternal {
  constexpr SessionOptionValue_StringListValueDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~SessionOptionValue_StringListValueDefaultTypeInternal() {}
  union {
    SessionOptionValue_StringListValue _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 SessionOptionValue_StringListValueDefaultTypeInternal _SessionOptionValue_StringListValue_default_instance_;

inline constexpr SchemaResult::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        schema_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()) {}

template <typename>
constexpr SchemaResult::SchemaResult(::_pbi::ConstantInitialized)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(SchemaResult_class_data_.base()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
struct SchemaResultDefaultTypeInternal {
  constexpr SchemaResultDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~SchemaResultDefaultTypeInternal() {}
  union {
    SchemaResult _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 SchemaResultDefaultTypeInternal _SchemaResult_default_instance_;

inline constexpr Result::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        body_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()) {}

template <typename>
constexpr Result::Result(::_pbi::ConstantInitialized)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(Result_class_data_.base()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
struct ResultDefaultTypeInternal {
  constexpr ResultDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~ResultDefaultTypeInternal() {}
  union {
    Result _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 ResultDefaultTypeInternal _Result_default_instance_;

inline constexpr PutResult::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        app_metadata_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()) {}

template <typename>
constexpr PutResult::PutResult(::_pbi::ConstantInitialized)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(PutResult_class_data_.base()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
struct PutResultDefaultTypeInternal {
  constexpr PutResultDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~PutResultDefaultTypeInternal() {}
  union {
    PutResult _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PutResultDefaultTypeInternal _PutResult_default_instance_;

inline constexpr Location::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        uri_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()) {}

template <typename>
constexpr Location::Location(::_pbi::ConstantInitialized)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(Location_class_data_.base()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
struct LocationDefaultTypeInternal {
  constexpr LocationDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~LocationDefaultTypeInternal() {}
  union {
    Location _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 LocationDefaultTypeInternal _Location_default_instance_;

inline constexpr HandshakeResponse::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        payload_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        protocol_version_{::uint64_t{0u}} {}

template <typename>
constexpr HandshakeResponse::HandshakeResponse(::_pbi::ConstantInitialized)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(HandshakeResponse_class_data_.base()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
struct HandshakeResponseDefaultTypeInternal {
  constexpr HandshakeResponseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~HandshakeResponseDefaultTypeInternal() {}
  union {
    HandshakeResponse _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 HandshakeResponseDefaultTypeInternal _HandshakeResponse_default_instance_;

inline constexpr HandshakeRequest::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        payload_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        protocol_version_{::uint64_t{0u}} {}

template <typename>
constexpr HandshakeRequest::HandshakeRequest(::_pbi::ConstantInitialized)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(HandshakeRequest_class_data_.base()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
struct HandshakeRequestDefaultTypeInternal {
  constexpr HandshakeRequestDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~HandshakeRequestDefaultTypeInternal() {}
  union {
    HandshakeRequest _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 HandshakeRequestDefaultTypeInternal _HandshakeRequest_default_instance_;
template <typename>
constexpr GetSessionOptionsRequest::GetSessionOptionsRequest(::_pbi::ConstantInitialized)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::internal::ZeroFieldsBase(GetSessionOptionsRequest_class_data_.base()){}
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::internal::ZeroFieldsBase() {
}
#endif  // PROTOBUF_CUSTOM_VTABLE
struct GetSessionOptionsRequestDefaultTypeInternal {
  constexpr GetSessionOptionsRequestDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~GetSessionOptionsRequestDefaultTypeInternal() {}
  union {
    GetSessionOptionsRequest _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 GetSessionOptionsRequestDefaultTypeInternal _GetSessionOptionsRequest_default_instance_;

inline constexpr FlightDescriptor::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        #ifdef PROTOBUF_INTERNAL_REMOVE_ARENA_PTRS_REPEATED_PTR_FIELD
        path_{visibility, ::_pbi::InternalMetadataOffset::Build<
            ::arrow::flight::protocol::FlightDescriptor,
            PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::FlightDescriptor, _impl_.path_)>()
        }
        #else  // !PROTOBUF_INTERNAL_REMOVE_ARENA_PTRS_REPEATED_PTR_FIELD
        path_ {}
        #endif
        ,
        cmd_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        type_{static_cast< ::arrow::flight::protocol::FlightDescriptor_DescriptorType >(0)} {}

template <typename>
constexpr FlightDescriptor::FlightDescriptor(::_pbi::ConstantInitialized)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(FlightDescriptor_class_data_.base()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
struct FlightDescriptorDefaultTypeInternal {
  constexpr FlightDescriptorDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~FlightDescriptorDefaultTypeInternal() {}
  union {
    FlightDescriptor _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 FlightDescriptorDefaultTypeInternal _FlightDescriptor_default_instance_;
template <typename>
constexpr Empty::Empty(::_pbi::ConstantInitialized)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::internal::ZeroFieldsBase(Empty_class_data_.base()){}
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::internal::ZeroFieldsBase() {
}
#endif  // PROTOBUF_CUSTOM_VTABLE
struct EmptyDefaultTypeInternal {
  constexpr EmptyDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~EmptyDefaultTypeInternal() {}
  union {
    Empty _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 EmptyDefaultTypeInternal _Empty_default_instance_;

inline constexpr Criteria::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        expression_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()) {}

template <typename>
constexpr Criteria::Criteria(::_pbi::ConstantInitialized)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(Criteria_class_data_.base()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
struct CriteriaDefaultTypeInternal {
  constexpr CriteriaDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~CriteriaDefaultTypeInternal() {}
  union {
    Criteria _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 CriteriaDefaultTypeInternal _Criteria_default_instance_;

inline constexpr CloseSessionResult::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        status_{static_cast< ::arrow::flight::protocol::CloseSessionResult_Status >(0)} {}

template <typename>
constexpr CloseSessionResult::CloseSessionResult(::_pbi::ConstantInitialized)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(CloseSessionResult_class_data_.base()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
struct CloseSessionResultDefaultTypeInternal {
  constexpr CloseSessionResultDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~CloseSessionResultDefaultTypeInternal() {}
  union {
    CloseSessionResult _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 CloseSessionResultDefaultTypeInternal _CloseSessionResult_default_instance_;
template <typename>
constexpr CloseSessionRequest::CloseSessionRequest(::_pbi::ConstantInitialized)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::internal::ZeroFieldsBase(CloseSessionRequest_class_data_.base()){}
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::internal::ZeroFieldsBase() {
}
#endif  // PROTOBUF_CUSTOM_VTABLE
struct CloseSessionRequestDefaultTypeInternal {
  constexpr CloseSessionRequestDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~CloseSessionRequestDefaultTypeInternal() {}
  union {
    CloseSessionRequest _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 CloseSessionRequestDefaultTypeInternal _CloseSessionRequest_default_instance_;

inline constexpr CancelFlightInfoResult::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        status_{static_cast< ::arrow::flight::protocol::CancelStatus >(0)} {}

template <typename>
constexpr CancelFlightInfoResult::CancelFlightInfoResult(::_pbi::ConstantInitialized)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(CancelFlightInfoResult_class_data_.base()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
struct CancelFlightInfoResultDefaultTypeInternal {
  constexpr CancelFlightInfoResultDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~CancelFlightInfoResultDefaultTypeInternal() {}
  union {
    CancelFlightInfoResult _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 CancelFlightInfoResultDefaultTypeInternal _CancelFlightInfoResult_default_instance_;

inline constexpr BasicAuth::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        username_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        password_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()) {}

template <typename>
constexpr BasicAuth::BasicAuth(::_pbi::ConstantInitialized)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(BasicAuth_class_data_.base()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
struct BasicAuthDefaultTypeInternal {
  constexpr BasicAuthDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~BasicAuthDefaultTypeInternal() {}
  union {
    BasicAuth _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 BasicAuthDefaultTypeInternal _BasicAuth_default_instance_;

inline constexpr ActionType::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        type_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        description_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()) {}

template <typename>
constexpr ActionType::ActionType(::_pbi::ConstantInitialized)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(ActionType_class_data_.base()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
struct ActionTypeDefaultTypeInternal {
  constexpr ActionTypeDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~ActionTypeDefaultTypeInternal() {}
  union {
    ActionType _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 ActionTypeDefaultTypeInternal _ActionType_default_instance_;

inline constexpr Action::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        type_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        body_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()) {}

template <typename>
constexpr Action::Action(::_pbi::ConstantInitialized)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(Action_class_data_.base()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
struct ActionDefaultTypeInternal {
  constexpr ActionDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~ActionDefaultTypeInternal() {}
  union {
    Action _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 ActionDefaultTypeInternal _Action_default_instance_;
template <typename>
constexpr SetSessionOptionsResult_ErrorsEntry_DoNotUse::SetSessionOptionsResult_ErrorsEntry_DoNotUse(::_pbi::ConstantInitialized)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : SetSessionOptionsResult_ErrorsEntry_DoNotUse::MapEntry(SetSessionOptionsResult_ErrorsEntry_DoNotUse_class_data_.base()){}
#else   // PROTOBUF_CUSTOM_VTABLE
    : SetSessionOptionsResult_ErrorsEntry_DoNotUse::MapEntry() {
}
#endif  // PROTOBUF_CUSTOM_VTABLE
struct SetSessionOptionsResult_ErrorsEntry_DoNotUseDefaultTypeInternal {
  constexpr SetSessionOptionsResult_ErrorsEntry_DoNotUseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~SetSessionOptionsResult_ErrorsEntry_DoNotUseDefaultTypeInternal() {}
  union {
    SetSessionOptionsResult_ErrorsEntry_DoNotUse _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 SetSessionOptionsResult_ErrorsEntry_DoNotUseDefaultTypeInternal _SetSessionOptionsResult_ErrorsEntry_DoNotUse_default_instance_;

inline constexpr SessionOptionValue::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : option_value_{},
        _cached_size_{0},
        _oneof_case_{} {}

template <typename>
constexpr SessionOptionValue::SessionOptionValue(::_pbi::ConstantInitialized)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(SessionOptionValue_class_data_.base()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
struct SessionOptionValueDefaultTypeInternal {
  constexpr SessionOptionValueDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~SessionOptionValueDefaultTypeInternal() {}
  union {
    SessionOptionValue _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 SessionOptionValueDefaultTypeInternal _SessionOptionValue_default_instance_;

inline constexpr FlightEndpoint::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        #ifdef PROTOBUF_INTERNAL_REMOVE_ARENA_PTRS_REPEATED_PTR_FIELD
        location_{visibility, ::_pbi::InternalMetadataOffset::Build<
            ::arrow::flight::protocol::FlightEndpoint,
            PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::FlightEndpoint, _impl_.location_)>()
        }
        #else  // !PROTOBUF_INTERNAL_REMOVE_ARENA_PTRS_REPEATED_PTR_FIELD
        location_ {}
        #endif
        ,
        app_metadata_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        ticket_{nullptr},
        expiration_time_{nullptr} {}

template <typename>
constexpr FlightEndpoint::FlightEndpoint(::_pbi::ConstantInitialized)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(FlightEndpoint_class_data_.base()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
struct FlightEndpointDefaultTypeInternal {
  constexpr FlightEndpointDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~FlightEndpointDefaultTypeInternal() {}
  union {
    FlightEndpoint _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 FlightEndpointDefaultTypeInternal _FlightEndpoint_default_instance_;

inline constexpr FlightData::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        data_header_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        app_metadata_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        data_body_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        flight_descriptor_{nullptr} {}

template <typename>
constexpr FlightData::FlightData(::_pbi::ConstantInitialized)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(FlightData_class_data_.base()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
struct FlightDataDefaultTypeInternal {
  constexpr FlightDataDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~FlightDataDefaultTypeInternal() {}
  union {
    FlightData _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 FlightDataDefaultTypeInternal _FlightData_default_instance_;

inline constexpr SetSessionOptionsResult::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        #ifdef PROTOBUF_INTERNAL_REMOVE_ARENA_PTRS_MAP_FIELD
        errors_{visibility, ::_pbi::InternalMetadataOffset::Build<
            ::arrow::flight::protocol::SetSessionOptionsResult,
            PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::SetSessionOptionsResult, _impl_.errors_)>()
        }
        #else
        errors_ {}
        #endif
     {}

template <typename>
constexpr SetSessionOptionsResult::SetSessionOptionsResult(::_pbi::ConstantInitialized)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(SetSessionOptionsResult_class_data_.base()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
struct SetSessionOptionsResultDefaultTypeInternal {
  constexpr SetSessionOptionsResultDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~SetSessionOptionsResultDefaultTypeInternal() {}
  union {
    SetSessionOptionsResult _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 SetSessionOptionsResultDefaultTypeInternal _SetSessionOptionsResult_default_instance_;
template <typename>
constexpr SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse::SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse(::_pbi::ConstantInitialized)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse::MapEntry(SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse_class_data_.base()){}
#else   // PROTOBUF_CUSTOM_VTABLE
    : SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse::MapEntry() {
}
#endif  // PROTOBUF_CUSTOM_VTABLE
struct SetSessionOptionsRequest_SessionOptionsEntry_DoNotUseDefaultTypeInternal {
  constexpr SetSessionOptionsRequest_SessionOptionsEntry_DoNotUseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~SetSessionOptionsRequest_SessionOptionsEntry_DoNotUseDefaultTypeInternal() {}
  union {
    SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 SetSessionOptionsRequest_SessionOptionsEntry_DoNotUseDefaultTypeInternal _SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse_default_instance_;

inline constexpr RenewFlightEndpointRequest::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        endpoint_{nullptr} {}

template <typename>
constexpr RenewFlightEndpointRequest::RenewFlightEndpointRequest(::_pbi::ConstantInitialized)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(RenewFlightEndpointRequest_class_data_.base()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
struct RenewFlightEndpointRequestDefaultTypeInternal {
  constexpr RenewFlightEndpointRequestDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~RenewFlightEndpointRequestDefaultTypeInternal() {}
  union {
    RenewFlightEndpointRequest _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 RenewFlightEndpointRequestDefaultTypeInternal _RenewFlightEndpointRequest_default_instance_;
template <typename>
constexpr GetSessionOptionsResult_SessionOptionsEntry_DoNotUse::GetSessionOptionsResult_SessionOptionsEntry_DoNotUse(::_pbi::ConstantInitialized)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : GetSessionOptionsResult_SessionOptionsEntry_DoNotUse::MapEntry(GetSessionOptionsResult_SessionOptionsEntry_DoNotUse_class_data_.base()){}
#else   // PROTOBUF_CUSTOM_VTABLE
    : GetSessionOptionsResult_SessionOptionsEntry_DoNotUse::MapEntry() {
}
#endif  // PROTOBUF_CUSTOM_VTABLE
struct GetSessionOptionsResult_SessionOptionsEntry_DoNotUseDefaultTypeInternal {
  constexpr GetSessionOptionsResult_SessionOptionsEntry_DoNotUseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~GetSessionOptionsResult_SessionOptionsEntry_DoNotUseDefaultTypeInternal() {}
  union {
    GetSessionOptionsResult_SessionOptionsEntry_DoNotUse _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 GetSessionOptionsResult_SessionOptionsEntry_DoNotUseDefaultTypeInternal _GetSessionOptionsResult_SessionOptionsEntry_DoNotUse_default_instance_;

inline constexpr FlightInfo::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        #ifdef PROTOBUF_INTERNAL_REMOVE_ARENA_PTRS_REPEATED_PTR_FIELD
        endpoint_{visibility, ::_pbi::InternalMetadataOffset::Build<
            ::arrow::flight::protocol::FlightInfo,
            PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::FlightInfo, _impl_.endpoint_)>()
        }
        #else  // !PROTOBUF_INTERNAL_REMOVE_ARENA_PTRS_REPEATED_PTR_FIELD
        endpoint_ {}
        #endif
        ,
        schema_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        app_metadata_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        flight_descriptor_{nullptr},
        total_records_{::int64_t{0}},
        total_bytes_{::int64_t{0}},
        ordered_{false} {}

template <typename>
constexpr FlightInfo::FlightInfo(::_pbi::ConstantInitialized)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(FlightInfo_class_data_.base()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
struct FlightInfoDefaultTypeInternal {
  constexpr FlightInfoDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~FlightInfoDefaultTypeInternal() {}
  union {
    FlightInfo _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 FlightInfoDefaultTypeInternal _FlightInfo_default_instance_;

inline constexpr SetSessionOptionsRequest::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        #ifdef PROTOBUF_INTERNAL_REMOVE_ARENA_PTRS_MAP_FIELD
        session_options_{visibility, ::_pbi::InternalMetadataOffset::Build<
            ::arrow::flight::protocol::SetSessionOptionsRequest,
            PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::SetSessionOptionsRequest, _impl_.session_options_)>()
        }
        #else
        session_options_ {}
        #endif
     {}

template <typename>
constexpr SetSessionOptionsRequest::SetSessionOptionsRequest(::_pbi::ConstantInitialized)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(SetSessionOptionsRequest_class_data_.base()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
struct SetSessionOptionsRequestDefaultTypeInternal {
  constexpr SetSessionOptionsRequestDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~SetSessionOptionsRequestDefaultTypeInternal() {}
  union {
    SetSessionOptionsRequest _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 SetSessionOptionsRequestDefaultTypeInternal _SetSessionOptionsRequest_default_instance_;

inline constexpr PollInfo::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        info_{nullptr},
        flight_descriptor_{nullptr},
        expiration_time_{nullptr},
        progress_{0} {}

template <typename>
constexpr PollInfo::PollInfo(::_pbi::ConstantInitialized)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(PollInfo_class_data_.base()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
struct PollInfoDefaultTypeInternal {
  constexpr PollInfoDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~PollInfoDefaultTypeInternal() {}
  union {
    PollInfo _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PollInfoDefaultTypeInternal _PollInfo_default_instance_;

inline constexpr GetSessionOptionsResult::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        #ifdef PROTOBUF_INTERNAL_REMOVE_ARENA_PTRS_MAP_FIELD
        session_options_{visibility, ::_pbi::InternalMetadataOffset::Build<
            ::arrow::flight::protocol::GetSessionOptionsResult,
            PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::GetSessionOptionsResult, _impl_.session_options_)>()
        }
        #else
        session_options_ {}
        #endif
     {}

template <typename>
constexpr GetSessionOptionsResult::GetSessionOptionsResult(::_pbi::ConstantInitialized)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(GetSessionOptionsResult_class_data_.base()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
struct GetSessionOptionsResultDefaultTypeInternal {
  constexpr GetSessionOptionsResultDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~GetSessionOptionsResultDefaultTypeInternal() {}
  union {
    GetSessionOptionsResult _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 GetSessionOptionsResultDefaultTypeInternal _GetSessionOptionsResult_default_instance_;

inline constexpr CancelFlightInfoRequest::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        info_{nullptr} {}

template <typename>
constexpr CancelFlightInfoRequest::CancelFlightInfoRequest(::_pbi::ConstantInitialized)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(CancelFlightInfoRequest_class_data_.base()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
struct CancelFlightInfoRequestDefaultTypeInternal {
  constexpr CancelFlightInfoRequestDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~CancelFlightInfoRequestDefaultTypeInternal() {}
  union {
    CancelFlightInfoRequest _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 CancelFlightInfoRequestDefaultTypeInternal _CancelFlightInfoRequest_default_instance_;
}  // namespace protocol
}  // namespace flight
}  // namespace arrow
static const ::_pb::EnumDescriptor* PROTOBUF_NONNULL
    file_level_enum_descriptors_Flight_2eproto[4];
static constexpr const ::_pb::ServiceDescriptor* PROTOBUF_NONNULL* PROTOBUF_NULLABLE
    file_level_service_descriptors_Flight_2eproto = nullptr;
const ::uint32_t
    TableStruct_Flight_2eproto::offsets[] ABSL_ATTRIBUTE_SECTION_VARIABLE(
        protodesc_cold) = {
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::HandshakeRequest, _impl_._has_bits_),
        5, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::HandshakeRequest, _impl_.protocol_version_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::HandshakeRequest, _impl_.payload_),
        1,
        0,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::HandshakeResponse, _impl_._has_bits_),
        5, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::HandshakeResponse, _impl_.protocol_version_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::HandshakeResponse, _impl_.payload_),
        1,
        0,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::BasicAuth, _impl_._has_bits_),
        5, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::BasicAuth, _impl_.username_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::BasicAuth, _impl_.password_),
        0,
        1,
        0x000, // bitmap
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::ActionType, _impl_._has_bits_),
        5, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::ActionType, _impl_.type_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::ActionType, _impl_.description_),
        0,
        1,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::Criteria, _impl_._has_bits_),
        4, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::Criteria, _impl_.expression_),
        0,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::Action, _impl_._has_bits_),
        5, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::Action, _impl_.type_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::Action, _impl_.body_),
        0,
        1,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::Result, _impl_._has_bits_),
        4, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::Result, _impl_.body_),
        0,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::SchemaResult, _impl_._has_bits_),
        4, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::SchemaResult, _impl_.schema_),
        0,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::FlightDescriptor, _impl_._has_bits_),
        6, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::FlightDescriptor, _impl_.type_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::FlightDescriptor, _impl_.cmd_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::FlightDescriptor, _impl_.path_),
        2,
        1,
        0,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::FlightInfo, _impl_._has_bits_),
        10, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::FlightInfo, _impl_.schema_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::FlightInfo, _impl_.flight_descriptor_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::FlightInfo, _impl_.endpoint_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::FlightInfo, _impl_.total_records_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::FlightInfo, _impl_.total_bytes_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::FlightInfo, _impl_.ordered_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::FlightInfo, _impl_.app_metadata_),
        1,
        3,
        0,
        4,
        5,
        6,
        2,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::PollInfo, _impl_._has_bits_),
        7, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::PollInfo, _impl_.info_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::PollInfo, _impl_.flight_descriptor_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::PollInfo, _impl_.progress_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::PollInfo, _impl_.expiration_time_),
        0,
        1,
        3,
        2,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::CancelFlightInfoRequest, _impl_._has_bits_),
        4, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::CancelFlightInfoRequest, _impl_.info_),
        0,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::CancelFlightInfoResult, _impl_._has_bits_),
        4, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::CancelFlightInfoResult, _impl_.status_),
        0,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::Ticket, _impl_._has_bits_),
        4, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::Ticket, _impl_.ticket_),
        0,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::Location, _impl_._has_bits_),
        4, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::Location, _impl_.uri_),
        0,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::FlightEndpoint, _impl_._has_bits_),
        7, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::FlightEndpoint, _impl_.ticket_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::FlightEndpoint, _impl_.location_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::FlightEndpoint, _impl_.expiration_time_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::FlightEndpoint, _impl_.app_metadata_),
        2,
        0,
        3,
        1,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::RenewFlightEndpointRequest, _impl_._has_bits_),
        4, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::RenewFlightEndpointRequest, _impl_.endpoint_),
        0,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::FlightData, _impl_._has_bits_),
        7, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::FlightData, _impl_.flight_descriptor_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::FlightData, _impl_.data_header_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::FlightData, _impl_.app_metadata_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::FlightData, _impl_.data_body_),
        3,
        0,
        1,
        2,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::PutResult, _impl_._has_bits_),
        4, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::PutResult, _impl_.app_metadata_),
        0,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::SessionOptionValue_StringListValue, _impl_._has_bits_),
        4, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::SessionOptionValue_StringListValue, _impl_.values_),
        0,
        0x004, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::SessionOptionValue, _impl_._oneof_case_[0]),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::SessionOptionValue, _impl_.option_value_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::SessionOptionValue, _impl_.option_value_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::SessionOptionValue, _impl_.option_value_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::SessionOptionValue, _impl_.option_value_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::SessionOptionValue, _impl_.option_value_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::SessionOptionValue, _impl_.option_value_),
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse, _impl_._has_bits_),
        5, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse, _impl_.key_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse, _impl_.value_),
        0,
        1,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::SetSessionOptionsRequest, _impl_._has_bits_),
        4, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::SetSessionOptionsRequest, _impl_.session_options_),
        0,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::SetSessionOptionsResult_Error, _impl_._has_bits_),
        4, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::SetSessionOptionsResult_Error, _impl_.value_),
        0,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::SetSessionOptionsResult_ErrorsEntry_DoNotUse, _impl_._has_bits_),
        5, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::SetSessionOptionsResult_ErrorsEntry_DoNotUse, _impl_.key_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::SetSessionOptionsResult_ErrorsEntry_DoNotUse, _impl_.value_),
        0,
        1,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::SetSessionOptionsResult, _impl_._has_bits_),
        4, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::SetSessionOptionsResult, _impl_.errors_),
        0,
        0x000, // bitmap
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::GetSessionOptionsResult_SessionOptionsEntry_DoNotUse, _impl_._has_bits_),
        5, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::GetSessionOptionsResult_SessionOptionsEntry_DoNotUse, _impl_.key_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::GetSessionOptionsResult_SessionOptionsEntry_DoNotUse, _impl_.value_),
        0,
        1,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::GetSessionOptionsResult, _impl_._has_bits_),
        4, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::GetSessionOptionsResult, _impl_.session_options_),
        0,
        0x000, // bitmap
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::CloseSessionResult, _impl_._has_bits_),
        4, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::CloseSessionResult, _impl_.status_),
        0,
};

static const ::_pbi::MigrationSchema
    schemas[] ABSL_ATTRIBUTE_SECTION_VARIABLE(protodesc_cold) = {
        {0, sizeof(::arrow::flight::protocol::HandshakeRequest)},
        {7, sizeof(::arrow::flight::protocol::HandshakeResponse)},
        {14, sizeof(::arrow::flight::protocol::BasicAuth)},
        {21, sizeof(::arrow::flight::protocol::Empty)},
        {22, sizeof(::arrow::flight::protocol::ActionType)},
        {29, sizeof(::arrow::flight::protocol::Criteria)},
        {34, sizeof(::arrow::flight::protocol::Action)},
        {41, sizeof(::arrow::flight::protocol::Result)},
        {46, sizeof(::arrow::flight::protocol::SchemaResult)},
        {51, sizeof(::arrow::flight::protocol::FlightDescriptor)},
        {60, sizeof(::arrow::flight::protocol::FlightInfo)},
        {77, sizeof(::arrow::flight::protocol::PollInfo)},
        {88, sizeof(::arrow::flight::protocol::CancelFlightInfoRequest)},
        {93, sizeof(::arrow::flight::protocol::CancelFlightInfoResult)},
        {98, sizeof(::arrow::flight::protocol::Ticket)},
        {103, sizeof(::arrow::flight::protocol::Location)},
        {108, sizeof(::arrow::flight::protocol::FlightEndpoint)},
        {119, sizeof(::arrow::flight::protocol::RenewFlightEndpointRequest)},
        {124, sizeof(::arrow::flight::protocol::FlightData)},
        {135, sizeof(::arrow::flight::protocol::PutResult)},
        {140, sizeof(::arrow::flight::protocol::SessionOptionValue_StringListValue)},
        {145, sizeof(::arrow::flight::protocol::SessionOptionValue)},
        {153, sizeof(::arrow::flight::protocol::SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse)},
        {160, sizeof(::arrow::flight::protocol::SetSessionOptionsRequest)},
        {165, sizeof(::arrow::flight::protocol::SetSessionOptionsResult_Error)},
        {170, sizeof(::arrow::flight::protocol::SetSessionOptionsResult_ErrorsEntry_DoNotUse)},
        {177, sizeof(::arrow::flight::protocol::SetSessionOptionsResult)},
        {182, sizeof(::arrow::flight::protocol::GetSessionOptionsRequest)},
        {183, sizeof(::arrow::flight::protocol::GetSessionOptionsResult_SessionOptionsEntry_DoNotUse)},
        {190, sizeof(::arrow::flight::protocol::GetSessionOptionsResult)},
        {195, sizeof(::arrow::flight::protocol::CloseSessionRequest)},
        {196, sizeof(::arrow::flight::protocol::CloseSessionResult)},
};
static const ::_pb::Message* PROTOBUF_NONNULL const file_default_instances[] = {
    &::arrow::flight::protocol::_HandshakeRequest_default_instance_._instance,
    &::arrow::flight::protocol::_HandshakeResponse_default_instance_._instance,
    &::arrow::flight::protocol::_BasicAuth_default_instance_._instance,
    &::arrow::flight::protocol::_Empty_default_instance_._instance,
    &::arrow::flight::protocol::_ActionType_default_instance_._instance,
    &::arrow::flight::protocol::_Criteria_default_instance_._instance,
    &::arrow::flight::protocol::_Action_default_instance_._instance,
    &::arrow::flight::protocol::_Result_default_instance_._instance,
    &::arrow::flight::protocol::_SchemaResult_default_instance_._instance,
    &::arrow::flight::protocol::_FlightDescriptor_default_instance_._instance,
    &::arrow::flight::protocol::_FlightInfo_default_instance_._instance,
    &::arrow::flight::protocol::_PollInfo_default_instance_._instance,
    &::arrow::flight::protocol::_CancelFlightInfoRequest_default_instance_._instance,
    &::arrow::flight::protocol::_CancelFlightInfoResult_default_instance_._instance,
    &::arrow::flight::protocol::_Ticket_default_instance_._instance,
    &::arrow::flight::protocol::_Location_default_instance_._instance,
    &::arrow::flight::protocol::_FlightEndpoint_default_instance_._instance,
    &::arrow::flight::protocol::_RenewFlightEndpointRequest_default_instance_._instance,
    &::arrow::flight::protocol::_FlightData_default_instance_._instance,
    &::arrow::flight::protocol::_PutResult_default_instance_._instance,
    &::arrow::flight::protocol::_SessionOptionValue_StringListValue_default_instance_._instance,
    &::arrow::flight::protocol::_SessionOptionValue_default_instance_._instance,
    &::arrow::flight::protocol::_SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse_default_instance_._instance,
    &::arrow::flight::protocol::_SetSessionOptionsRequest_default_instance_._instance,
    &::arrow::flight::protocol::_SetSessionOptionsResult_Error_default_instance_._instance,
    &::arrow::flight::protocol::_SetSessionOptionsResult_ErrorsEntry_DoNotUse_default_instance_._instance,
    &::arrow::flight::protocol::_SetSessionOptionsResult_default_instance_._instance,
    &::arrow::flight::protocol::_GetSessionOptionsRequest_default_instance_._instance,
    &::arrow::flight::protocol::_GetSessionOptionsResult_SessionOptionsEntry_DoNotUse_default_instance_._instance,
    &::arrow::flight::protocol::_GetSessionOptionsResult_default_instance_._instance,
    &::arrow::flight::protocol::_CloseSessionRequest_default_instance_._instance,
    &::arrow::flight::protocol::_CloseSessionResult_default_instance_._instance,
};
const char descriptor_table_protodef_Flight_2eproto[] ABSL_ATTRIBUTE_SECTION_VARIABLE(
    protodesc_cold) = {
    "\n\014Flight.proto\022\025arrow.flight.protocol\032\037g"
    "oogle/protobuf/timestamp.proto\"=\n\020Handsh"
    "akeRequest\022\030\n\020protocol_version\030\001 \001(\004\022\017\n\007"
    "payload\030\002 \001(\014\">\n\021HandshakeResponse\022\030\n\020pr"
    "otocol_version\030\001 \001(\004\022\017\n\007payload\030\002 \001(\014\"/\n"
    "\tBasicAuth\022\020\n\010username\030\002 \001(\t\022\020\n\010password"
    "\030\003 \001(\t\"\007\n\005Empty\"/\n\nActionType\022\014\n\004type\030\001 "
    "\001(\t\022\023\n\013description\030\002 \001(\t\"\036\n\010Criteria\022\022\n\n"
    "expression\030\001 \001(\014\"$\n\006Action\022\014\n\004type\030\001 \001(\t"
    "\022\014\n\004body\030\002 \001(\014\"\026\n\006Result\022\014\n\004body\030\001 \001(\014\"\036"
    "\n\014SchemaResult\022\016\n\006schema\030\001 \001(\014\"\245\001\n\020Fligh"
    "tDescriptor\022D\n\004type\030\001 \001(\01626.arrow.flight"
    ".protocol.FlightDescriptor.DescriptorTyp"
    "e\022\013\n\003cmd\030\002 \001(\014\022\014\n\004path\030\003 \003(\t\"0\n\016Descript"
    "orType\022\013\n\007UNKNOWN\020\000\022\010\n\004PATH\020\001\022\007\n\003CMD\020\002\"\354"
    "\001\n\nFlightInfo\022\016\n\006schema\030\001 \001(\014\022B\n\021flight_"
    "descriptor\030\002 \001(\0132\'.arrow.flight.protocol"
    ".FlightDescriptor\0227\n\010endpoint\030\003 \003(\0132%.ar"
    "row.flight.protocol.FlightEndpoint\022\025\n\rto"
    "tal_records\030\004 \001(\003\022\023\n\013total_bytes\030\005 \001(\003\022\017"
    "\n\007ordered\030\006 \001(\010\022\024\n\014app_metadata\030\007 \001(\014\"\330\001"
    "\n\010PollInfo\022/\n\004info\030\001 \001(\0132!.arrow.flight."
    "protocol.FlightInfo\022B\n\021flight_descriptor"
    "\030\002 \001(\0132\'.arrow.flight.protocol.FlightDes"
    "criptor\022\025\n\010progress\030\003 \001(\001H\000\210\001\001\0223\n\017expira"
    "tion_time\030\004 \001(\0132\032.google.protobuf.Timest"
    "ampB\013\n\t_progress\"J\n\027CancelFlightInfoRequ"
    "est\022/\n\004info\030\001 \001(\0132!.arrow.flight.protoco"
    "l.FlightInfo\"M\n\026CancelFlightInfoResult\0223"
    "\n\006status\030\001 \001(\0162#.arrow.flight.protocol.C"
    "ancelStatus\"\030\n\006Ticket\022\016\n\006ticket\030\001 \001(\014\"\027\n"
    "\010Location\022\013\n\003uri\030\001 \001(\t\"\275\001\n\016FlightEndpoin"
    "t\022-\n\006ticket\030\001 \001(\0132\035.arrow.flight.protoco"
    "l.Ticket\0221\n\010location\030\002 \003(\0132\037.arrow.fligh"
    "t.protocol.Location\0223\n\017expiration_time\030\003"
    " \001(\0132\032.google.protobuf.Timestamp\022\024\n\014app_"
    "metadata\030\004 \001(\014\"U\n\032RenewFlightEndpointReq"
    "uest\0227\n\010endpoint\030\001 \001(\0132%.arrow.flight.pr"
    "otocol.FlightEndpoint\"\217\001\n\nFlightData\022B\n\021"
    "flight_descriptor\030\001 \001(\0132\'.arrow.flight.p"
    "rotocol.FlightDescriptor\022\023\n\013data_header\030"
    "\002 \001(\014\022\024\n\014app_metadata\030\003 \001(\014\022\022\n\tdata_body"
    "\030\350\007 \001(\014\"!\n\tPutResult\022\024\n\014app_metadata\030\001 \001"
    "(\014\"\374\001\n\022SessionOptionValue\022\026\n\014string_valu"
    "e\030\001 \001(\tH\000\022\024\n\nbool_value\030\002 \001(\010H\000\022\025\n\013int64"
    "_value\030\003 \001(\020H\000\022\026\n\014double_value\030\004 \001(\001H\000\022V"
    "\n\021string_list_value\030\005 \001(\01329.arrow.flight"
    ".protocol.SessionOptionValue.StringListV"
    "alueH\000\032!\n\017StringListValue\022\016\n\006values\030\001 \003("
    "\tB\016\n\014option_value\"\332\001\n\030SetSessionOptionsR"
    "equest\022\\\n\017session_options\030\001 \003(\0132C.arrow."
    "flight.protocol.SetSessionOptionsRequest"
    ".SessionOptionsEntry\032`\n\023SessionOptionsEn"
    "try\022\013\n\003key\030\001 \001(\t\0228\n\005value\030\002 \001(\0132).arrow."
    "flight.protocol.SessionOptionValue:\0028\001\"\354"
    "\002\n\027SetSessionOptionsResult\022J\n\006errors\030\001 \003"
    "(\0132:.arrow.flight.protocol.SetSessionOpt"
    "ionsResult.ErrorsEntry\032Q\n\005Error\022H\n\005value"
    "\030\001 \001(\01629.arrow.flight.protocol.SetSessio"
    "nOptionsResult.ErrorValue\032c\n\013ErrorsEntry"
    "\022\013\n\003key\030\001 \001(\t\022C\n\005value\030\002 \001(\01324.arrow.fli"
    "ght.protocol.SetSessionOptionsResult.Err"
    "or:\0028\001\"M\n\nErrorValue\022\017\n\013UNSPECIFIED\020\000\022\020\n"
    "\014INVALID_NAME\020\001\022\021\n\rINVALID_VALUE\020\002\022\t\n\005ER"
    "ROR\020\003\"\032\n\030GetSessionOptionsRequest\"\330\001\n\027Ge"
    "tSessionOptionsResult\022[\n\017session_options"
    "\030\001 \003(\0132B.arrow.flight.protocol.GetSessio"
    "nOptionsResult.SessionOptionsEntry\032`\n\023Se"
    "ssionOptionsEntry\022\013\n\003key\030\001 \001(\t\0228\n\005value\030"
    "\002 \001(\0132).arrow.flight.protocol.SessionOpt"
    "ionValue:\0028\001\"\025\n\023CloseSessionRequest\"\235\001\n\022"
    "CloseSessionResult\022@\n\006status\030\001 \001(\01620.arr"
    "ow.flight.protocol.CloseSessionResult.St"
    "atus\"E\n\006Status\022\017\n\013UNSPECIFIED\020\000\022\n\n\006CLOSE"
    "D\020\001\022\013\n\007CLOSING\020\002\022\021\n\rNOT_CLOSEABLE\020\003*\213\001\n\014"
    "CancelStatus\022\035\n\031CANCEL_STATUS_UNSPECIFIE"
    "D\020\000\022\033\n\027CANCEL_STATUS_CANCELLED\020\001\022\034\n\030CANC"
    "EL_STATUS_CANCELLING\020\002\022!\n\035CANCEL_STATUS_"
    "NOT_CANCELLABLE\020\0032\205\007\n\rFlightService\022d\n\tH"
    "andshake\022\'.arrow.flight.protocol.Handsha"
    "keRequest\032(.arrow.flight.protocol.Handsh"
    "akeResponse\"\000(\0010\001\022U\n\013ListFlights\022\037.arrow"
    ".flight.protocol.Criteria\032!.arrow.flight"
    ".protocol.FlightInfo\"\0000\001\022]\n\rGetFlightInf"
    "o\022\'.arrow.flight.protocol.FlightDescript"
    "or\032!.arrow.flight.protocol.FlightInfo\"\000\022"
    "\\\n\016PollFlightInfo\022\'.arrow.flight.protoco"
    "l.FlightDescriptor\032\037.arrow.flight.protoc"
    "ol.PollInfo\"\000\022[\n\tGetSchema\022\'.arrow.fligh"
    "t.protocol.FlightDescriptor\032#.arrow.flig"
    "ht.protocol.SchemaResult\"\000\022M\n\005DoGet\022\035.ar"
    "row.flight.protocol.Ticket\032!.arrow.fligh"
    "t.protocol.FlightData\"\0000\001\022R\n\005DoPut\022!.arr"
    "ow.flight.protocol.FlightData\032 .arrow.fl"
    "ight.protocol.PutResult\"\000(\0010\001\022X\n\nDoExcha"
    "nge\022!.arrow.flight.protocol.FlightData\032!"
    ".arrow.flight.protocol.FlightData\"\000(\0010\001\022"
    "L\n\010DoAction\022\035.arrow.flight.protocol.Acti"
    "on\032\035.arrow.flight.protocol.Result\"\0000\001\022R\n"
    "\013ListActions\022\034.arrow.flight.protocol.Emp"
    "ty\032!.arrow.flight.protocol.ActionType\"\0000"
    "\001Bq\n\034org.apache.arrow.flight.implZ2githu"
    "b.com/apache/arrow-go/arrow/flight/gen/f"
    "light\252\002\034Apache.Arrow.Flight.Protocolb\006pr"
    "oto3"
};
static const ::_pbi::DescriptorTable* PROTOBUF_NONNULL const
    descriptor_table_Flight_2eproto_deps[1] = {
        &::descriptor_table_google_2fprotobuf_2ftimestamp_2eproto,
};
static ::absl::once_flag descriptor_table_Flight_2eproto_once;
PROTOBUF_CONSTINIT const ::_pbi::DescriptorTable descriptor_table_Flight_2eproto = {
    false,
    false,
    4164,
    descriptor_table_protodef_Flight_2eproto,
    "Flight.proto",
    &descriptor_table_Flight_2eproto_once,
    descriptor_table_Flight_2eproto_deps,
    1,
    32,
    schemas,
    file_default_instances,
    TableStruct_Flight_2eproto::offsets,
    file_level_enum_descriptors_Flight_2eproto,
    file_level_service_descriptors_Flight_2eproto,
};
namespace arrow {
namespace flight {
namespace protocol {
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
FlightDescriptor_DescriptorType_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_Flight_2eproto);
  return file_level_enum_descriptors_Flight_2eproto[0];
}
PROTOBUF_CONSTINIT const uint32_t FlightDescriptor_DescriptorType_internal_data_[] = {
    196608u, 0u, };
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
SetSessionOptionsResult_ErrorValue_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_Flight_2eproto);
  return file_level_enum_descriptors_Flight_2eproto[1];
}
PROTOBUF_CONSTINIT const uint32_t SetSessionOptionsResult_ErrorValue_internal_data_[] = {
    262144u, 0u, };
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
CloseSessionResult_Status_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_Flight_2eproto);
  return file_level_enum_descriptors_Flight_2eproto[2];
}
PROTOBUF_CONSTINIT const uint32_t CloseSessionResult_Status_internal_data_[] = {
    262144u, 0u, };
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
CancelStatus_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_Flight_2eproto);
  return file_level_enum_descriptors_Flight_2eproto[3];
}
PROTOBUF_CONSTINIT const uint32_t CancelStatus_internal_data_[] = {
    262144u, 0u, };
// ===================================================================

class HandshakeRequest::_Internal {
 public:
  using HasBits =
      decltype(::std::declval<HandshakeRequest>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(HandshakeRequest, _impl_._has_bits_);
};

HandshakeRequest::HandshakeRequest(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, HandshakeRequest_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.HandshakeRequest)
}
PROTOBUF_NDEBUG_INLINE HandshakeRequest::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::HandshakeRequest& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        payload_(arena, from.payload_) {}

HandshakeRequest::HandshakeRequest(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const HandshakeRequest& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, HandshakeRequest_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  HandshakeRequest* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  _impl_.protocol_version_ = from._impl_.protocol_version_;

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.HandshakeRequest)
}
PROTOBUF_NDEBUG_INLINE HandshakeRequest::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        payload_(arena) {}

inline void HandshakeRequest::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.protocol_version_ = {};
}
HandshakeRequest::~HandshakeRequest() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.HandshakeRequest)
  SharedDtor(*this);
}
inline void HandshakeRequest::SharedDtor(MessageLite& self) {
  HandshakeRequest& this_ = static_cast<HandshakeRequest&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.payload_.Destroy();
  this_._impl_.~Impl_();
}

inline void* PROTOBUF_NONNULL HandshakeRequest::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) HandshakeRequest(arena);
}
constexpr auto HandshakeRequest::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(HandshakeRequest),
                                            alignof(HandshakeRequest));
}
constexpr auto HandshakeRequest::InternalGenerateClassData_() {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &_HandshakeRequest_default_instance_._instance,
          &_table_.header,
          nullptr,  // IsInitialized
          &HandshakeRequest::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<HandshakeRequest>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &HandshakeRequest::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<HandshakeRequest>(), &HandshakeRequest::ByteSizeLong,
              &HandshakeRequest::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(HandshakeRequest, _impl_._cached_size_),
          false,
      },
      &HandshakeRequest::kDescriptorMethods,
      &descriptor_table_Flight_2eproto,
      nullptr,  // tracker
  };
}

PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull HandshakeRequest_class_data_ =
        HandshakeRequest::InternalGenerateClassData_();

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
HandshakeRequest::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&HandshakeRequest_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(HandshakeRequest_class_data_.tc_table);
  return HandshakeRequest_class_data_.base();
}
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<1, 2, 0, 0, 2>
HandshakeRequest::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(HandshakeRequest, _impl_._has_bits_),
    0, // no _extensions_
    2, 8,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967292,  // skipmap
    offsetof(decltype(_table_), field_entries),
    2,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    HandshakeRequest_class_data_.base(),
    nullptr,  // post_loop_handler
    ::_pbi::TcParser::GenericFallback,  // fallback
    #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
    ::_pbi::TcParser::GetTable<::arrow::flight::protocol::HandshakeRequest>(),  // to_prefetch
    #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
  }, {{
    // bytes payload = 2;
    {::_pbi::TcParser::FastBS1,
     {18, 0, 0,
      PROTOBUF_FIELD_OFFSET(HandshakeRequest, _impl_.payload_)}},
    // uint64 protocol_version = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(HandshakeRequest, _impl_.protocol_version_), 1>(),
     {8, 1, 0,
      PROTOBUF_FIELD_OFFSET(HandshakeRequest, _impl_.protocol_version_)}},
  }}, {{
    65535, 65535
  }}, {{
    // uint64 protocol_version = 1;
    {PROTOBUF_FIELD_OFFSET(HandshakeRequest, _impl_.protocol_version_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // bytes payload = 2;
    {PROTOBUF_FIELD_OFFSET(HandshakeRequest, _impl_.payload_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
  }},
  // no aux_entries
  {{
  }},
};
PROTOBUF_NOINLINE void HandshakeRequest::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.HandshakeRequest)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.payload_.ClearNonDefaultToEmpty();
  }
  _impl_.protocol_version_ = ::uint64_t{0u};
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL HandshakeRequest::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const HandshakeRequest& this_ = static_cast<const HandshakeRequest&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL HandshakeRequest::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const HandshakeRequest& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.HandshakeRequest)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // uint64 protocol_version = 1;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (this_._internal_protocol_version() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
          1, this_._internal_protocol_version(), target);
    }
  }

  // bytes payload = 2;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_payload().empty()) {
      const ::std::string& _s = this_._internal_payload();
      target = stream->WriteBytesMaybeAliased(2, _s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.HandshakeRequest)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t HandshakeRequest::ByteSizeLong(const MessageLite& base) {
  const HandshakeRequest& this_ = static_cast<const HandshakeRequest&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t HandshakeRequest::ByteSizeLong() const {
  const HandshakeRequest& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.HandshakeRequest)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    // bytes payload = 2;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_payload().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                        this_._internal_payload());
      }
    }
    // uint64 protocol_version = 1;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (this_._internal_protocol_version() != 0) {
        total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
            this_._internal_protocol_version());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void HandshakeRequest::MergeImpl(::google::protobuf::MessageLite& to_msg,
                            const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this =
      static_cast<HandshakeRequest*>(&to_msg);
  auto& from = static_cast<const HandshakeRequest&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.HandshakeRequest)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!from._internal_payload().empty()) {
        _this->_internal_set_payload(from._internal_payload());
      } else {
        if (_this->_impl_.payload_.IsDefault()) {
          _this->_internal_set_payload("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (from._internal_protocol_version() != 0) {
        _this->_impl_.protocol_version_ = from._impl_.protocol_version_;
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void HandshakeRequest::CopyFrom(const HandshakeRequest& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.HandshakeRequest)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void HandshakeRequest::InternalSwap(HandshakeRequest* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.payload_, &other->_impl_.payload_, arena);
  swap(_impl_.protocol_version_, other->_impl_.protocol_version_);
}

::google::protobuf::Metadata HandshakeRequest::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

class HandshakeResponse::_Internal {
 public:
  using HasBits =
      decltype(::std::declval<HandshakeResponse>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(HandshakeResponse, _impl_._has_bits_);
};

HandshakeResponse::HandshakeResponse(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, HandshakeResponse_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.HandshakeResponse)
}
PROTOBUF_NDEBUG_INLINE HandshakeResponse::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::HandshakeResponse& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        payload_(arena, from.payload_) {}

HandshakeResponse::HandshakeResponse(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const HandshakeResponse& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, HandshakeResponse_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  HandshakeResponse* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  _impl_.protocol_version_ = from._impl_.protocol_version_;

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.HandshakeResponse)
}
PROTOBUF_NDEBUG_INLINE HandshakeResponse::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        payload_(arena) {}

inline void HandshakeResponse::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.protocol_version_ = {};
}
HandshakeResponse::~HandshakeResponse() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.HandshakeResponse)
  SharedDtor(*this);
}
inline void HandshakeResponse::SharedDtor(MessageLite& self) {
  HandshakeResponse& this_ = static_cast<HandshakeResponse&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.payload_.Destroy();
  this_._impl_.~Impl_();
}

inline void* PROTOBUF_NONNULL HandshakeResponse::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) HandshakeResponse(arena);
}
constexpr auto HandshakeResponse::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(HandshakeResponse),
                                            alignof(HandshakeResponse));
}
constexpr auto HandshakeResponse::InternalGenerateClassData_() {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &_HandshakeResponse_default_instance_._instance,
          &_table_.header,
          nullptr,  // IsInitialized
          &HandshakeResponse::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<HandshakeResponse>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &HandshakeResponse::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<HandshakeResponse>(), &HandshakeResponse::ByteSizeLong,
              &HandshakeResponse::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(HandshakeResponse, _impl_._cached_size_),
          false,
      },
      &HandshakeResponse::kDescriptorMethods,
      &descriptor_table_Flight_2eproto,
      nullptr,  // tracker
  };
}

PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull HandshakeResponse_class_data_ =
        HandshakeResponse::InternalGenerateClassData_();

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
HandshakeResponse::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&HandshakeResponse_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(HandshakeResponse_class_data_.tc_table);
  return HandshakeResponse_class_data_.base();
}
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<1, 2, 0, 0, 2>
HandshakeResponse::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(HandshakeResponse, _impl_._has_bits_),
    0, // no _extensions_
    2, 8,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967292,  // skipmap
    offsetof(decltype(_table_), field_entries),
    2,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    HandshakeResponse_class_data_.base(),
    nullptr,  // post_loop_handler
    ::_pbi::TcParser::GenericFallback,  // fallback
    #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
    ::_pbi::TcParser::GetTable<::arrow::flight::protocol::HandshakeResponse>(),  // to_prefetch
    #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
  }, {{
    // bytes payload = 2;
    {::_pbi::TcParser::FastBS1,
     {18, 0, 0,
      PROTOBUF_FIELD_OFFSET(HandshakeResponse, _impl_.payload_)}},
    // uint64 protocol_version = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(HandshakeResponse, _impl_.protocol_version_), 1>(),
     {8, 1, 0,
      PROTOBUF_FIELD_OFFSET(HandshakeResponse, _impl_.protocol_version_)}},
  }}, {{
    65535, 65535
  }}, {{
    // uint64 protocol_version = 1;
    {PROTOBUF_FIELD_OFFSET(HandshakeResponse, _impl_.protocol_version_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // bytes payload = 2;
    {PROTOBUF_FIELD_OFFSET(HandshakeResponse, _impl_.payload_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
  }},
  // no aux_entries
  {{
  }},
};
PROTOBUF_NOINLINE void HandshakeResponse::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.HandshakeResponse)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.payload_.ClearNonDefaultToEmpty();
  }
  _impl_.protocol_version_ = ::uint64_t{0u};
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL HandshakeResponse::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const HandshakeResponse& this_ = static_cast<const HandshakeResponse&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL HandshakeResponse::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const HandshakeResponse& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.HandshakeResponse)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // uint64 protocol_version = 1;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (this_._internal_protocol_version() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
          1, this_._internal_protocol_version(), target);
    }
  }

  // bytes payload = 2;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_payload().empty()) {
      const ::std::string& _s = this_._internal_payload();
      target = stream->WriteBytesMaybeAliased(2, _s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.HandshakeResponse)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t HandshakeResponse::ByteSizeLong(const MessageLite& base) {
  const HandshakeResponse& this_ = static_cast<const HandshakeResponse&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t HandshakeResponse::ByteSizeLong() const {
  const HandshakeResponse& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.HandshakeResponse)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    // bytes payload = 2;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_payload().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                        this_._internal_payload());
      }
    }
    // uint64 protocol_version = 1;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (this_._internal_protocol_version() != 0) {
        total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
            this_._internal_protocol_version());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void HandshakeResponse::MergeImpl(::google::protobuf::MessageLite& to_msg,
                            const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this =
      static_cast<HandshakeResponse*>(&to_msg);
  auto& from = static_cast<const HandshakeResponse&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.HandshakeResponse)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!from._internal_payload().empty()) {
        _this->_internal_set_payload(from._internal_payload());
      } else {
        if (_this->_impl_.payload_.IsDefault()) {
          _this->_internal_set_payload("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (from._internal_protocol_version() != 0) {
        _this->_impl_.protocol_version_ = from._impl_.protocol_version_;
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void HandshakeResponse::CopyFrom(const HandshakeResponse& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.HandshakeResponse)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void HandshakeResponse::InternalSwap(HandshakeResponse* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.payload_, &other->_impl_.payload_, arena);
  swap(_impl_.protocol_version_, other->_impl_.protocol_version_);
}

::google::protobuf::Metadata HandshakeResponse::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

class BasicAuth::_Internal {
 public:
  using HasBits =
      decltype(::std::declval<BasicAuth>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(BasicAuth, _impl_._has_bits_);
};

BasicAuth::BasicAuth(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, BasicAuth_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.BasicAuth)
}
PROTOBUF_NDEBUG_INLINE BasicAuth::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::BasicAuth& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        username_(arena, from.username_),
        password_(arena, from.password_) {}

BasicAuth::BasicAuth(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const BasicAuth& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, BasicAuth_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  BasicAuth* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.BasicAuth)
}
PROTOBUF_NDEBUG_INLINE BasicAuth::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        username_(arena),
        password_(arena) {}

inline void BasicAuth::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
BasicAuth::~BasicAuth() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.BasicAuth)
  SharedDtor(*this);
}
inline void BasicAuth::SharedDtor(MessageLite& self) {
  BasicAuth& this_ = static_cast<BasicAuth&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.username_.Destroy();
  this_._impl_.password_.Destroy();
  this_._impl_.~Impl_();
}

inline void* PROTOBUF_NONNULL BasicAuth::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) BasicAuth(arena);
}
constexpr auto BasicAuth::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(BasicAuth),
                                            alignof(BasicAuth));
}
constexpr auto BasicAuth::InternalGenerateClassData_() {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &_BasicAuth_default_instance_._instance,
          &_table_.header,
          nullptr,  // IsInitialized
          &BasicAuth::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<BasicAuth>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &BasicAuth::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<BasicAuth>(), &BasicAuth::ByteSizeLong,
              &BasicAuth::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(BasicAuth, _impl_._cached_size_),
          false,
      },
      &BasicAuth::kDescriptorMethods,
      &descriptor_table_Flight_2eproto,
      nullptr,  // tracker
  };
}

PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull BasicAuth_class_data_ =
        BasicAuth::InternalGenerateClassData_();

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
BasicAuth::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&BasicAuth_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(BasicAuth_class_data_.tc_table);
  return BasicAuth_class_data_.base();
}
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<1, 2, 0, 56, 2>
BasicAuth::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(BasicAuth, _impl_._has_bits_),
    0, // no _extensions_
    3, 8,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967289,  // skipmap
    offsetof(decltype(_table_), field_entries),
    2,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    BasicAuth_class_data_.base(),
    nullptr,  // post_loop_handler
    ::_pbi::TcParser::GenericFallback,  // fallback
    #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
    ::_pbi::TcParser::GetTable<::arrow::flight::protocol::BasicAuth>(),  // to_prefetch
    #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
  }, {{
    // string username = 2;
    {::_pbi::TcParser::FastUS1,
     {18, 0, 0,
      PROTOBUF_FIELD_OFFSET(BasicAuth, _impl_.username_)}},
    // string password = 3;
    {::_pbi::TcParser::FastUS1,
     {26, 1, 0,
      PROTOBUF_FIELD_OFFSET(BasicAuth, _impl_.password_)}},
  }}, {{
    65535, 65535
  }}, {{
    // string username = 2;
    {PROTOBUF_FIELD_OFFSET(BasicAuth, _impl_.username_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
    // string password = 3;
    {PROTOBUF_FIELD_OFFSET(BasicAuth, _impl_.password_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
  }},
  // no aux_entries
  {{
    "\37\10\10\0\0\0\0\0"
    "arrow.flight.protocol.BasicAuth"
    "username"
    "password"
  }},
};
PROTOBUF_NOINLINE void BasicAuth::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.BasicAuth)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.username_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.password_.ClearNonDefaultToEmpty();
    }
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL BasicAuth::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const BasicAuth& this_ = static_cast<const BasicAuth&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL BasicAuth::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const BasicAuth& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.BasicAuth)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // string username = 2;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_username().empty()) {
      const ::std::string& _s = this_._internal_username();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.BasicAuth.username");
      target = stream->WriteStringMaybeAliased(2, _s, target);
    }
  }

  // string password = 3;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (!this_._internal_password().empty()) {
      const ::std::string& _s = this_._internal_password();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.BasicAuth.password");
      target = stream->WriteStringMaybeAliased(3, _s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.BasicAuth)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t BasicAuth::ByteSizeLong(const MessageLite& base) {
  const BasicAuth& this_ = static_cast<const BasicAuth&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t BasicAuth::ByteSizeLong() const {
  const BasicAuth& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.BasicAuth)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    // string username = 2;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_username().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_username());
      }
    }
    // string password = 3;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!this_._internal_password().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_password());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void BasicAuth::MergeImpl(::google::protobuf::MessageLite& to_msg,
                            const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this =
      static_cast<BasicAuth*>(&to_msg);
  auto& from = static_cast<const BasicAuth&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.BasicAuth)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!from._internal_username().empty()) {
        _this->_internal_set_username(from._internal_username());
      } else {
        if (_this->_impl_.username_.IsDefault()) {
          _this->_internal_set_username("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!from._internal_password().empty()) {
        _this->_internal_set_password(from._internal_password());
      } else {
        if (_this->_impl_.password_.IsDefault()) {
          _this->_internal_set_password("");
        }
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void BasicAuth::CopyFrom(const BasicAuth& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.BasicAuth)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void BasicAuth::InternalSwap(BasicAuth* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.username_, &other->_impl_.username_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.password_, &other->_impl_.password_, arena);
}

::google::protobuf::Metadata BasicAuth::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

class Empty::_Internal {
 public:
};

Empty::Empty(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::internal::ZeroFieldsBase(arena, Empty_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::internal::ZeroFieldsBase(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.Empty)
}
Empty::Empty(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const Empty& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::internal::ZeroFieldsBase(arena, Empty_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::internal::ZeroFieldsBase(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  Empty* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.Empty)
}

inline void* PROTOBUF_NONNULL Empty::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) Empty(arena);
}
constexpr auto Empty::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::ZeroInit(sizeof(Empty),
                                            alignof(Empty));
}
constexpr auto Empty::InternalGenerateClassData_() {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &_Empty_default_instance_._instance,
          &_table_.header,
          nullptr,  // IsInitialized
          &Empty::MergeImpl,
          ::google::protobuf::internal::ZeroFieldsBase::GetNewImpl<Empty>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &Empty::SharedDtor,
          ::google::protobuf::internal::ZeroFieldsBase::GetClearImpl<Empty>(), &Empty::ByteSizeLong,
              &Empty::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(Empty, _impl_._cached_size_),
          false,
      },
      &Empty::kDescriptorMethods,
      &descriptor_table_Flight_2eproto,
      nullptr,  // tracker
  };
}

PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull Empty_class_data_ =
        Empty::InternalGenerateClassData_();

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Empty::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Empty_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(Empty_class_data_.tc_table);
  return Empty_class_data_.base();
}
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 0, 0, 0, 2>
Empty::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(Empty,
                          _impl_._cached_size_),  // no hasbits
    0, // no _extensions_
    0, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967295,  // skipmap
    offsetof(decltype(_table_), field_names),  // no field_entries
    0,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    Empty_class_data_.base(),
    nullptr,  // post_loop_handler
    ::_pbi::TcParser::GenericFallback,  // fallback
    #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
    ::_pbi::TcParser::GetTable<::arrow::flight::protocol::Empty>(),  // to_prefetch
    #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
  }, {{
    {::_pbi::TcParser::MiniParse, {}},
  }}, {{
    65535, 65535
  }}, // no field_entries, or aux_entries
  {{
  }},
};







::google::protobuf::Metadata Empty::GetMetadata() const {
  return ::google::protobuf::internal::ZeroFieldsBase::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

class ActionType::_Internal {
 public:
  using HasBits =
      decltype(::std::declval<ActionType>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(ActionType, _impl_._has_bits_);
};

ActionType::ActionType(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, ActionType_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.ActionType)
}
PROTOBUF_NDEBUG_INLINE ActionType::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::ActionType& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        type_(arena, from.type_),
        description_(arena, from.description_) {}

ActionType::ActionType(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const ActionType& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, ActionType_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  ActionType* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.ActionType)
}
PROTOBUF_NDEBUG_INLINE ActionType::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        type_(arena),
        description_(arena) {}

inline void ActionType::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
ActionType::~ActionType() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.ActionType)
  SharedDtor(*this);
}
inline void ActionType::SharedDtor(MessageLite& self) {
  ActionType& this_ = static_cast<ActionType&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.type_.Destroy();
  this_._impl_.description_.Destroy();
  this_._impl_.~Impl_();
}

inline void* PROTOBUF_NONNULL ActionType::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) ActionType(arena);
}
constexpr auto ActionType::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(ActionType),
                                            alignof(ActionType));
}
constexpr auto ActionType::InternalGenerateClassData_() {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &_ActionType_default_instance_._instance,
          &_table_.header,
          nullptr,  // IsInitialized
          &ActionType::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<ActionType>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &ActionType::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<ActionType>(), &ActionType::ByteSizeLong,
              &ActionType::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(ActionType, _impl_._cached_size_),
          false,
      },
      &ActionType::kDescriptorMethods,
      &descriptor_table_Flight_2eproto,
      nullptr,  // tracker
  };
}

PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull ActionType_class_data_ =
        ActionType::InternalGenerateClassData_();

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ActionType::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ActionType_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(ActionType_class_data_.tc_table);
  return ActionType_class_data_.base();
}
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<1, 2, 0, 56, 2>
ActionType::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(ActionType, _impl_._has_bits_),
    0, // no _extensions_
    2, 8,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967292,  // skipmap
    offsetof(decltype(_table_), field_entries),
    2,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    ActionType_class_data_.base(),
    nullptr,  // post_loop_handler
    ::_pbi::TcParser::GenericFallback,  // fallback
    #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
    ::_pbi::TcParser::GetTable<::arrow::flight::protocol::ActionType>(),  // to_prefetch
    #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
  }, {{
    // string description = 2;
    {::_pbi::TcParser::FastUS1,
     {18, 1, 0,
      PROTOBUF_FIELD_OFFSET(ActionType, _impl_.description_)}},
    // string type = 1;
    {::_pbi::TcParser::FastUS1,
     {10, 0, 0,
      PROTOBUF_FIELD_OFFSET(ActionType, _impl_.type_)}},
  }}, {{
    65535, 65535
  }}, {{
    // string type = 1;
    {PROTOBUF_FIELD_OFFSET(ActionType, _impl_.type_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
    // string description = 2;
    {PROTOBUF_FIELD_OFFSET(ActionType, _impl_.description_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
  }},
  // no aux_entries
  {{
    "\40\4\13\0\0\0\0\0"
    "arrow.flight.protocol.ActionType"
    "type"
    "description"
  }},
};
PROTOBUF_NOINLINE void ActionType::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.ActionType)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.type_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.description_.ClearNonDefaultToEmpty();
    }
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL ActionType::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const ActionType& this_ = static_cast<const ActionType&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL ActionType::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const ActionType& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.ActionType)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // string type = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_type().empty()) {
      const ::std::string& _s = this_._internal_type();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.ActionType.type");
      target = stream->WriteStringMaybeAliased(1, _s, target);
    }
  }

  // string description = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (!this_._internal_description().empty()) {
      const ::std::string& _s = this_._internal_description();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.ActionType.description");
      target = stream->WriteStringMaybeAliased(2, _s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.ActionType)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t ActionType::ByteSizeLong(const MessageLite& base) {
  const ActionType& this_ = static_cast<const ActionType&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t ActionType::ByteSizeLong() const {
  const ActionType& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.ActionType)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    // string type = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_type().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_type());
      }
    }
    // string description = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!this_._internal_description().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_description());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void ActionType::MergeImpl(::google::protobuf::MessageLite& to_msg,
                            const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this =
      static_cast<ActionType*>(&to_msg);
  auto& from = static_cast<const ActionType&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.ActionType)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!from._internal_type().empty()) {
        _this->_internal_set_type(from._internal_type());
      } else {
        if (_this->_impl_.type_.IsDefault()) {
          _this->_internal_set_type("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!from._internal_description().empty()) {
        _this->_internal_set_description(from._internal_description());
      } else {
        if (_this->_impl_.description_.IsDefault()) {
          _this->_internal_set_description("");
        }
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void ActionType::CopyFrom(const ActionType& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.ActionType)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void ActionType::InternalSwap(ActionType* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.type_, &other->_impl_.type_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.description_, &other->_impl_.description_, arena);
}

::google::protobuf::Metadata ActionType::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

class Criteria::_Internal {
 public:
  using HasBits =
      decltype(::std::declval<Criteria>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(Criteria, _impl_._has_bits_);
};

Criteria::Criteria(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Criteria_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.Criteria)
}
PROTOBUF_NDEBUG_INLINE Criteria::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::Criteria& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        expression_(arena, from.expression_) {}

Criteria::Criteria(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const Criteria& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Criteria_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  Criteria* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.Criteria)
}
PROTOBUF_NDEBUG_INLINE Criteria::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        expression_(arena) {}

inline void Criteria::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
Criteria::~Criteria() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.Criteria)
  SharedDtor(*this);
}
inline void Criteria::SharedDtor(MessageLite& self) {
  Criteria& this_ = static_cast<Criteria&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.expression_.Destroy();
  this_._impl_.~Impl_();
}

inline void* PROTOBUF_NONNULL Criteria::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) Criteria(arena);
}
constexpr auto Criteria::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(Criteria),
                                            alignof(Criteria));
}
constexpr auto Criteria::InternalGenerateClassData_() {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &_Criteria_default_instance_._instance,
          &_table_.header,
          nullptr,  // IsInitialized
          &Criteria::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<Criteria>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &Criteria::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<Criteria>(), &Criteria::ByteSizeLong,
              &Criteria::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(Criteria, _impl_._cached_size_),
          false,
      },
      &Criteria::kDescriptorMethods,
      &descriptor_table_Flight_2eproto,
      nullptr,  // tracker
  };
}

PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull Criteria_class_data_ =
        Criteria::InternalGenerateClassData_();

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Criteria::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Criteria_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(Criteria_class_data_.tc_table);
  return Criteria_class_data_.base();
}
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2>
Criteria::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(Criteria, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    Criteria_class_data_.base(),
    nullptr,  // post_loop_handler
    ::_pbi::TcParser::GenericFallback,  // fallback
    #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
    ::_pbi::TcParser::GetTable<::arrow::flight::protocol::Criteria>(),  // to_prefetch
    #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
  }, {{
    // bytes expression = 1;
    {::_pbi::TcParser::FastBS1,
     {10, 0, 0,
      PROTOBUF_FIELD_OFFSET(Criteria, _impl_.expression_)}},
  }}, {{
    65535, 65535
  }}, {{
    // bytes expression = 1;
    {PROTOBUF_FIELD_OFFSET(Criteria, _impl_.expression_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
  }},
  // no aux_entries
  {{
  }},
};
PROTOBUF_NOINLINE void Criteria::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.Criteria)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.expression_.ClearNonDefaultToEmpty();
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL Criteria::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const Criteria& this_ = static_cast<const Criteria&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL Criteria::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const Criteria& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.Criteria)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // bytes expression = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_expression().empty()) {
      const ::std::string& _s = this_._internal_expression();
      target = stream->WriteBytesMaybeAliased(1, _s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.Criteria)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t Criteria::ByteSizeLong(const MessageLite& base) {
  const Criteria& this_ = static_cast<const Criteria&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t Criteria::ByteSizeLong() const {
  const Criteria& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.Criteria)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

   {
    // bytes expression = 1;
    cached_has_bits = this_._impl_._has_bits_[0];
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_expression().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                        this_._internal_expression());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void Criteria::MergeImpl(::google::protobuf::MessageLite& to_msg,
                            const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this =
      static_cast<Criteria*>(&to_msg);
  auto& from = static_cast<const Criteria&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.Criteria)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!from._internal_expression().empty()) {
      _this->_internal_set_expression(from._internal_expression());
    } else {
      if (_this->_impl_.expression_.IsDefault()) {
        _this->_internal_set_expression("");
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void Criteria::CopyFrom(const Criteria& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.Criteria)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void Criteria::InternalSwap(Criteria* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.expression_, &other->_impl_.expression_, arena);
}

::google::protobuf::Metadata Criteria::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

class Action::_Internal {
 public:
  using HasBits =
      decltype(::std::declval<Action>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(Action, _impl_._has_bits_);
};

Action::Action(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Action_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.Action)
}
PROTOBUF_NDEBUG_INLINE Action::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::Action& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        type_(arena, from.type_),
        body_(arena, from.body_) {}

Action::Action(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const Action& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Action_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  Action* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.Action)
}
PROTOBUF_NDEBUG_INLINE Action::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        type_(arena),
        body_(arena) {}

inline void Action::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
Action::~Action() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.Action)
  SharedDtor(*this);
}
inline void Action::SharedDtor(MessageLite& self) {
  Action& this_ = static_cast<Action&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.type_.Destroy();
  this_._impl_.body_.Destroy();
  this_._impl_.~Impl_();
}

inline void* PROTOBUF_NONNULL Action::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) Action(arena);
}
constexpr auto Action::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(Action),
                                            alignof(Action));
}
constexpr auto Action::InternalGenerateClassData_() {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &_Action_default_instance_._instance,
          &_table_.header,
          nullptr,  // IsInitialized
          &Action::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<Action>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &Action::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<Action>(), &Action::ByteSizeLong,
              &Action::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(Action, _impl_._cached_size_),
          false,
      },
      &Action::kDescriptorMethods,
      &descriptor_table_Flight_2eproto,
      nullptr,  // tracker
  };
}

PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull Action_class_data_ =
        Action::InternalGenerateClassData_();

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Action::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Action_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(Action_class_data_.tc_table);
  return Action_class_data_.base();
}
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<1, 2, 0, 41, 2>
Action::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(Action, _impl_._has_bits_),
    0, // no _extensions_
    2, 8,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967292,  // skipmap
    offsetof(decltype(_table_), field_entries),
    2,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    Action_class_data_.base(),
    nullptr,  // post_loop_handler
    ::_pbi::TcParser::GenericFallback,  // fallback
    #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
    ::_pbi::TcParser::GetTable<::arrow::flight::protocol::Action>(),  // to_prefetch
    #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
  }, {{
    // bytes body = 2;
    {::_pbi::TcParser::FastBS1,
     {18, 1, 0,
      PROTOBUF_FIELD_OFFSET(Action, _impl_.body_)}},
    // string type = 1;
    {::_pbi::TcParser::FastUS1,
     {10, 0, 0,
      PROTOBUF_FIELD_OFFSET(Action, _impl_.type_)}},
  }}, {{
    65535, 65535
  }}, {{
    // string type = 1;
    {PROTOBUF_FIELD_OFFSET(Action, _impl_.type_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
    // bytes body = 2;
    {PROTOBUF_FIELD_OFFSET(Action, _impl_.body_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
  }},
  // no aux_entries
  {{
    "\34\4\0\0\0\0\0\0"
    "arrow.flight.protocol.Action"
    "type"
  }},
};
PROTOBUF_NOINLINE void Action::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.Action)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.type_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.body_.ClearNonDefaultToEmpty();
    }
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL Action::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const Action& this_ = static_cast<const Action&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL Action::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const Action& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.Action)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // string type = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_type().empty()) {
      const ::std::string& _s = this_._internal_type();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.Action.type");
      target = stream->WriteStringMaybeAliased(1, _s, target);
    }
  }

  // bytes body = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (!this_._internal_body().empty()) {
      const ::std::string& _s = this_._internal_body();
      target = stream->WriteBytesMaybeAliased(2, _s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.Action)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t Action::ByteSizeLong(const MessageLite& base) {
  const Action& this_ = static_cast<const Action&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t Action::ByteSizeLong() const {
  const Action& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.Action)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    // string type = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_type().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_type());
      }
    }
    // bytes body = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!this_._internal_body().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                        this_._internal_body());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void Action::MergeImpl(::google::protobuf::MessageLite& to_msg,
                            const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this =
      static_cast<Action*>(&to_msg);
  auto& from = static_cast<const Action&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.Action)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!from._internal_type().empty()) {
        _this->_internal_set_type(from._internal_type());
      } else {
        if (_this->_impl_.type_.IsDefault()) {
          _this->_internal_set_type("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!from._internal_body().empty()) {
        _this->_internal_set_body(from._internal_body());
      } else {
        if (_this->_impl_.body_.IsDefault()) {
          _this->_internal_set_body("");
        }
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void Action::CopyFrom(const Action& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.Action)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void Action::InternalSwap(Action* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.type_, &other->_impl_.type_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.body_, &other->_impl_.body_, arena);
}

::google::protobuf::Metadata Action::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

class Result::_Internal {
 public:
  using HasBits =
      decltype(::std::declval<Result>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(Result, _impl_._has_bits_);
};

Result::Result(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Result_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.Result)
}
PROTOBUF_NDEBUG_INLINE Result::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::Result& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        body_(arena, from.body_) {}

Result::Result(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const Result& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Result_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  Result* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.Result)
}
PROTOBUF_NDEBUG_INLINE Result::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        body_(arena) {}

inline void Result::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
Result::~Result() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.Result)
  SharedDtor(*this);
}
inline void Result::SharedDtor(MessageLite& self) {
  Result& this_ = static_cast<Result&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.body_.Destroy();
  this_._impl_.~Impl_();
}

inline void* PROTOBUF_NONNULL Result::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) Result(arena);
}
constexpr auto Result::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(Result),
                                            alignof(Result));
}
constexpr auto Result::InternalGenerateClassData_() {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &_Result_default_instance_._instance,
          &_table_.header,
          nullptr,  // IsInitialized
          &Result::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<Result>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &Result::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<Result>(), &Result::ByteSizeLong,
              &Result::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(Result, _impl_._cached_size_),
          false,
      },
      &Result::kDescriptorMethods,
      &descriptor_table_Flight_2eproto,
      nullptr,  // tracker
  };
}

PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull Result_class_data_ =
        Result::InternalGenerateClassData_();

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Result::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Result_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(Result_class_data_.tc_table);
  return Result_class_data_.base();
}
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2>
Result::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(Result, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    Result_class_data_.base(),
    nullptr,  // post_loop_handler
    ::_pbi::TcParser::GenericFallback,  // fallback
    #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
    ::_pbi::TcParser::GetTable<::arrow::flight::protocol::Result>(),  // to_prefetch
    #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
  }, {{
    // bytes body = 1;
    {::_pbi::TcParser::FastBS1,
     {10, 0, 0,
      PROTOBUF_FIELD_OFFSET(Result, _impl_.body_)}},
  }}, {{
    65535, 65535
  }}, {{
    // bytes body = 1;
    {PROTOBUF_FIELD_OFFSET(Result, _impl_.body_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
  }},
  // no aux_entries
  {{
  }},
};
PROTOBUF_NOINLINE void Result::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.Result)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.body_.ClearNonDefaultToEmpty();
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL Result::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const Result& this_ = static_cast<const Result&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL Result::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const Result& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.Result)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // bytes body = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_body().empty()) {
      const ::std::string& _s = this_._internal_body();
      target = stream->WriteBytesMaybeAliased(1, _s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.Result)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t Result::ByteSizeLong(const MessageLite& base) {
  const Result& this_ = static_cast<const Result&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t Result::ByteSizeLong() const {
  const Result& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.Result)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

   {
    // bytes body = 1;
    cached_has_bits = this_._impl_._has_bits_[0];
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_body().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                        this_._internal_body());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void Result::MergeImpl(::google::protobuf::MessageLite& to_msg,
                            const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this =
      static_cast<Result*>(&to_msg);
  auto& from = static_cast<const Result&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.Result)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!from._internal_body().empty()) {
      _this->_internal_set_body(from._internal_body());
    } else {
      if (_this->_impl_.body_.IsDefault()) {
        _this->_internal_set_body("");
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void Result::CopyFrom(const Result& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.Result)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void Result::InternalSwap(Result* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.body_, &other->_impl_.body_, arena);
}

::google::protobuf::Metadata Result::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

class SchemaResult::_Internal {
 public:
  using HasBits =
      decltype(::std::declval<SchemaResult>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(SchemaResult, _impl_._has_bits_);
};

SchemaResult::SchemaResult(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, SchemaResult_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.SchemaResult)
}
PROTOBUF_NDEBUG_INLINE SchemaResult::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::SchemaResult& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        schema_(arena, from.schema_) {}

SchemaResult::SchemaResult(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const SchemaResult& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, SchemaResult_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SchemaResult* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.SchemaResult)
}
PROTOBUF_NDEBUG_INLINE SchemaResult::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        schema_(arena) {}

inline void SchemaResult::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
SchemaResult::~SchemaResult() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.SchemaResult)
  SharedDtor(*this);
}
inline void SchemaResult::SharedDtor(MessageLite& self) {
  SchemaResult& this_ = static_cast<SchemaResult&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.schema_.Destroy();
  this_._impl_.~Impl_();
}

inline void* PROTOBUF_NONNULL SchemaResult::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) SchemaResult(arena);
}
constexpr auto SchemaResult::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(SchemaResult),
                                            alignof(SchemaResult));
}
constexpr auto SchemaResult::InternalGenerateClassData_() {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &_SchemaResult_default_instance_._instance,
          &_table_.header,
          nullptr,  // IsInitialized
          &SchemaResult::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<SchemaResult>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &SchemaResult::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<SchemaResult>(), &SchemaResult::ByteSizeLong,
              &SchemaResult::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(SchemaResult, _impl_._cached_size_),
          false,
      },
      &SchemaResult::kDescriptorMethods,
      &descriptor_table_Flight_2eproto,
      nullptr,  // tracker
  };
}

PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull SchemaResult_class_data_ =
        SchemaResult::InternalGenerateClassData_();

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
SchemaResult::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&SchemaResult_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(SchemaResult_class_data_.tc_table);
  return SchemaResult_class_data_.base();
}
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2>
SchemaResult::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(SchemaResult, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    SchemaResult_class_data_.base(),
    nullptr,  // post_loop_handler
    ::_pbi::TcParser::GenericFallback,  // fallback
    #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
    ::_pbi::TcParser::GetTable<::arrow::flight::protocol::SchemaResult>(),  // to_prefetch
    #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
  }, {{
    // bytes schema = 1;
    {::_pbi::TcParser::FastBS1,
     {10, 0, 0,
      PROTOBUF_FIELD_OFFSET(SchemaResult, _impl_.schema_)}},
  }}, {{
    65535, 65535
  }}, {{
    // bytes schema = 1;
    {PROTOBUF_FIELD_OFFSET(SchemaResult, _impl_.schema_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
  }},
  // no aux_entries
  {{
  }},
};
PROTOBUF_NOINLINE void SchemaResult::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.SchemaResult)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.schema_.ClearNonDefaultToEmpty();
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL SchemaResult::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const SchemaResult& this_ = static_cast<const SchemaResult&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL SchemaResult::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const SchemaResult& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.SchemaResult)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // bytes schema = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_schema().empty()) {
      const ::std::string& _s = this_._internal_schema();
      target = stream->WriteBytesMaybeAliased(1, _s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.SchemaResult)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t SchemaResult::ByteSizeLong(const MessageLite& base) {
  const SchemaResult& this_ = static_cast<const SchemaResult&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t SchemaResult::ByteSizeLong() const {
  const SchemaResult& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.SchemaResult)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

   {
    // bytes schema = 1;
    cached_has_bits = this_._impl_._has_bits_[0];
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_schema().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                        this_._internal_schema());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void SchemaResult::MergeImpl(::google::protobuf::MessageLite& to_msg,
                            const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this =
      static_cast<SchemaResult*>(&to_msg);
  auto& from = static_cast<const SchemaResult&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.SchemaResult)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!from._internal_schema().empty()) {
      _this->_internal_set_schema(from._internal_schema());
    } else {
      if (_this->_impl_.schema_.IsDefault()) {
        _this->_internal_set_schema("");
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void SchemaResult::CopyFrom(const SchemaResult& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.SchemaResult)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void SchemaResult::InternalSwap(SchemaResult* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.schema_, &other->_impl_.schema_, arena);
}

::google::protobuf::Metadata SchemaResult::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

class FlightDescriptor::_Internal {
 public:
  using HasBits =
      decltype(::std::declval<FlightDescriptor>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(FlightDescriptor, _impl_._has_bits_);
};

FlightDescriptor::FlightDescriptor(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, FlightDescriptor_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.FlightDescriptor)
}
PROTOBUF_NDEBUG_INLINE FlightDescriptor::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::FlightDescriptor& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        #ifdef PROTOBUF_INTERNAL_REMOVE_ARENA_PTRS_REPEATED_PTR_FIELD
        path_{visibility, ::_pbi::InternalMetadataOffset::Build<
            ::arrow::flight::protocol::FlightDescriptor,
            PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::FlightDescriptor, _impl_.path_)>()
        , from.path_}
        #else
        path_ { visibility, arena, from.path_ }
        #endif
        ,
        cmd_(arena, from.cmd_) {}

FlightDescriptor::FlightDescriptor(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const FlightDescriptor& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, FlightDescriptor_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  FlightDescriptor* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  _impl_.type_ = from._impl_.type_;

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.FlightDescriptor)
}
PROTOBUF_NDEBUG_INLINE FlightDescriptor::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        #ifdef PROTOBUF_INTERNAL_REMOVE_ARENA_PTRS_REPEATED_PTR_FIELD
        path_{visibility, ::_pbi::InternalMetadataOffset::Build<
            ::arrow::flight::protocol::FlightDescriptor,
            PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::FlightDescriptor, _impl_.path_)>()
        }
        #else
        path_ { visibility, arena }
        #endif
        ,
        cmd_(arena) {}

inline void FlightDescriptor::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.type_ = {};
}
FlightDescriptor::~FlightDescriptor() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.FlightDescriptor)
  SharedDtor(*this);
}
inline void FlightDescriptor::SharedDtor(MessageLite& self) {
  FlightDescriptor& this_ = static_cast<FlightDescriptor&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.cmd_.Destroy();
  this_._impl_.~Impl_();
}

inline void* PROTOBUF_NONNULL FlightDescriptor::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) FlightDescriptor(arena);
}
#ifdef PROTOBUF_INTERNAL_REMOVE_ARENA_PTRS_REPEATED_PTR_FIELD
constexpr auto FlightDescriptor::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(FlightDescriptor),
                                            alignof(FlightDescriptor));
}
#else  // !PROTOBUF_INTERNAL_REMOVE_ARENA_PTRS_REPEATED_PTR_FIELD
constexpr auto FlightDescriptor::InternalNewImpl_() {
  constexpr auto arena_bits = ::google::protobuf::internal::EncodePlacementArenaOffsets({
      PROTOBUF_FIELD_OFFSET(FlightDescriptor, _impl_.path_) +
          decltype(FlightDescriptor::_impl_.path_)::
              InternalGetArenaOffset(
                  ::google::protobuf::Message::internal_visibility()),
  });
  if (arena_bits.has_value()) {
    return ::google::protobuf::internal::MessageCreator::CopyInit(
        sizeof(FlightDescriptor), alignof(FlightDescriptor), *arena_bits);
  } else {
    return ::google::protobuf::internal::MessageCreator(&FlightDescriptor::PlacementNew_,
                                 sizeof(FlightDescriptor),
                                 alignof(FlightDescriptor));
  }
}
#endif
constexpr auto FlightDescriptor::InternalGenerateClassData_() {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &_FlightDescriptor_default_instance_._instance,
          &_table_.header,
          nullptr,  // IsInitialized
          &FlightDescriptor::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<FlightDescriptor>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &FlightDescriptor::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<FlightDescriptor>(), &FlightDescriptor::ByteSizeLong,
              &FlightDescriptor::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(FlightDescriptor, _impl_._cached_size_),
          false,
      },
      &FlightDescriptor::kDescriptorMethods,
      &descriptor_table_Flight_2eproto,
      nullptr,  // tracker
  };
}

PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull FlightDescriptor_class_data_ =
        FlightDescriptor::InternalGenerateClassData_();

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
FlightDescriptor::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&FlightDescriptor_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(FlightDescriptor_class_data_.tc_table);
  return FlightDescriptor_class_data_.base();
}
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<2, 3, 0, 51, 2>
FlightDescriptor::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(FlightDescriptor, _impl_._has_bits_),
    0, // no _extensions_
    3, 24,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967288,  // skipmap
    offsetof(decltype(_table_), field_entries),
    3,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    FlightDescriptor_class_data_.base(),
    nullptr,  // post_loop_handler
    ::_pbi::TcParser::GenericFallback,  // fallback
    #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
    ::_pbi::TcParser::GetTable<::arrow::flight::protocol::FlightDescriptor>(),  // to_prefetch
    #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
  }, {{
    {::_pbi::TcParser::MiniParse, {}},
    // .arrow.flight.protocol.FlightDescriptor.DescriptorType type = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(FlightDescriptor, _impl_.type_), 2>(),
     {8, 2, 0,
      PROTOBUF_FIELD_OFFSET(FlightDescriptor, _impl_.type_)}},
    // bytes cmd = 2;
    {::_pbi::TcParser::FastBS1,
     {18, 1, 0,
      PROTOBUF_FIELD_OFFSET(FlightDescriptor, _impl_.cmd_)}},
    // repeated string path = 3;
    {::_pbi::TcParser::FastUR1,
     {26, 0, 0,
      PROTOBUF_FIELD_OFFSET(FlightDescriptor, _impl_.path_)}},
  }}, {{
    65535, 65535
  }}, {{
    // .arrow.flight.protocol.FlightDescriptor.DescriptorType type = 1;
    {PROTOBUF_FIELD_OFFSET(FlightDescriptor, _impl_.type_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kOpenEnum)},
    // bytes cmd = 2;
    {PROTOBUF_FIELD_OFFSET(FlightDescriptor, _impl_.cmd_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    // repeated string path = 3;
    {PROTOBUF_FIELD_OFFSET(FlightDescriptor, _impl_.path_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcRepeated | ::_fl::kUtf8String | ::_fl::kRepSString)},
  }},
  // no aux_entries
  {{
    "\46\0\0\4\0\0\0\0"
    "arrow.flight.protocol.FlightDescriptor"
    "path"
  }},
};
PROTOBUF_NOINLINE void FlightDescriptor::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.FlightDescriptor)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBitForRepeated(cached_has_bits, 0x00000001U)) {
      _impl_.path_.Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.cmd_.ClearNonDefaultToEmpty();
    }
  }
  _impl_.type_ = 0;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL FlightDescriptor::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const FlightDescriptor& this_ = static_cast<const FlightDescriptor&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL FlightDescriptor::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const FlightDescriptor& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.FlightDescriptor)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // .arrow.flight.protocol.FlightDescriptor.DescriptorType type = 1;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    if (this_._internal_type() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteEnumToArray(
          1, this_._internal_type(), target);
    }
  }

  // bytes cmd = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (!this_._internal_cmd().empty()) {
      const ::std::string& _s = this_._internal_cmd();
      target = stream->WriteBytesMaybeAliased(2, _s, target);
    }
  }

  // repeated string path = 3;
  if (CheckHasBitForRepeated(cached_has_bits, 0x00000001U)) {
    for (int i = 0, n = this_._internal_path_size(); i < n; ++i) {
      const auto& s = this_._internal_path().Get(i);
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          s.data(), static_cast<int>(s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.FlightDescriptor.path");
      target = stream->WriteString(3, s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.FlightDescriptor)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t FlightDescriptor::ByteSizeLong(const MessageLite& base) {
  const FlightDescriptor& this_ = static_cast<const FlightDescriptor&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t FlightDescriptor::ByteSizeLong() const {
  const FlightDescriptor& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.FlightDescriptor)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    // repeated string path = 3;
    if (CheckHasBitForRepeated(cached_has_bits, 0x00000001U)) {
      total_size +=
          1 * ::google::protobuf::internal::FromIntSize(this_._internal_path().size());
      for (int i = 0, n = this_._internal_path().size(); i < n; ++i) {
        total_size += ::google::protobuf::internal::WireFormatLite::StringSize(
            this_._internal_path().Get(i));
      }
    }
    // bytes cmd = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!this_._internal_cmd().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                        this_._internal_cmd());
      }
    }
    // .arrow.flight.protocol.FlightDescriptor.DescriptorType type = 1;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (this_._internal_type() != 0) {
        total_size += 1 +
                      ::_pbi::WireFormatLite::EnumSize(this_._internal_type());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void FlightDescriptor::MergeImpl(::google::protobuf::MessageLite& to_msg,
                            const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this =
      static_cast<FlightDescriptor*>(&to_msg);
  auto& from = static_cast<const FlightDescriptor&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.FlightDescriptor)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBitForRepeated(cached_has_bits, 0x00000001U)) {
      _this->_internal_mutable_path()->InternalMergeFromWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), arena,
          from._internal_path());
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!from._internal_cmd().empty()) {
        _this->_internal_set_cmd(from._internal_cmd());
      } else {
        if (_this->_impl_.cmd_.IsDefault()) {
          _this->_internal_set_cmd("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (from._internal_type() != 0) {
        _this->_impl_.type_ = from._impl_.type_;
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void FlightDescriptor::CopyFrom(const FlightDescriptor& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.FlightDescriptor)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void FlightDescriptor::InternalSwap(FlightDescriptor* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.path_.InternalSwap(&other->_impl_.path_);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.cmd_, &other->_impl_.cmd_, arena);
  swap(_impl_.type_, other->_impl_.type_);
}

::google::protobuf::Metadata FlightDescriptor::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

class FlightInfo::_Internal {
 public:
  using HasBits =
      decltype(::std::declval<FlightInfo>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(FlightInfo, _impl_._has_bits_);
};

FlightInfo::FlightInfo(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, FlightInfo_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.FlightInfo)
}
PROTOBUF_NDEBUG_INLINE FlightInfo::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::FlightInfo& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        #ifdef PROTOBUF_INTERNAL_REMOVE_ARENA_PTRS_REPEATED_PTR_FIELD
        endpoint_{visibility, ::_pbi::InternalMetadataOffset::Build<
            ::arrow::flight::protocol::FlightInfo,
            PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::FlightInfo, _impl_.endpoint_)>()
        , from.endpoint_}
        #else
        endpoint_ { visibility, arena, from.endpoint_ }
        #endif
        ,
        schema_(arena, from.schema_),
        app_metadata_(arena, from.app_metadata_) {}

FlightInfo::FlightInfo(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const FlightInfo& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, FlightInfo_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  FlightInfo* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  ::uint32_t cached_has_bits = _impl_._has_bits_[0];
  _impl_.flight_descriptor_ = (CheckHasBit(cached_has_bits, 0x00000008U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.flight_descriptor_)
                : nullptr;
  ::memcpy(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, total_records_),
           reinterpret_cast<const char*>(&from._impl_) +
               offsetof(Impl_, total_records_),
           offsetof(Impl_, ordered_) -
               offsetof(Impl_, total_records_) +
               sizeof(Impl_::ordered_));

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.FlightInfo)
}
PROTOBUF_NDEBUG_INLINE FlightInfo::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        #ifdef PROTOBUF_INTERNAL_REMOVE_ARENA_PTRS_REPEATED_PTR_FIELD
        endpoint_{visibility, ::_pbi::InternalMetadataOffset::Build<
            ::arrow::flight::protocol::FlightInfo,
            PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::FlightInfo, _impl_.endpoint_)>()
        }
        #else
        endpoint_ { visibility, arena }
        #endif
        ,
        schema_(arena),
        app_metadata_(arena) {}

inline void FlightInfo::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, flight_descriptor_),
           0,
           offsetof(Impl_, ordered_) -
               offsetof(Impl_, flight_descriptor_) +
               sizeof(Impl_::ordered_));
}
FlightInfo::~FlightInfo() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.FlightInfo)
  SharedDtor(*this);
}
inline void FlightInfo::SharedDtor(MessageLite& self) {
  FlightInfo& this_ = static_cast<FlightInfo&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.schema_.Destroy();
  this_._impl_.app_metadata_.Destroy();
  delete this_._impl_.flight_descriptor_;
  this_._impl_.~Impl_();
}

inline void* PROTOBUF_NONNULL FlightInfo::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) FlightInfo(arena);
}
#ifdef PROTOBUF_INTERNAL_REMOVE_ARENA_PTRS_REPEATED_PTR_FIELD
constexpr auto FlightInfo::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(FlightInfo),
                                            alignof(FlightInfo));
}
#else  // !PROTOBUF_INTERNAL_REMOVE_ARENA_PTRS_REPEATED_PTR_FIELD
constexpr auto FlightInfo::InternalNewImpl_() {
  constexpr auto arena_bits = ::google::protobuf::internal::EncodePlacementArenaOffsets({
      PROTOBUF_FIELD_OFFSET(FlightInfo, _impl_.endpoint_) +
          decltype(FlightInfo::_impl_.endpoint_)::
              InternalGetArenaOffset(
                  ::google::protobuf::Message::internal_visibility()),
  });
  if (arena_bits.has_value()) {
    return ::google::protobuf::internal::MessageCreator::CopyInit(
        sizeof(FlightInfo), alignof(FlightInfo), *arena_bits);
  } else {
    return ::google::protobuf::internal::MessageCreator(&FlightInfo::PlacementNew_,
                                 sizeof(FlightInfo),
                                 alignof(FlightInfo));
  }
}
#endif
constexpr auto FlightInfo::InternalGenerateClassData_() {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &_FlightInfo_default_instance_._instance,
          &_table_.header,
          nullptr,  // IsInitialized
          &FlightInfo::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<FlightInfo>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &FlightInfo::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<FlightInfo>(), &FlightInfo::ByteSizeLong,
              &FlightInfo::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(FlightInfo, _impl_._cached_size_),
          false,
      },
      &FlightInfo::kDescriptorMethods,
      &descriptor_table_Flight_2eproto,
      nullptr,  // tracker
  };
}

PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull FlightInfo_class_data_ =
        FlightInfo::InternalGenerateClassData_();

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
FlightInfo::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&FlightInfo_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(FlightInfo_class_data_.tc_table);
  return FlightInfo_class_data_.base();
}
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<3, 7, 2, 0, 2>
FlightInfo::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(FlightInfo, _impl_._has_bits_),
    0, // no _extensions_
    7, 56,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967168,  // skipmap
    offsetof(decltype(_table_), field_entries),
    7,  // num_field_entries
    2,  // num_aux_entries
    offsetof(decltype(_table_), aux_entries),
    FlightInfo_class_data_.base(),
    nullptr,  // post_loop_handler
    ::_pbi::TcParser::GenericFallback,  // fallback
    #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
    ::_pbi::TcParser::GetTable<::arrow::flight::protocol::FlightInfo>(),  // to_prefetch
    #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
  }, {{
    {::_pbi::TcParser::MiniParse, {}},
    // bytes schema = 1;
    {::_pbi::TcParser::FastBS1,
     {10, 1, 0,
      PROTOBUF_FIELD_OFFSET(FlightInfo, _impl_.schema_)}},
    // .arrow.flight.protocol.FlightDescriptor flight_descriptor = 2;
    {::_pbi::TcParser::FastMtS1,
     {18, 3, 0,
      PROTOBUF_FIELD_OFFSET(FlightInfo, _impl_.flight_descriptor_)}},
    // repeated .arrow.flight.protocol.FlightEndpoint endpoint = 3;
    {::_pbi::TcParser::FastMtR1,
     {26, 0, 1,
      PROTOBUF_FIELD_OFFSET(FlightInfo, _impl_.endpoint_)}},
    // int64 total_records = 4;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(FlightInfo, _impl_.total_records_), 4>(),
     {32, 4, 0,
      PROTOBUF_FIELD_OFFSET(FlightInfo, _impl_.total_records_)}},
    // int64 total_bytes = 5;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(FlightInfo, _impl_.total_bytes_), 5>(),
     {40, 5, 0,
      PROTOBUF_FIELD_OFFSET(FlightInfo, _impl_.total_bytes_)}},
    // bool ordered = 6;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(FlightInfo, _impl_.ordered_), 6>(),
     {48, 6, 0,
      PROTOBUF_FIELD_OFFSET(FlightInfo, _impl_.ordered_)}},
    // bytes app_metadata = 7;
    {::_pbi::TcParser::FastBS1,
     {58, 2, 0,
      PROTOBUF_FIELD_OFFSET(FlightInfo, _impl_.app_metadata_)}},
  }}, {{
    65535, 65535
  }}, {{
    // bytes schema = 1;
    {PROTOBUF_FIELD_OFFSET(FlightInfo, _impl_.schema_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    // .arrow.flight.protocol.FlightDescriptor flight_descriptor = 2;
    {PROTOBUF_FIELD_OFFSET(FlightInfo, _impl_.flight_descriptor_), _Internal::kHasBitsOffset + 3, 0, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
    // repeated .arrow.flight.protocol.FlightEndpoint endpoint = 3;
    {PROTOBUF_FIELD_OFFSET(FlightInfo, _impl_.endpoint_), _Internal::kHasBitsOffset + 0, 1, (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
    // int64 total_records = 4;
    {PROTOBUF_FIELD_OFFSET(FlightInfo, _impl_.total_records_), _Internal::kHasBitsOffset + 4, 0, (0 | ::_fl::kFcOptional | ::_fl::kInt64)},
    // int64 total_bytes = 5;
    {PROTOBUF_FIELD_OFFSET(FlightInfo, _impl_.total_bytes_), _Internal::kHasBitsOffset + 5, 0, (0 | ::_fl::kFcOptional | ::_fl::kInt64)},
    // bool ordered = 6;
    {PROTOBUF_FIELD_OFFSET(FlightInfo, _impl_.ordered_), _Internal::kHasBitsOffset + 6, 0, (0 | ::_fl::kFcOptional | ::_fl::kBool)},
    // bytes app_metadata = 7;
    {PROTOBUF_FIELD_OFFSET(FlightInfo, _impl_.app_metadata_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
  }},
  {{
      {::_pbi::TcParser::GetTable<::arrow::flight::protocol::FlightDescriptor>()},
      {::_pbi::TcParser::GetTable<::arrow::flight::protocol::FlightEndpoint>()},
  }},
  {{
  }},
};
PROTOBUF_NOINLINE void FlightInfo::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.FlightInfo)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000000fU)) {
    if (CheckHasBitForRepeated(cached_has_bits, 0x00000001U)) {
      _impl_.endpoint_.Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.schema_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _impl_.app_metadata_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      ABSL_DCHECK(_impl_.flight_descriptor_ != nullptr);
      _impl_.flight_descriptor_->Clear();
    }
  }
  if (BatchCheckHasBit(cached_has_bits, 0x00000070U)) {
    ::memset(&_impl_.total_records_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.ordered_) -
        reinterpret_cast<char*>(&_impl_.total_records_)) + sizeof(_impl_.ordered_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL FlightInfo::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const FlightInfo& this_ = static_cast<const FlightInfo&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL FlightInfo::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const FlightInfo& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.FlightInfo)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // bytes schema = 1;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (!this_._internal_schema().empty()) {
      const ::std::string& _s = this_._internal_schema();
      target = stream->WriteBytesMaybeAliased(1, _s, target);
    }
  }

  // .arrow.flight.protocol.FlightDescriptor flight_descriptor = 2;
  if (CheckHasBit(cached_has_bits, 0x00000008U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        2, *this_._impl_.flight_descriptor_, this_._impl_.flight_descriptor_->GetCachedSize(), target,
        stream);
  }

  // repeated .arrow.flight.protocol.FlightEndpoint endpoint = 3;
  if (CheckHasBitForRepeated(cached_has_bits, 0x00000001U)) {
    for (unsigned i = 0, n = static_cast<unsigned>(
                             this_._internal_endpoint_size());
         i < n; i++) {
      const auto& repfield = this_._internal_endpoint().Get(i);
      target =
          ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
              3, repfield, repfield.GetCachedSize(),
              target, stream);
    }
  }

  // int64 total_records = 4;
  if (CheckHasBit(cached_has_bits, 0x00000010U)) {
    if (this_._internal_total_records() != 0) {
      target =
          ::google::protobuf::internal::WireFormatLite::WriteInt64ToArrayWithField<4>(
              stream, this_._internal_total_records(), target);
    }
  }

  // int64 total_bytes = 5;
  if (CheckHasBit(cached_has_bits, 0x00000020U)) {
    if (this_._internal_total_bytes() != 0) {
      target =
          ::google::protobuf::internal::WireFormatLite::WriteInt64ToArrayWithField<5>(
              stream, this_._internal_total_bytes(), target);
    }
  }

  // bool ordered = 6;
  if (CheckHasBit(cached_has_bits, 0x00000040U)) {
    if (this_._internal_ordered() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteBoolToArray(
          6, this_._internal_ordered(), target);
    }
  }

  // bytes app_metadata = 7;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    if (!this_._internal_app_metadata().empty()) {
      const ::std::string& _s = this_._internal_app_metadata();
      target = stream->WriteBytesMaybeAliased(7, _s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.FlightInfo)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t FlightInfo::ByteSizeLong(const MessageLite& base) {
  const FlightInfo& this_ = static_cast<const FlightInfo&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t FlightInfo::ByteSizeLong() const {
  const FlightInfo& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.FlightInfo)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000007fU)) {
    // repeated .arrow.flight.protocol.FlightEndpoint endpoint = 3;
    if (CheckHasBitForRepeated(cached_has_bits, 0x00000001U)) {
      total_size += 1UL * this_._internal_endpoint_size();
      for (const auto& msg : this_._internal_endpoint()) {
        total_size += ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
      }
    }
    // bytes schema = 1;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!this_._internal_schema().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                        this_._internal_schema());
      }
    }
    // bytes app_metadata = 7;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (!this_._internal_app_metadata().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                        this_._internal_app_metadata());
      }
    }
    // .arrow.flight.protocol.FlightDescriptor flight_descriptor = 2;
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.flight_descriptor_);
    }
    // int64 total_records = 4;
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      if (this_._internal_total_records() != 0) {
        total_size += ::_pbi::WireFormatLite::Int64SizePlusOne(
            this_._internal_total_records());
      }
    }
    // int64 total_bytes = 5;
    if (CheckHasBit(cached_has_bits, 0x00000020U)) {
      if (this_._internal_total_bytes() != 0) {
        total_size += ::_pbi::WireFormatLite::Int64SizePlusOne(
            this_._internal_total_bytes());
      }
    }
    // bool ordered = 6;
    if (CheckHasBit(cached_has_bits, 0x00000040U)) {
      if (this_._internal_ordered() != 0) {
        total_size += 2;
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void FlightInfo::MergeImpl(::google::protobuf::MessageLite& to_msg,
                            const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this =
      static_cast<FlightInfo*>(&to_msg);
  auto& from = static_cast<const FlightInfo&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.FlightInfo)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000007fU)) {
    if (CheckHasBitForRepeated(cached_has_bits, 0x00000001U)) {
      _this->_internal_mutable_endpoint()->InternalMergeFromWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), arena,
          from._internal_endpoint());
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!from._internal_schema().empty()) {
        _this->_internal_set_schema(from._internal_schema());
      } else {
        if (_this->_impl_.schema_.IsDefault()) {
          _this->_internal_set_schema("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (!from._internal_app_metadata().empty()) {
        _this->_internal_set_app_metadata(from._internal_app_metadata());
      } else {
        if (_this->_impl_.app_metadata_.IsDefault()) {
          _this->_internal_set_app_metadata("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      ABSL_DCHECK(from._impl_.flight_descriptor_ != nullptr);
      if (_this->_impl_.flight_descriptor_ == nullptr) {
        _this->_impl_.flight_descriptor_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.flight_descriptor_);
      } else {
        _this->_impl_.flight_descriptor_->MergeFrom(*from._impl_.flight_descriptor_);
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      if (from._internal_total_records() != 0) {
        _this->_impl_.total_records_ = from._impl_.total_records_;
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000020U)) {
      if (from._internal_total_bytes() != 0) {
        _this->_impl_.total_bytes_ = from._impl_.total_bytes_;
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000040U)) {
      if (from._internal_ordered() != 0) {
        _this->_impl_.ordered_ = from._impl_.ordered_;
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void FlightInfo::CopyFrom(const FlightInfo& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.FlightInfo)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void FlightInfo::InternalSwap(FlightInfo* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.endpoint_.InternalSwap(&other->_impl_.endpoint_);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.schema_, &other->_impl_.schema_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.app_metadata_, &other->_impl_.app_metadata_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(FlightInfo, _impl_.ordered_)
      + sizeof(FlightInfo::_impl_.ordered_)
      - PROTOBUF_FIELD_OFFSET(FlightInfo, _impl_.flight_descriptor_)>(
          reinterpret_cast<char*>(&_impl_.flight_descriptor_),
          reinterpret_cast<char*>(&other->_impl_.flight_descriptor_));
}

::google::protobuf::Metadata FlightInfo::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

class PollInfo::_Internal {
 public:
  using HasBits =
      decltype(::std::declval<PollInfo>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(PollInfo, _impl_._has_bits_);
};

void PollInfo::clear_expiration_time() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (_impl_.expiration_time_ != nullptr) _impl_.expiration_time_->Clear();
  ClearHasBit(_impl_._has_bits_[0],
                  0x00000004U);
}
PollInfo::PollInfo(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, PollInfo_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.PollInfo)
}
PROTOBUF_NDEBUG_INLINE PollInfo::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::PollInfo& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0} {}

PollInfo::PollInfo(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const PollInfo& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, PollInfo_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  PollInfo* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  ::uint32_t cached_has_bits = _impl_._has_bits_[0];
  _impl_.info_ = (CheckHasBit(cached_has_bits, 0x00000001U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.info_)
                : nullptr;
  _impl_.flight_descriptor_ = (CheckHasBit(cached_has_bits, 0x00000002U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.flight_descriptor_)
                : nullptr;
  _impl_.expiration_time_ = (CheckHasBit(cached_has_bits, 0x00000004U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.expiration_time_)
                : nullptr;
  _impl_.progress_ = from._impl_.progress_;

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.PollInfo)
}
PROTOBUF_NDEBUG_INLINE PollInfo::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0} {}

inline void PollInfo::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, info_),
           0,
           offsetof(Impl_, progress_) -
               offsetof(Impl_, info_) +
               sizeof(Impl_::progress_));
}
PollInfo::~PollInfo() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.PollInfo)
  SharedDtor(*this);
}
inline void PollInfo::SharedDtor(MessageLite& self) {
  PollInfo& this_ = static_cast<PollInfo&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  delete this_._impl_.info_;
  delete this_._impl_.flight_descriptor_;
  delete this_._impl_.expiration_time_;
  this_._impl_.~Impl_();
}

inline void* PROTOBUF_NONNULL PollInfo::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) PollInfo(arena);
}
constexpr auto PollInfo::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::ZeroInit(sizeof(PollInfo),
                                            alignof(PollInfo));
}
constexpr auto PollInfo::InternalGenerateClassData_() {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &_PollInfo_default_instance_._instance,
          &_table_.header,
          nullptr,  // IsInitialized
          &PollInfo::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<PollInfo>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &PollInfo::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<PollInfo>(), &PollInfo::ByteSizeLong,
              &PollInfo::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(PollInfo, _impl_._cached_size_),
          false,
      },
      &PollInfo::kDescriptorMethods,
      &descriptor_table_Flight_2eproto,
      nullptr,  // tracker
  };
}

PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull PollInfo_class_data_ =
        PollInfo::InternalGenerateClassData_();

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
PollInfo::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&PollInfo_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(PollInfo_class_data_.tc_table);
  return PollInfo_class_data_.base();
}
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<2, 4, 3, 0, 2>
PollInfo::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(PollInfo, _impl_._has_bits_),
    0, // no _extensions_
    4, 24,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967280,  // skipmap
    offsetof(decltype(_table_), field_entries),
    4,  // num_field_entries
    3,  // num_aux_entries
    offsetof(decltype(_table_), aux_entries),
    PollInfo_class_data_.base(),
    nullptr,  // post_loop_handler
    ::_pbi::TcParser::GenericFallback,  // fallback
    #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
    ::_pbi::TcParser::GetTable<::arrow::flight::protocol::PollInfo>(),  // to_prefetch
    #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
  }, {{
    // .google.protobuf.Timestamp expiration_time = 4;
    {::_pbi::TcParser::FastMtS1,
     {34, 2, 2,
      PROTOBUF_FIELD_OFFSET(PollInfo, _impl_.expiration_time_)}},
    // .arrow.flight.protocol.FlightInfo info = 1;
    {::_pbi::TcParser::FastMtS1,
     {10, 0, 0,
      PROTOBUF_FIELD_OFFSET(PollInfo, _impl_.info_)}},
    // .arrow.flight.protocol.FlightDescriptor flight_descriptor = 2;
    {::_pbi::TcParser::FastMtS1,
     {18, 1, 1,
      PROTOBUF_FIELD_OFFSET(PollInfo, _impl_.flight_descriptor_)}},
    // optional double progress = 3;
    {::_pbi::TcParser::FastF64S1,
     {25, 3, 0,
      PROTOBUF_FIELD_OFFSET(PollInfo, _impl_.progress_)}},
  }}, {{
    65535, 65535
  }}, {{
    // .arrow.flight.protocol.FlightInfo info = 1;
    {PROTOBUF_FIELD_OFFSET(PollInfo, _impl_.info_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
    // .arrow.flight.protocol.FlightDescriptor flight_descriptor = 2;
    {PROTOBUF_FIELD_OFFSET(PollInfo, _impl_.flight_descriptor_), _Internal::kHasBitsOffset + 1, 1, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
    // optional double progress = 3;
    {PROTOBUF_FIELD_OFFSET(PollInfo, _impl_.progress_), _Internal::kHasBitsOffset + 3, 0, (0 | ::_fl::kFcOptional | ::_fl::kDouble)},
    // .google.protobuf.Timestamp expiration_time = 4;
    {PROTOBUF_FIELD_OFFSET(PollInfo, _impl_.expiration_time_), _Internal::kHasBitsOffset + 2, 2, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
  }},
  {{
      {::_pbi::TcParser::GetTable<::arrow::flight::protocol::FlightInfo>()},
      {::_pbi::TcParser::GetTable<::arrow::flight::protocol::FlightDescriptor>()},
      {::_pbi::TcParser::GetTable<::google::protobuf::Timestamp>()},
  }},
  {{
  }},
};
PROTOBUF_NOINLINE void PollInfo::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.PollInfo)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      ABSL_DCHECK(_impl_.info_ != nullptr);
      _impl_.info_->Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      ABSL_DCHECK(_impl_.flight_descriptor_ != nullptr);
      _impl_.flight_descriptor_->Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      ABSL_DCHECK(_impl_.expiration_time_ != nullptr);
      _impl_.expiration_time_->Clear();
    }
  }
  _impl_.progress_ = 0;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL PollInfo::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const PollInfo& this_ = static_cast<const PollInfo&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL PollInfo::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const PollInfo& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.PollInfo)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // .arrow.flight.protocol.FlightInfo info = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        1, *this_._impl_.info_, this_._impl_.info_->GetCachedSize(), target,
        stream);
  }

  // .arrow.flight.protocol.FlightDescriptor flight_descriptor = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        2, *this_._impl_.flight_descriptor_, this_._impl_.flight_descriptor_->GetCachedSize(), target,
        stream);
  }

  // optional double progress = 3;
  if (CheckHasBit(cached_has_bits, 0x00000008U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteDoubleToArray(
        3, this_._internal_progress(), target);
  }

  // .google.protobuf.Timestamp expiration_time = 4;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        4, *this_._impl_.expiration_time_, this_._impl_.expiration_time_->GetCachedSize(), target,
        stream);
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.PollInfo)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t PollInfo::ByteSizeLong(const MessageLite& base) {
  const PollInfo& this_ = static_cast<const PollInfo&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t PollInfo::ByteSizeLong() const {
  const PollInfo& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.PollInfo)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  total_size += static_cast<bool>(0x00000008U & cached_has_bits) * 9;
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    // .arrow.flight.protocol.FlightInfo info = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.info_);
    }
    // .arrow.flight.protocol.FlightDescriptor flight_descriptor = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.flight_descriptor_);
    }
    // .google.protobuf.Timestamp expiration_time = 4;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.expiration_time_);
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void PollInfo::MergeImpl(::google::protobuf::MessageLite& to_msg,
                            const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this =
      static_cast<PollInfo*>(&to_msg);
  auto& from = static_cast<const PollInfo&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.PollInfo)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000000fU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      ABSL_DCHECK(from._impl_.info_ != nullptr);
      if (_this->_impl_.info_ == nullptr) {
        _this->_impl_.info_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.info_);
      } else {
        _this->_impl_.info_->MergeFrom(*from._impl_.info_);
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      ABSL_DCHECK(from._impl_.flight_descriptor_ != nullptr);
      if (_this->_impl_.flight_descriptor_ == nullptr) {
        _this->_impl_.flight_descriptor_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.flight_descriptor_);
      } else {
        _this->_impl_.flight_descriptor_->MergeFrom(*from._impl_.flight_descriptor_);
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      ABSL_DCHECK(from._impl_.expiration_time_ != nullptr);
      if (_this->_impl_.expiration_time_ == nullptr) {
        _this->_impl_.expiration_time_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.expiration_time_);
      } else {
        _this->_impl_.expiration_time_->MergeFrom(*from._impl_.expiration_time_);
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      _this->_impl_.progress_ = from._impl_.progress_;
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void PollInfo::CopyFrom(const PollInfo& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.PollInfo)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void PollInfo::InternalSwap(PollInfo* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(PollInfo, _impl_.progress_)
      + sizeof(PollInfo::_impl_.progress_)
      - PROTOBUF_FIELD_OFFSET(PollInfo, _impl_.info_)>(
          reinterpret_cast<char*>(&_impl_.info_),
          reinterpret_cast<char*>(&other->_impl_.info_));
}

::google::protobuf::Metadata PollInfo::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

class CancelFlightInfoRequest::_Internal {
 public:
  using HasBits =
      decltype(::std::declval<CancelFlightInfoRequest>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(CancelFlightInfoRequest, _impl_._has_bits_);
};

CancelFlightInfoRequest::CancelFlightInfoRequest(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CancelFlightInfoRequest_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.CancelFlightInfoRequest)
}
PROTOBUF_NDEBUG_INLINE CancelFlightInfoRequest::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::CancelFlightInfoRequest& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0} {}

CancelFlightInfoRequest::CancelFlightInfoRequest(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const CancelFlightInfoRequest& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CancelFlightInfoRequest_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  CancelFlightInfoRequest* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  ::uint32_t cached_has_bits = _impl_._has_bits_[0];
  _impl_.info_ = (CheckHasBit(cached_has_bits, 0x00000001U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.info_)
                : nullptr;

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.CancelFlightInfoRequest)
}
PROTOBUF_NDEBUG_INLINE CancelFlightInfoRequest::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0} {}

inline void CancelFlightInfoRequest::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.info_ = {};
}
CancelFlightInfoRequest::~CancelFlightInfoRequest() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.CancelFlightInfoRequest)
  SharedDtor(*this);
}
inline void CancelFlightInfoRequest::SharedDtor(MessageLite& self) {
  CancelFlightInfoRequest& this_ = static_cast<CancelFlightInfoRequest&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  delete this_._impl_.info_;
  this_._impl_.~Impl_();
}

inline void* PROTOBUF_NONNULL CancelFlightInfoRequest::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) CancelFlightInfoRequest(arena);
}
constexpr auto CancelFlightInfoRequest::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::ZeroInit(sizeof(CancelFlightInfoRequest),
                                            alignof(CancelFlightInfoRequest));
}
constexpr auto CancelFlightInfoRequest::InternalGenerateClassData_() {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &_CancelFlightInfoRequest_default_instance_._instance,
          &_table_.header,
          nullptr,  // IsInitialized
          &CancelFlightInfoRequest::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<CancelFlightInfoRequest>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &CancelFlightInfoRequest::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<CancelFlightInfoRequest>(), &CancelFlightInfoRequest::ByteSizeLong,
              &CancelFlightInfoRequest::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(CancelFlightInfoRequest, _impl_._cached_size_),
          false,
      },
      &CancelFlightInfoRequest::kDescriptorMethods,
      &descriptor_table_Flight_2eproto,
      nullptr,  // tracker
  };
}

PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull CancelFlightInfoRequest_class_data_ =
        CancelFlightInfoRequest::InternalGenerateClassData_();

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CancelFlightInfoRequest::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CancelFlightInfoRequest_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(CancelFlightInfoRequest_class_data_.tc_table);
  return CancelFlightInfoRequest_class_data_.base();
}
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 1, 0, 2>
CancelFlightInfoRequest::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(CancelFlightInfoRequest, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    1,  // num_aux_entries
    offsetof(decltype(_table_), aux_entries),
    CancelFlightInfoRequest_class_data_.base(),
    nullptr,  // post_loop_handler
    ::_pbi::TcParser::GenericFallback,  // fallback
    #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
    ::_pbi::TcParser::GetTable<::arrow::flight::protocol::CancelFlightInfoRequest>(),  // to_prefetch
    #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
  }, {{
    // .arrow.flight.protocol.FlightInfo info = 1;
    {::_pbi::TcParser::FastMtS1,
     {10, 0, 0,
      PROTOBUF_FIELD_OFFSET(CancelFlightInfoRequest, _impl_.info_)}},
  }}, {{
    65535, 65535
  }}, {{
    // .arrow.flight.protocol.FlightInfo info = 1;
    {PROTOBUF_FIELD_OFFSET(CancelFlightInfoRequest, _impl_.info_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
  }},
  {{
      {::_pbi::TcParser::GetTable<::arrow::flight::protocol::FlightInfo>()},
  }},
  {{
  }},
};
PROTOBUF_NOINLINE void CancelFlightInfoRequest::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.CancelFlightInfoRequest)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    ABSL_DCHECK(_impl_.info_ != nullptr);
    _impl_.info_->Clear();
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL CancelFlightInfoRequest::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const CancelFlightInfoRequest& this_ = static_cast<const CancelFlightInfoRequest&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL CancelFlightInfoRequest::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const CancelFlightInfoRequest& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.CancelFlightInfoRequest)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // .arrow.flight.protocol.FlightInfo info = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        1, *this_._impl_.info_, this_._impl_.info_->GetCachedSize(), target,
        stream);
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.CancelFlightInfoRequest)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t CancelFlightInfoRequest::ByteSizeLong(const MessageLite& base) {
  const CancelFlightInfoRequest& this_ = static_cast<const CancelFlightInfoRequest&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t CancelFlightInfoRequest::ByteSizeLong() const {
  const CancelFlightInfoRequest& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.CancelFlightInfoRequest)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

   {
    // .arrow.flight.protocol.FlightInfo info = 1;
    cached_has_bits = this_._impl_._has_bits_[0];
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.info_);
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void CancelFlightInfoRequest::MergeImpl(::google::protobuf::MessageLite& to_msg,
                            const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this =
      static_cast<CancelFlightInfoRequest*>(&to_msg);
  auto& from = static_cast<const CancelFlightInfoRequest&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.CancelFlightInfoRequest)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    ABSL_DCHECK(from._impl_.info_ != nullptr);
    if (_this->_impl_.info_ == nullptr) {
      _this->_impl_.info_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.info_);
    } else {
      _this->_impl_.info_->MergeFrom(*from._impl_.info_);
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void CancelFlightInfoRequest::CopyFrom(const CancelFlightInfoRequest& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.CancelFlightInfoRequest)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void CancelFlightInfoRequest::InternalSwap(CancelFlightInfoRequest* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.info_, other->_impl_.info_);
}

::google::protobuf::Metadata CancelFlightInfoRequest::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

class CancelFlightInfoResult::_Internal {
 public:
  using HasBits =
      decltype(::std::declval<CancelFlightInfoResult>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(CancelFlightInfoResult, _impl_._has_bits_);
};

CancelFlightInfoResult::CancelFlightInfoResult(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CancelFlightInfoResult_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.CancelFlightInfoResult)
}
CancelFlightInfoResult::CancelFlightInfoResult(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const CancelFlightInfoResult& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CancelFlightInfoResult_class_data_.base()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(from._impl_) {
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}
PROTOBUF_NDEBUG_INLINE CancelFlightInfoResult::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0} {}

inline void CancelFlightInfoResult::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.status_ = {};
}
CancelFlightInfoResult::~CancelFlightInfoResult() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.CancelFlightInfoResult)
  SharedDtor(*this);
}
inline void CancelFlightInfoResult::SharedDtor(MessageLite& self) {
  CancelFlightInfoResult& this_ = static_cast<CancelFlightInfoResult&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.~Impl_();
}

inline void* PROTOBUF_NONNULL CancelFlightInfoResult::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) CancelFlightInfoResult(arena);
}
constexpr auto CancelFlightInfoResult::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::ZeroInit(sizeof(CancelFlightInfoResult),
                                            alignof(CancelFlightInfoResult));
}
constexpr auto CancelFlightInfoResult::InternalGenerateClassData_() {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &_CancelFlightInfoResult_default_instance_._instance,
          &_table_.header,
          nullptr,  // IsInitialized
          &CancelFlightInfoResult::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<CancelFlightInfoResult>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &CancelFlightInfoResult::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<CancelFlightInfoResult>(), &CancelFlightInfoResult::ByteSizeLong,
              &CancelFlightInfoResult::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(CancelFlightInfoResult, _impl_._cached_size_),
          false,
      },
      &CancelFlightInfoResult::kDescriptorMethods,
      &descriptor_table_Flight_2eproto,
      nullptr,  // tracker
  };
}

PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull CancelFlightInfoResult_class_data_ =
        CancelFlightInfoResult::InternalGenerateClassData_();

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CancelFlightInfoResult::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CancelFlightInfoResult_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(CancelFlightInfoResult_class_data_.tc_table);
  return CancelFlightInfoResult_class_data_.base();
}
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2>
CancelFlightInfoResult::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(CancelFlightInfoResult, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    CancelFlightInfoResult_class_data_.base(),
    nullptr,  // post_loop_handler
    ::_pbi::TcParser::GenericFallback,  // fallback
    #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
    ::_pbi::TcParser::GetTable<::arrow::flight::protocol::CancelFlightInfoResult>(),  // to_prefetch
    #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
  }, {{
    // .arrow.flight.protocol.CancelStatus status = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(CancelFlightInfoResult, _impl_.status_), 0>(),
     {8, 0, 0,
      PROTOBUF_FIELD_OFFSET(CancelFlightInfoResult, _impl_.status_)}},
  }}, {{
    65535, 65535
  }}, {{
    // .arrow.flight.protocol.CancelStatus status = 1;
    {PROTOBUF_FIELD_OFFSET(CancelFlightInfoResult, _impl_.status_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kOpenEnum)},
  }},
  // no aux_entries
  {{
  }},
};
PROTOBUF_NOINLINE void CancelFlightInfoResult::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.CancelFlightInfoResult)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.status_ = 0;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL CancelFlightInfoResult::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const CancelFlightInfoResult& this_ = static_cast<const CancelFlightInfoResult&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL CancelFlightInfoResult::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const CancelFlightInfoResult& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.CancelFlightInfoResult)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // .arrow.flight.protocol.CancelStatus status = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (this_._internal_status() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteEnumToArray(
          1, this_._internal_status(), target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.CancelFlightInfoResult)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t CancelFlightInfoResult::ByteSizeLong(const MessageLite& base) {
  const CancelFlightInfoResult& this_ = static_cast<const CancelFlightInfoResult&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t CancelFlightInfoResult::ByteSizeLong() const {
  const CancelFlightInfoResult& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.CancelFlightInfoResult)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

   {
    // .arrow.flight.protocol.CancelStatus status = 1;
    cached_has_bits = this_._impl_._has_bits_[0];
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (this_._internal_status() != 0) {
        total_size += 1 +
                      ::_pbi::WireFormatLite::EnumSize(this_._internal_status());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void CancelFlightInfoResult::MergeImpl(::google::protobuf::MessageLite& to_msg,
                            const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this =
      static_cast<CancelFlightInfoResult*>(&to_msg);
  auto& from = static_cast<const CancelFlightInfoResult&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.CancelFlightInfoResult)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (from._internal_status() != 0) {
      _this->_impl_.status_ = from._impl_.status_;
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void CancelFlightInfoResult::CopyFrom(const CancelFlightInfoResult& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.CancelFlightInfoResult)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void CancelFlightInfoResult::InternalSwap(CancelFlightInfoResult* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.status_, other->_impl_.status_);
}

::google::protobuf::Metadata CancelFlightInfoResult::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

class Ticket::_Internal {
 public:
  using HasBits =
      decltype(::std::declval<Ticket>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(Ticket, _impl_._has_bits_);
};

Ticket::Ticket(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Ticket_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.Ticket)
}
PROTOBUF_NDEBUG_INLINE Ticket::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::Ticket& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        ticket_(arena, from.ticket_) {}

Ticket::Ticket(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const Ticket& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Ticket_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  Ticket* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.Ticket)
}
PROTOBUF_NDEBUG_INLINE Ticket::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        ticket_(arena) {}

inline void Ticket::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
Ticket::~Ticket() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.Ticket)
  SharedDtor(*this);
}
inline void Ticket::SharedDtor(MessageLite& self) {
  Ticket& this_ = static_cast<Ticket&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.ticket_.Destroy();
  this_._impl_.~Impl_();
}

inline void* PROTOBUF_NONNULL Ticket::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) Ticket(arena);
}
constexpr auto Ticket::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(Ticket),
                                            alignof(Ticket));
}
constexpr auto Ticket::InternalGenerateClassData_() {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &_Ticket_default_instance_._instance,
          &_table_.header,
          nullptr,  // IsInitialized
          &Ticket::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<Ticket>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &Ticket::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<Ticket>(), &Ticket::ByteSizeLong,
              &Ticket::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(Ticket, _impl_._cached_size_),
          false,
      },
      &Ticket::kDescriptorMethods,
      &descriptor_table_Flight_2eproto,
      nullptr,  // tracker
  };
}

PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull Ticket_class_data_ =
        Ticket::InternalGenerateClassData_();

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Ticket::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Ticket_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(Ticket_class_data_.tc_table);
  return Ticket_class_data_.base();
}
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2>
Ticket::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(Ticket, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    Ticket_class_data_.base(),
    nullptr,  // post_loop_handler
    ::_pbi::TcParser::GenericFallback,  // fallback
    #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
    ::_pbi::TcParser::GetTable<::arrow::flight::protocol::Ticket>(),  // to_prefetch
    #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
  }, {{
    // bytes ticket = 1;
    {::_pbi::TcParser::FastBS1,
     {10, 0, 0,
      PROTOBUF_FIELD_OFFSET(Ticket, _impl_.ticket_)}},
  }}, {{
    65535, 65535
  }}, {{
    // bytes ticket = 1;
    {PROTOBUF_FIELD_OFFSET(Ticket, _impl_.ticket_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
  }},
  // no aux_entries
  {{
  }},
};
PROTOBUF_NOINLINE void Ticket::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.Ticket)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.ticket_.ClearNonDefaultToEmpty();
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL Ticket::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const Ticket& this_ = static_cast<const Ticket&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL Ticket::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const Ticket& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.Ticket)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // bytes ticket = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_ticket().empty()) {
      const ::std::string& _s = this_._internal_ticket();
      target = stream->WriteBytesMaybeAliased(1, _s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.Ticket)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t Ticket::ByteSizeLong(const MessageLite& base) {
  const Ticket& this_ = static_cast<const Ticket&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t Ticket::ByteSizeLong() const {
  const Ticket& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.Ticket)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

   {
    // bytes ticket = 1;
    cached_has_bits = this_._impl_._has_bits_[0];
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_ticket().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                        this_._internal_ticket());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void Ticket::MergeImpl(::google::protobuf::MessageLite& to_msg,
                            const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this =
      static_cast<Ticket*>(&to_msg);
  auto& from = static_cast<const Ticket&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.Ticket)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!from._internal_ticket().empty()) {
      _this->_internal_set_ticket(from._internal_ticket());
    } else {
      if (_this->_impl_.ticket_.IsDefault()) {
        _this->_internal_set_ticket("");
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void Ticket::CopyFrom(const Ticket& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.Ticket)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void Ticket::InternalSwap(Ticket* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.ticket_, &other->_impl_.ticket_, arena);
}

::google::protobuf::Metadata Ticket::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

class Location::_Internal {
 public:
  using HasBits =
      decltype(::std::declval<Location>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(Location, _impl_._has_bits_);
};

Location::Location(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Location_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.Location)
}
PROTOBUF_NDEBUG_INLINE Location::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::Location& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        uri_(arena, from.uri_) {}

Location::Location(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const Location& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Location_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  Location* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.Location)
}
PROTOBUF_NDEBUG_INLINE Location::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        uri_(arena) {}

inline void Location::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
Location::~Location() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.Location)
  SharedDtor(*this);
}
inline void Location::SharedDtor(MessageLite& self) {
  Location& this_ = static_cast<Location&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.uri_.Destroy();
  this_._impl_.~Impl_();
}

inline void* PROTOBUF_NONNULL Location::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) Location(arena);
}
constexpr auto Location::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(Location),
                                            alignof(Location));
}
constexpr auto Location::InternalGenerateClassData_() {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &_Location_default_instance_._instance,
          &_table_.header,
          nullptr,  // IsInitialized
          &Location::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<Location>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &Location::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<Location>(), &Location::ByteSizeLong,
              &Location::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(Location, _impl_._cached_size_),
          false,
      },
      &Location::kDescriptorMethods,
      &descriptor_table_Flight_2eproto,
      nullptr,  // tracker
  };
}

PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull Location_class_data_ =
        Location::InternalGenerateClassData_();

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Location::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Location_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(Location_class_data_.tc_table);
  return Location_class_data_.base();
}
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 42, 2>
Location::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(Location, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    Location_class_data_.base(),
    nullptr,  // post_loop_handler
    ::_pbi::TcParser::GenericFallback,  // fallback
    #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
    ::_pbi::TcParser::GetTable<::arrow::flight::protocol::Location>(),  // to_prefetch
    #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
  }, {{
    // string uri = 1;
    {::_pbi::TcParser::FastUS1,
     {10, 0, 0,
      PROTOBUF_FIELD_OFFSET(Location, _impl_.uri_)}},
  }}, {{
    65535, 65535
  }}, {{
    // string uri = 1;
    {PROTOBUF_FIELD_OFFSET(Location, _impl_.uri_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
  }},
  // no aux_entries
  {{
    "\36\3\0\0\0\0\0\0"
    "arrow.flight.protocol.Location"
    "uri"
  }},
};
PROTOBUF_NOINLINE void Location::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.Location)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.uri_.ClearNonDefaultToEmpty();
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL Location::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const Location& this_ = static_cast<const Location&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL Location::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const Location& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.Location)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // string uri = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_uri().empty()) {
      const ::std::string& _s = this_._internal_uri();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.Location.uri");
      target = stream->WriteStringMaybeAliased(1, _s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.Location)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t Location::ByteSizeLong(const MessageLite& base) {
  const Location& this_ = static_cast<const Location&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t Location::ByteSizeLong() const {
  const Location& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.Location)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

   {
    // string uri = 1;
    cached_has_bits = this_._impl_._has_bits_[0];
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_uri().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_uri());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void Location::MergeImpl(::google::protobuf::MessageLite& to_msg,
                            const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this =
      static_cast<Location*>(&to_msg);
  auto& from = static_cast<const Location&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.Location)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!from._internal_uri().empty()) {
      _this->_internal_set_uri(from._internal_uri());
    } else {
      if (_this->_impl_.uri_.IsDefault()) {
        _this->_internal_set_uri("");
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void Location::CopyFrom(const Location& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.Location)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void Location::InternalSwap(Location* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.uri_, &other->_impl_.uri_, arena);
}

::google::protobuf::Metadata Location::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

class FlightEndpoint::_Internal {
 public:
  using HasBits =
      decltype(::std::declval<FlightEndpoint>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(FlightEndpoint, _impl_._has_bits_);
};

void FlightEndpoint::clear_expiration_time() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (_impl_.expiration_time_ != nullptr) _impl_.expiration_time_->Clear();
  ClearHasBit(_impl_._has_bits_[0],
                  0x00000008U);
}
FlightEndpoint::FlightEndpoint(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, FlightEndpoint_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.FlightEndpoint)
}
PROTOBUF_NDEBUG_INLINE FlightEndpoint::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::FlightEndpoint& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        #ifdef PROTOBUF_INTERNAL_REMOVE_ARENA_PTRS_REPEATED_PTR_FIELD
        location_{visibility, ::_pbi::InternalMetadataOffset::Build<
            ::arrow::flight::protocol::FlightEndpoint,
            PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::FlightEndpoint, _impl_.location_)>()
        , from.location_}
        #else
        location_ { visibility, arena, from.location_ }
        #endif
        ,
        app_metadata_(arena, from.app_metadata_) {}

FlightEndpoint::FlightEndpoint(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const FlightEndpoint& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, FlightEndpoint_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  FlightEndpoint* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  ::uint32_t cached_has_bits = _impl_._has_bits_[0];
  _impl_.ticket_ = (CheckHasBit(cached_has_bits, 0x00000004U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.ticket_)
                : nullptr;
  _impl_.expiration_time_ = (CheckHasBit(cached_has_bits, 0x00000008U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.expiration_time_)
                : nullptr;

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.FlightEndpoint)
}
PROTOBUF_NDEBUG_INLINE FlightEndpoint::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        #ifdef PROTOBUF_INTERNAL_REMOVE_ARENA_PTRS_REPEATED_PTR_FIELD
        location_{visibility, ::_pbi::InternalMetadataOffset::Build<
            ::arrow::flight::protocol::FlightEndpoint,
            PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::FlightEndpoint, _impl_.location_)>()
        }
        #else
        location_ { visibility, arena }
        #endif
        ,
        app_metadata_(arena) {}

inline void FlightEndpoint::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, ticket_),
           0,
           offsetof(Impl_, expiration_time_) -
               offsetof(Impl_, ticket_) +
               sizeof(Impl_::expiration_time_));
}
FlightEndpoint::~FlightEndpoint() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.FlightEndpoint)
  SharedDtor(*this);
}
inline void FlightEndpoint::SharedDtor(MessageLite& self) {
  FlightEndpoint& this_ = static_cast<FlightEndpoint&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.app_metadata_.Destroy();
  delete this_._impl_.ticket_;
  delete this_._impl_.expiration_time_;
  this_._impl_.~Impl_();
}

inline void* PROTOBUF_NONNULL FlightEndpoint::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) FlightEndpoint(arena);
}
#ifdef PROTOBUF_INTERNAL_REMOVE_ARENA_PTRS_REPEATED_PTR_FIELD
constexpr auto FlightEndpoint::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(FlightEndpoint),
                                            alignof(FlightEndpoint));
}
#else  // !PROTOBUF_INTERNAL_REMOVE_ARENA_PTRS_REPEATED_PTR_FIELD
constexpr auto FlightEndpoint::InternalNewImpl_() {
  constexpr auto arena_bits = ::google::protobuf::internal::EncodePlacementArenaOffsets({
      PROTOBUF_FIELD_OFFSET(FlightEndpoint, _impl_.location_) +
          decltype(FlightEndpoint::_impl_.location_)::
              InternalGetArenaOffset(
                  ::google::protobuf::Message::internal_visibility()),
  });
  if (arena_bits.has_value()) {
    return ::google::protobuf::internal::MessageCreator::CopyInit(
        sizeof(FlightEndpoint), alignof(FlightEndpoint), *arena_bits);
  } else {
    return ::google::protobuf::internal::MessageCreator(&FlightEndpoint::PlacementNew_,
                                 sizeof(FlightEndpoint),
                                 alignof(FlightEndpoint));
  }
}
#endif
constexpr auto FlightEndpoint::InternalGenerateClassData_() {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &_FlightEndpoint_default_instance_._instance,
          &_table_.header,
          nullptr,  // IsInitialized
          &FlightEndpoint::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<FlightEndpoint>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &FlightEndpoint::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<FlightEndpoint>(), &FlightEndpoint::ByteSizeLong,
              &FlightEndpoint::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(FlightEndpoint, _impl_._cached_size_),
          false,
      },
      &FlightEndpoint::kDescriptorMethods,
      &descriptor_table_Flight_2eproto,
      nullptr,  // tracker
  };
}

PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull FlightEndpoint_class_data_ =
        FlightEndpoint::InternalGenerateClassData_();

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
FlightEndpoint::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&FlightEndpoint_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(FlightEndpoint_class_data_.tc_table);
  return FlightEndpoint_class_data_.base();
}
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<2, 4, 3, 0, 2>
FlightEndpoint::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(FlightEndpoint, _impl_._has_bits_),
    0, // no _extensions_
    4, 24,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967280,  // skipmap
    offsetof(decltype(_table_), field_entries),
    4,  // num_field_entries
    3,  // num_aux_entries
    offsetof(decltype(_table_), aux_entries),
    FlightEndpoint_class_data_.base(),
    nullptr,  // post_loop_handler
    ::_pbi::TcParser::GenericFallback,  // fallback
    #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
    ::_pbi::TcParser::GetTable<::arrow::flight::protocol::FlightEndpoint>(),  // to_prefetch
    #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
  }, {{
    // bytes app_metadata = 4;
    {::_pbi::TcParser::FastBS1,
     {34, 1, 0,
      PROTOBUF_FIELD_OFFSET(FlightEndpoint, _impl_.app_metadata_)}},
    // .arrow.flight.protocol.Ticket ticket = 1;
    {::_pbi::TcParser::FastMtS1,
     {10, 2, 0,
      PROTOBUF_FIELD_OFFSET(FlightEndpoint, _impl_.ticket_)}},
    // repeated .arrow.flight.protocol.Location location = 2;
    {::_pbi::TcParser::FastMtR1,
     {18, 0, 1,
      PROTOBUF_FIELD_OFFSET(FlightEndpoint, _impl_.location_)}},
    // .google.protobuf.Timestamp expiration_time = 3;
    {::_pbi::TcParser::FastMtS1,
     {26, 3, 2,
      PROTOBUF_FIELD_OFFSET(FlightEndpoint, _impl_.expiration_time_)}},
  }}, {{
    65535, 65535
  }}, {{
    // .arrow.flight.protocol.Ticket ticket = 1;
    {PROTOBUF_FIELD_OFFSET(FlightEndpoint, _impl_.ticket_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
    // repeated .arrow.flight.protocol.Location location = 2;
    {PROTOBUF_FIELD_OFFSET(FlightEndpoint, _impl_.location_), _Internal::kHasBitsOffset + 0, 1, (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
    // .google.protobuf.Timestamp expiration_time = 3;
    {PROTOBUF_FIELD_OFFSET(FlightEndpoint, _impl_.expiration_time_), _Internal::kHasBitsOffset + 3, 2, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
    // bytes app_metadata = 4;
    {PROTOBUF_FIELD_OFFSET(FlightEndpoint, _impl_.app_metadata_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
  }},
  {{
      {::_pbi::TcParser::GetTable<::arrow::flight::protocol::Ticket>()},
      {::_pbi::TcParser::GetTable<::arrow::flight::protocol::Location>()},
      {::_pbi::TcParser::GetTable<::google::protobuf::Timestamp>()},
  }},
  {{
  }},
};
PROTOBUF_NOINLINE void FlightEndpoint::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.FlightEndpoint)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000000fU)) {
    if (CheckHasBitForRepeated(cached_has_bits, 0x00000001U)) {
      _impl_.location_.Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.app_metadata_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      ABSL_DCHECK(_impl_.ticket_ != nullptr);
      _impl_.ticket_->Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      ABSL_DCHECK(_impl_.expiration_time_ != nullptr);
      _impl_.expiration_time_->Clear();
    }
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL FlightEndpoint::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const FlightEndpoint& this_ = static_cast<const FlightEndpoint&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL FlightEndpoint::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const FlightEndpoint& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.FlightEndpoint)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // .arrow.flight.protocol.Ticket ticket = 1;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        1, *this_._impl_.ticket_, this_._impl_.ticket_->GetCachedSize(), target,
        stream);
  }

  // repeated .arrow.flight.protocol.Location location = 2;
  if (CheckHasBitForRepeated(cached_has_bits, 0x00000001U)) {
    for (unsigned i = 0, n = static_cast<unsigned>(
                             this_._internal_location_size());
         i < n; i++) {
      const auto& repfield = this_._internal_location().Get(i);
      target =
          ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
              2, repfield, repfield.GetCachedSize(),
              target, stream);
    }
  }

  // .google.protobuf.Timestamp expiration_time = 3;
  if (CheckHasBit(cached_has_bits, 0x00000008U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        3, *this_._impl_.expiration_time_, this_._impl_.expiration_time_->GetCachedSize(), target,
        stream);
  }

  // bytes app_metadata = 4;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (!this_._internal_app_metadata().empty()) {
      const ::std::string& _s = this_._internal_app_metadata();
      target = stream->WriteBytesMaybeAliased(4, _s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.FlightEndpoint)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t FlightEndpoint::ByteSizeLong(const MessageLite& base) {
  const FlightEndpoint& this_ = static_cast<const FlightEndpoint&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t FlightEndpoint::ByteSizeLong() const {
  const FlightEndpoint& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.FlightEndpoint)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000000fU)) {
    // repeated .arrow.flight.protocol.Location location = 2;
    if (CheckHasBitForRepeated(cached_has_bits, 0x00000001U)) {
      total_size += 1UL * this_._internal_location_size();
      for (const auto& msg : this_._internal_location()) {
        total_size += ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
      }
    }
    // bytes app_metadata = 4;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!this_._internal_app_metadata().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                        this_._internal_app_metadata());
      }
    }
    // .arrow.flight.protocol.Ticket ticket = 1;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.ticket_);
    }
    // .google.protobuf.Timestamp expiration_time = 3;
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.expiration_time_);
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void FlightEndpoint::MergeImpl(::google::protobuf::MessageLite& to_msg,
                            const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this =
      static_cast<FlightEndpoint*>(&to_msg);
  auto& from = static_cast<const FlightEndpoint&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.FlightEndpoint)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000000fU)) {
    if (CheckHasBitForRepeated(cached_has_bits, 0x00000001U)) {
      _this->_internal_mutable_location()->InternalMergeFromWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), arena,
          from._internal_location());
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!from._internal_app_metadata().empty()) {
        _this->_internal_set_app_metadata(from._internal_app_metadata());
      } else {
        if (_this->_impl_.app_metadata_.IsDefault()) {
          _this->_internal_set_app_metadata("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      ABSL_DCHECK(from._impl_.ticket_ != nullptr);
      if (_this->_impl_.ticket_ == nullptr) {
        _this->_impl_.ticket_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.ticket_);
      } else {
        _this->_impl_.ticket_->MergeFrom(*from._impl_.ticket_);
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      ABSL_DCHECK(from._impl_.expiration_time_ != nullptr);
      if (_this->_impl_.expiration_time_ == nullptr) {
        _this->_impl_.expiration_time_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.expiration_time_);
      } else {
        _this->_impl_.expiration_time_->MergeFrom(*from._impl_.expiration_time_);
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void FlightEndpoint::CopyFrom(const FlightEndpoint& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.FlightEndpoint)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void FlightEndpoint::InternalSwap(FlightEndpoint* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.location_.InternalSwap(&other->_impl_.location_);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.app_metadata_, &other->_impl_.app_metadata_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(FlightEndpoint, _impl_.expiration_time_)
      + sizeof(FlightEndpoint::_impl_.expiration_time_)
      - PROTOBUF_FIELD_OFFSET(FlightEndpoint, _impl_.ticket_)>(
          reinterpret_cast<char*>(&_impl_.ticket_),
          reinterpret_cast<char*>(&other->_impl_.ticket_));
}

::google::protobuf::Metadata FlightEndpoint::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

class RenewFlightEndpointRequest::_Internal {
 public:
  using HasBits =
      decltype(::std::declval<RenewFlightEndpointRequest>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(RenewFlightEndpointRequest, _impl_._has_bits_);
};

RenewFlightEndpointRequest::RenewFlightEndpointRequest(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, RenewFlightEndpointRequest_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.RenewFlightEndpointRequest)
}
PROTOBUF_NDEBUG_INLINE RenewFlightEndpointRequest::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::RenewFlightEndpointRequest& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0} {}

RenewFlightEndpointRequest::RenewFlightEndpointRequest(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const RenewFlightEndpointRequest& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, RenewFlightEndpointRequest_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  RenewFlightEndpointRequest* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  ::uint32_t cached_has_bits = _impl_._has_bits_[0];
  _impl_.endpoint_ = (CheckHasBit(cached_has_bits, 0x00000001U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.endpoint_)
                : nullptr;

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.RenewFlightEndpointRequest)
}
PROTOBUF_NDEBUG_INLINE RenewFlightEndpointRequest::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0} {}

inline void RenewFlightEndpointRequest::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.endpoint_ = {};
}
RenewFlightEndpointRequest::~RenewFlightEndpointRequest() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.RenewFlightEndpointRequest)
  SharedDtor(*this);
}
inline void RenewFlightEndpointRequest::SharedDtor(MessageLite& self) {
  RenewFlightEndpointRequest& this_ = static_cast<RenewFlightEndpointRequest&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  delete this_._impl_.endpoint_;
  this_._impl_.~Impl_();
}

inline void* PROTOBUF_NONNULL RenewFlightEndpointRequest::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) RenewFlightEndpointRequest(arena);
}
constexpr auto RenewFlightEndpointRequest::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::ZeroInit(sizeof(RenewFlightEndpointRequest),
                                            alignof(RenewFlightEndpointRequest));
}
constexpr auto RenewFlightEndpointRequest::InternalGenerateClassData_() {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &_RenewFlightEndpointRequest_default_instance_._instance,
          &_table_.header,
          nullptr,  // IsInitialized
          &RenewFlightEndpointRequest::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<RenewFlightEndpointRequest>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &RenewFlightEndpointRequest::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<RenewFlightEndpointRequest>(), &RenewFlightEndpointRequest::ByteSizeLong,
              &RenewFlightEndpointRequest::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(RenewFlightEndpointRequest, _impl_._cached_size_),
          false,
      },
      &RenewFlightEndpointRequest::kDescriptorMethods,
      &descriptor_table_Flight_2eproto,
      nullptr,  // tracker
  };
}

PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull RenewFlightEndpointRequest_class_data_ =
        RenewFlightEndpointRequest::InternalGenerateClassData_();

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
RenewFlightEndpointRequest::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&RenewFlightEndpointRequest_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(RenewFlightEndpointRequest_class_data_.tc_table);
  return RenewFlightEndpointRequest_class_data_.base();
}
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 1, 0, 2>
RenewFlightEndpointRequest::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(RenewFlightEndpointRequest, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    1,  // num_aux_entries
    offsetof(decltype(_table_), aux_entries),
    RenewFlightEndpointRequest_class_data_.base(),
    nullptr,  // post_loop_handler
    ::_pbi::TcParser::GenericFallback,  // fallback
    #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
    ::_pbi::TcParser::GetTable<::arrow::flight::protocol::RenewFlightEndpointRequest>(),  // to_prefetch
    #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
  }, {{
    // .arrow.flight.protocol.FlightEndpoint endpoint = 1;
    {::_pbi::TcParser::FastMtS1,
     {10, 0, 0,
      PROTOBUF_FIELD_OFFSET(RenewFlightEndpointRequest, _impl_.endpoint_)}},
  }}, {{
    65535, 65535
  }}, {{
    // .arrow.flight.protocol.FlightEndpoint endpoint = 1;
    {PROTOBUF_FIELD_OFFSET(RenewFlightEndpointRequest, _impl_.endpoint_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
  }},
  {{
      {::_pbi::TcParser::GetTable<::arrow::flight::protocol::FlightEndpoint>()},
  }},
  {{
  }},
};
PROTOBUF_NOINLINE void RenewFlightEndpointRequest::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.RenewFlightEndpointRequest)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    ABSL_DCHECK(_impl_.endpoint_ != nullptr);
    _impl_.endpoint_->Clear();
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL RenewFlightEndpointRequest::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const RenewFlightEndpointRequest& this_ = static_cast<const RenewFlightEndpointRequest&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL RenewFlightEndpointRequest::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const RenewFlightEndpointRequest& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.RenewFlightEndpointRequest)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // .arrow.flight.protocol.FlightEndpoint endpoint = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        1, *this_._impl_.endpoint_, this_._impl_.endpoint_->GetCachedSize(), target,
        stream);
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.RenewFlightEndpointRequest)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t RenewFlightEndpointRequest::ByteSizeLong(const MessageLite& base) {
  const RenewFlightEndpointRequest& this_ = static_cast<const RenewFlightEndpointRequest&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t RenewFlightEndpointRequest::ByteSizeLong() const {
  const RenewFlightEndpointRequest& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.RenewFlightEndpointRequest)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

   {
    // .arrow.flight.protocol.FlightEndpoint endpoint = 1;
    cached_has_bits = this_._impl_._has_bits_[0];
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.endpoint_);
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void RenewFlightEndpointRequest::MergeImpl(::google::protobuf::MessageLite& to_msg,
                            const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this =
      static_cast<RenewFlightEndpointRequest*>(&to_msg);
  auto& from = static_cast<const RenewFlightEndpointRequest&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.RenewFlightEndpointRequest)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    ABSL_DCHECK(from._impl_.endpoint_ != nullptr);
    if (_this->_impl_.endpoint_ == nullptr) {
      _this->_impl_.endpoint_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.endpoint_);
    } else {
      _this->_impl_.endpoint_->MergeFrom(*from._impl_.endpoint_);
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void RenewFlightEndpointRequest::CopyFrom(const RenewFlightEndpointRequest& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.RenewFlightEndpointRequest)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void RenewFlightEndpointRequest::InternalSwap(RenewFlightEndpointRequest* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.endpoint_, other->_impl_.endpoint_);
}

::google::protobuf::Metadata RenewFlightEndpointRequest::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

class FlightData::_Internal {
 public:
  using HasBits =
      decltype(::std::declval<FlightData>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(FlightData, _impl_._has_bits_);
};

FlightData::FlightData(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, FlightData_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.FlightData)
}
PROTOBUF_NDEBUG_INLINE FlightData::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::FlightData& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        data_header_(arena, from.data_header_),
        app_metadata_(arena, from.app_metadata_),
        data_body_(arena, from.data_body_) {}

FlightData::FlightData(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const FlightData& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, FlightData_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  FlightData* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  ::uint32_t cached_has_bits = _impl_._has_bits_[0];
  _impl_.flight_descriptor_ = (CheckHasBit(cached_has_bits, 0x00000008U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.flight_descriptor_)
                : nullptr;

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.FlightData)
}
PROTOBUF_NDEBUG_INLINE FlightData::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        data_header_(arena),
        app_metadata_(arena),
        data_body_(arena) {}

inline void FlightData::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.flight_descriptor_ = {};
}
FlightData::~FlightData() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.FlightData)
  SharedDtor(*this);
}
inline void FlightData::SharedDtor(MessageLite& self) {
  FlightData& this_ = static_cast<FlightData&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.data_header_.Destroy();
  this_._impl_.app_metadata_.Destroy();
  this_._impl_.data_body_.Destroy();
  delete this_._impl_.flight_descriptor_;
  this_._impl_.~Impl_();
}

inline void* PROTOBUF_NONNULL FlightData::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) FlightData(arena);
}
constexpr auto FlightData::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(FlightData),
                                            alignof(FlightData));
}
constexpr auto FlightData::InternalGenerateClassData_() {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &_FlightData_default_instance_._instance,
          &_table_.header,
          nullptr,  // IsInitialized
          &FlightData::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<FlightData>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &FlightData::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<FlightData>(), &FlightData::ByteSizeLong,
              &FlightData::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(FlightData, _impl_._cached_size_),
          false,
      },
      &FlightData::kDescriptorMethods,
      &descriptor_table_Flight_2eproto,
      nullptr,  // tracker
  };
}

PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull FlightData_class_data_ =
        FlightData::InternalGenerateClassData_();

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
FlightData::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&FlightData_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(FlightData_class_data_.tc_table);
  return FlightData_class_data_.base();
}
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<2, 4, 1, 0, 7>
FlightData::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(FlightData, _impl_._has_bits_),
    0, // no _extensions_
    1000, 24,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967288,  // skipmap
    offsetof(decltype(_table_), field_entries),
    4,  // num_field_entries
    1,  // num_aux_entries
    offsetof(decltype(_table_), aux_entries),
    FlightData_class_data_.base(),
    nullptr,  // post_loop_handler
    ::_pbi::TcParser::GenericFallback,  // fallback
    #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
    ::_pbi::TcParser::GetTable<::arrow::flight::protocol::FlightData>(),  // to_prefetch
    #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
  }, {{
    // bytes data_body = 1000;
    {::_pbi::TcParser::FastBS2,
     {16066, 2, 0,
      PROTOBUF_FIELD_OFFSET(FlightData, _impl_.data_body_)}},
    // .arrow.flight.protocol.FlightDescriptor flight_descriptor = 1;
    {::_pbi::TcParser::FastMtS1,
     {10, 3, 0,
      PROTOBUF_FIELD_OFFSET(FlightData, _impl_.flight_descriptor_)}},
    // bytes data_header = 2;
    {::_pbi::TcParser::FastBS1,
     {18, 0, 0,
      PROTOBUF_FIELD_OFFSET(FlightData, _impl_.data_header_)}},
    // bytes app_metadata = 3;
    {::_pbi::TcParser::FastBS1,
     {26, 1, 0,
      PROTOBUF_FIELD_OFFSET(FlightData, _impl_.app_metadata_)}},
  }}, {{
    1000, 0, 1,
    65534, 3,
    65535, 65535
  }}, {{
    // .arrow.flight.protocol.FlightDescriptor flight_descriptor = 1;
    {PROTOBUF_FIELD_OFFSET(FlightData, _impl_.flight_descriptor_), _Internal::kHasBitsOffset + 3, 0, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
    // bytes data_header = 2;
    {PROTOBUF_FIELD_OFFSET(FlightData, _impl_.data_header_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    // bytes app_metadata = 3;
    {PROTOBUF_FIELD_OFFSET(FlightData, _impl_.app_metadata_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    // bytes data_body = 1000;
    {PROTOBUF_FIELD_OFFSET(FlightData, _impl_.data_body_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
  }},
  {{
      {::_pbi::TcParser::GetTable<::arrow::flight::protocol::FlightDescriptor>()},
  }},
  {{
  }},
};
PROTOBUF_NOINLINE void FlightData::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.FlightData)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000000fU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.data_header_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.app_metadata_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _impl_.data_body_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      ABSL_DCHECK(_impl_.flight_descriptor_ != nullptr);
      _impl_.flight_descriptor_->Clear();
    }
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL FlightData::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const FlightData& this_ = static_cast<const FlightData&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL FlightData::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const FlightData& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.FlightData)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // .arrow.flight.protocol.FlightDescriptor flight_descriptor = 1;
  if (CheckHasBit(cached_has_bits, 0x00000008U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        1, *this_._impl_.flight_descriptor_, this_._impl_.flight_descriptor_->GetCachedSize(), target,
        stream);
  }

  // bytes data_header = 2;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_data_header().empty()) {
      const ::std::string& _s = this_._internal_data_header();
      target = stream->WriteBytesMaybeAliased(2, _s, target);
    }
  }

  // bytes app_metadata = 3;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (!this_._internal_app_metadata().empty()) {
      const ::std::string& _s = this_._internal_app_metadata();
      target = stream->WriteBytesMaybeAliased(3, _s, target);
    }
  }

  // bytes data_body = 1000;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    if (!this_._internal_data_body().empty()) {
      const ::std::string& _s = this_._internal_data_body();
      target = stream->WriteBytesMaybeAliased(1000, _s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.FlightData)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t FlightData::ByteSizeLong(const MessageLite& base) {
  const FlightData& this_ = static_cast<const FlightData&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t FlightData::ByteSizeLong() const {
  const FlightData& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.FlightData)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000000fU)) {
    // bytes data_header = 2;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_data_header().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                        this_._internal_data_header());
      }
    }
    // bytes app_metadata = 3;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!this_._internal_app_metadata().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                        this_._internal_app_metadata());
      }
    }
    // bytes data_body = 1000;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (!this_._internal_data_body().empty()) {
        total_size += 2 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                        this_._internal_data_body());
      }
    }
    // .arrow.flight.protocol.FlightDescriptor flight_descriptor = 1;
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.flight_descriptor_);
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void FlightData::MergeImpl(::google::protobuf::MessageLite& to_msg,
                            const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this =
      static_cast<FlightData*>(&to_msg);
  auto& from = static_cast<const FlightData&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.FlightData)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000000fU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!from._internal_data_header().empty()) {
        _this->_internal_set_data_header(from._internal_data_header());
      } else {
        if (_this->_impl_.data_header_.IsDefault()) {
          _this->_internal_set_data_header("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!from._internal_app_metadata().empty()) {
        _this->_internal_set_app_metadata(from._internal_app_metadata());
      } else {
        if (_this->_impl_.app_metadata_.IsDefault()) {
          _this->_internal_set_app_metadata("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (!from._internal_data_body().empty()) {
        _this->_internal_set_data_body(from._internal_data_body());
      } else {
        if (_this->_impl_.data_body_.IsDefault()) {
          _this->_internal_set_data_body("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      ABSL_DCHECK(from._impl_.flight_descriptor_ != nullptr);
      if (_this->_impl_.flight_descriptor_ == nullptr) {
        _this->_impl_.flight_descriptor_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.flight_descriptor_);
      } else {
        _this->_impl_.flight_descriptor_->MergeFrom(*from._impl_.flight_descriptor_);
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void FlightData::CopyFrom(const FlightData& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.FlightData)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void FlightData::InternalSwap(FlightData* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.data_header_, &other->_impl_.data_header_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.app_metadata_, &other->_impl_.app_metadata_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.data_body_, &other->_impl_.data_body_, arena);
  swap(_impl_.flight_descriptor_, other->_impl_.flight_descriptor_);
}

::google::protobuf::Metadata FlightData::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

class PutResult::_Internal {
 public:
  using HasBits =
      decltype(::std::declval<PutResult>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(PutResult, _impl_._has_bits_);
};

PutResult::PutResult(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, PutResult_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.PutResult)
}
PROTOBUF_NDEBUG_INLINE PutResult::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::PutResult& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        app_metadata_(arena, from.app_metadata_) {}

PutResult::PutResult(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const PutResult& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, PutResult_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  PutResult* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.PutResult)
}
PROTOBUF_NDEBUG_INLINE PutResult::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        app_metadata_(arena) {}

inline void PutResult::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
PutResult::~PutResult() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.PutResult)
  SharedDtor(*this);
}
inline void PutResult::SharedDtor(MessageLite& self) {
  PutResult& this_ = static_cast<PutResult&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.app_metadata_.Destroy();
  this_._impl_.~Impl_();
}

inline void* PROTOBUF_NONNULL PutResult::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) PutResult(arena);
}
constexpr auto PutResult::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(PutResult),
                                            alignof(PutResult));
}
constexpr auto PutResult::InternalGenerateClassData_() {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &_PutResult_default_instance_._instance,
          &_table_.header,
          nullptr,  // IsInitialized
          &PutResult::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<PutResult>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &PutResult::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<PutResult>(), &PutResult::ByteSizeLong,
              &PutResult::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(PutResult, _impl_._cached_size_),
          false,
      },
      &PutResult::kDescriptorMethods,
      &descriptor_table_Flight_2eproto,
      nullptr,  // tracker
  };
}

PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull PutResult_class_data_ =
        PutResult::InternalGenerateClassData_();

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
PutResult::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&PutResult_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(PutResult_class_data_.tc_table);
  return PutResult_class_data_.base();
}
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2>
PutResult::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(PutResult, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    PutResult_class_data_.base(),
    nullptr,  // post_loop_handler
    ::_pbi::TcParser::GenericFallback,  // fallback
    #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
    ::_pbi::TcParser::GetTable<::arrow::flight::protocol::PutResult>(),  // to_prefetch
    #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
  }, {{
    // bytes app_metadata = 1;
    {::_pbi::TcParser::FastBS1,
     {10, 0, 0,
      PROTOBUF_FIELD_OFFSET(PutResult, _impl_.app_metadata_)}},
  }}, {{
    65535, 65535
  }}, {{
    // bytes app_metadata = 1;
    {PROTOBUF_FIELD_OFFSET(PutResult, _impl_.app_metadata_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
  }},
  // no aux_entries
  {{
  }},
};
PROTOBUF_NOINLINE void PutResult::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.PutResult)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.app_metadata_.ClearNonDefaultToEmpty();
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL PutResult::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const PutResult& this_ = static_cast<const PutResult&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL PutResult::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const PutResult& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.PutResult)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // bytes app_metadata = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_app_metadata().empty()) {
      const ::std::string& _s = this_._internal_app_metadata();
      target = stream->WriteBytesMaybeAliased(1, _s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.PutResult)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t PutResult::ByteSizeLong(const MessageLite& base) {
  const PutResult& this_ = static_cast<const PutResult&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t PutResult::ByteSizeLong() const {
  const PutResult& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.PutResult)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

   {
    // bytes app_metadata = 1;
    cached_has_bits = this_._impl_._has_bits_[0];
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_app_metadata().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                        this_._internal_app_metadata());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void PutResult::MergeImpl(::google::protobuf::MessageLite& to_msg,
                            const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this =
      static_cast<PutResult*>(&to_msg);
  auto& from = static_cast<const PutResult&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.PutResult)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!from._internal_app_metadata().empty()) {
      _this->_internal_set_app_metadata(from._internal_app_metadata());
    } else {
      if (_this->_impl_.app_metadata_.IsDefault()) {
        _this->_internal_set_app_metadata("");
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void PutResult::CopyFrom(const PutResult& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.PutResult)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void PutResult::InternalSwap(PutResult* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.app_metadata_, &other->_impl_.app_metadata_, arena);
}

::google::protobuf::Metadata PutResult::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

class SessionOptionValue_StringListValue::_Internal {
 public:
  using HasBits =
      decltype(::std::declval<SessionOptionValue_StringListValue>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(SessionOptionValue_StringListValue, _impl_._has_bits_);
};

SessionOptionValue_StringListValue::SessionOptionValue_StringListValue(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, SessionOptionValue_StringListValue_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.SessionOptionValue.StringListValue)
}
PROTOBUF_NDEBUG_INLINE SessionOptionValue_StringListValue::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::SessionOptionValue_StringListValue& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        #ifdef PROTOBUF_INTERNAL_REMOVE_ARENA_PTRS_REPEATED_PTR_FIELD
        values_{visibility, ::_pbi::InternalMetadataOffset::Build<
            ::arrow::flight::protocol::SessionOptionValue_StringListValue,
            PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::SessionOptionValue_StringListValue, _impl_.values_)>()
        , from.values_}
        #else
        values_ { visibility, arena, from.values_ }
        #endif
     {}

SessionOptionValue_StringListValue::SessionOptionValue_StringListValue(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const SessionOptionValue_StringListValue& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, SessionOptionValue_StringListValue_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SessionOptionValue_StringListValue* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.SessionOptionValue.StringListValue)
}
PROTOBUF_NDEBUG_INLINE SessionOptionValue_StringListValue::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        #ifdef PROTOBUF_INTERNAL_REMOVE_ARENA_PTRS_REPEATED_PTR_FIELD
        values_{visibility, ::_pbi::InternalMetadataOffset::Build<
            ::arrow::flight::protocol::SessionOptionValue_StringListValue,
            PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::SessionOptionValue_StringListValue, _impl_.values_)>()
        }
        #else
        values_ { visibility, arena }
        #endif
     {}

inline void SessionOptionValue_StringListValue::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
SessionOptionValue_StringListValue::~SessionOptionValue_StringListValue() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.SessionOptionValue.StringListValue)
  SharedDtor(*this);
}
inline void SessionOptionValue_StringListValue::SharedDtor(MessageLite& self) {
  SessionOptionValue_StringListValue& this_ = static_cast<SessionOptionValue_StringListValue&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.~Impl_();
}

inline void* PROTOBUF_NONNULL SessionOptionValue_StringListValue::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) SessionOptionValue_StringListValue(arena);
}
#ifdef PROTOBUF_INTERNAL_REMOVE_ARENA_PTRS_REPEATED_PTR_FIELD
constexpr auto SessionOptionValue_StringListValue::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(SessionOptionValue_StringListValue),
                                            alignof(SessionOptionValue_StringListValue));
}
#else  // !PROTOBUF_INTERNAL_REMOVE_ARENA_PTRS_REPEATED_PTR_FIELD
constexpr auto SessionOptionValue_StringListValue::InternalNewImpl_() {
  constexpr auto arena_bits = ::google::protobuf::internal::EncodePlacementArenaOffsets({
      PROTOBUF_FIELD_OFFSET(SessionOptionValue_StringListValue, _impl_.values_) +
          decltype(SessionOptionValue_StringListValue::_impl_.values_)::
              InternalGetArenaOffset(
                  ::google::protobuf::Message::internal_visibility()),
  });
  if (arena_bits.has_value()) {
    return ::google::protobuf::internal::MessageCreator::ZeroInit(
        sizeof(SessionOptionValue_StringListValue), alignof(SessionOptionValue_StringListValue), *arena_bits);
  } else {
    return ::google::protobuf::internal::MessageCreator(&SessionOptionValue_StringListValue::PlacementNew_,
                                 sizeof(SessionOptionValue_StringListValue),
                                 alignof(SessionOptionValue_StringListValue));
  }
}
#endif
constexpr auto SessionOptionValue_StringListValue::InternalGenerateClassData_() {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &_SessionOptionValue_StringListValue_default_instance_._instance,
          &_table_.header,
          nullptr,  // IsInitialized
          &SessionOptionValue_StringListValue::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<SessionOptionValue_StringListValue>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &SessionOptionValue_StringListValue::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<SessionOptionValue_StringListValue>(), &SessionOptionValue_StringListValue::ByteSizeLong,
              &SessionOptionValue_StringListValue::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(SessionOptionValue_StringListValue, _impl_._cached_size_),
          false,
      },
      &SessionOptionValue_StringListValue::kDescriptorMethods,
      &descriptor_table_Flight_2eproto,
      nullptr,  // tracker
  };
}

PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull SessionOptionValue_StringListValue_class_data_ =
        SessionOptionValue_StringListValue::InternalGenerateClassData_();

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
SessionOptionValue_StringListValue::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&SessionOptionValue_StringListValue_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(SessionOptionValue_StringListValue_class_data_.tc_table);
  return SessionOptionValue_StringListValue_class_data_.base();
}
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 71, 2>
SessionOptionValue_StringListValue::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(SessionOptionValue_StringListValue, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    SessionOptionValue_StringListValue_class_data_.base(),
    nullptr,  // post_loop_handler
    ::_pbi::TcParser::GenericFallback,  // fallback
    #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
    ::_pbi::TcParser::GetTable<::arrow::flight::protocol::SessionOptionValue_StringListValue>(),  // to_prefetch
    #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
  }, {{
    // repeated string values = 1;
    {::_pbi::TcParser::FastUR1,
     {10, 0, 0,
      PROTOBUF_FIELD_OFFSET(SessionOptionValue_StringListValue, _impl_.values_)}},
  }}, {{
    65535, 65535
  }}, {{
    // repeated string values = 1;
    {PROTOBUF_FIELD_OFFSET(SessionOptionValue_StringListValue, _impl_.values_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcRepeated | ::_fl::kUtf8String | ::_fl::kRepSString)},
  }},
  // no aux_entries
  {{
    "\70\6\0\0\0\0\0\0"
    "arrow.flight.protocol.SessionOptionValue.StringListValue"
    "values"
  }},
};
PROTOBUF_NOINLINE void SessionOptionValue_StringListValue::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.SessionOptionValue.StringListValue)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBitForRepeated(cached_has_bits, 0x00000001U)) {
    _impl_.values_.Clear();
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL SessionOptionValue_StringListValue::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const SessionOptionValue_StringListValue& this_ = static_cast<const SessionOptionValue_StringListValue&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL SessionOptionValue_StringListValue::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const SessionOptionValue_StringListValue& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.SessionOptionValue.StringListValue)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // repeated string values = 1;
  if (CheckHasBitForRepeated(cached_has_bits, 0x00000001U)) {
    for (int i = 0, n = this_._internal_values_size(); i < n; ++i) {
      const auto& s = this_._internal_values().Get(i);
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          s.data(), static_cast<int>(s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.SessionOptionValue.StringListValue.values");
      target = stream->WriteString(1, s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.SessionOptionValue.StringListValue)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t SessionOptionValue_StringListValue::ByteSizeLong(const MessageLite& base) {
  const SessionOptionValue_StringListValue& this_ = static_cast<const SessionOptionValue_StringListValue&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t SessionOptionValue_StringListValue::ByteSizeLong() const {
  const SessionOptionValue_StringListValue& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.SessionOptionValue.StringListValue)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
   {
    // repeated string values = 1;
    cached_has_bits = this_._impl_._has_bits_[0];
    if (CheckHasBitForRepeated(cached_has_bits, 0x00000001U)) {
      total_size +=
          1 * ::google::protobuf::internal::FromIntSize(this_._internal_values().size());
      for (int i = 0, n = this_._internal_values().size(); i < n; ++i) {
        total_size += ::google::protobuf::internal::WireFormatLite::StringSize(
            this_._internal_values().Get(i));
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void SessionOptionValue_StringListValue::MergeImpl(::google::protobuf::MessageLite& to_msg,
                            const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this =
      static_cast<SessionOptionValue_StringListValue*>(&to_msg);
  auto& from = static_cast<const SessionOptionValue_StringListValue&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.SessionOptionValue.StringListValue)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (CheckHasBitForRepeated(cached_has_bits, 0x00000001U)) {
    _this->_internal_mutable_values()->InternalMergeFromWithArena(
        ::google::protobuf::MessageLite::internal_visibility(), arena,
        from._internal_values());
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void SessionOptionValue_StringListValue::CopyFrom(const SessionOptionValue_StringListValue& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.SessionOptionValue.StringListValue)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void SessionOptionValue_StringListValue::InternalSwap(SessionOptionValue_StringListValue* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.values_.InternalSwap(&other->_impl_.values_);
}

::google::protobuf::Metadata SessionOptionValue_StringListValue::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

class SessionOptionValue::_Internal {
 public:
  static constexpr ::int32_t kOneofCaseOffset =
      PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::SessionOptionValue, _impl_._oneof_case_);
};

void SessionOptionValue::set_allocated_string_list_value(::arrow::flight::protocol::SessionOptionValue_StringListValue* PROTOBUF_NULLABLE string_list_value) {
  ::google::protobuf::Arena* message_arena = GetArena();
  clear_option_value();
  if (string_list_value) {
    ::google::protobuf::Arena* submessage_arena = string_list_value->GetArena();
    if (message_arena != submessage_arena) {
      string_list_value = ::google::protobuf::internal::GetOwnedMessage(message_arena, string_list_value, submessage_arena);
    }
    set_has_string_list_value();
    _impl_.option_value_.string_list_value_ = string_list_value;
  }
  // @@protoc_insertion_point(field_set_allocated:arrow.flight.protocol.SessionOptionValue.string_list_value)
}
SessionOptionValue::SessionOptionValue(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, SessionOptionValue_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.SessionOptionValue)
}
PROTOBUF_NDEBUG_INLINE SessionOptionValue::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::SessionOptionValue& from_msg)
      : option_value_{},
        _cached_size_{0},
        _oneof_case_{from._oneof_case_[0]} {}

SessionOptionValue::SessionOptionValue(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const SessionOptionValue& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, SessionOptionValue_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SessionOptionValue* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  switch (option_value_case()) {
    case OPTION_VALUE_NOT_SET:
      break;
      case kStringValue:
        new (&_impl_.option_value_.string_value_) decltype(_impl_.option_value_.string_value_){arena, from._impl_.option_value_.string_value_};
        break;
      case kBoolValue:
        _impl_.option_value_.bool_value_ = from._impl_.option_value_.bool_value_;
        break;
      case kInt64Value:
        _impl_.option_value_.int64_value_ = from._impl_.option_value_.int64_value_;
        break;
      case kDoubleValue:
        _impl_.option_value_.double_value_ = from._impl_.option_value_.double_value_;
        break;
      case kStringListValue:
        _impl_.option_value_.string_list_value_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.option_value_.string_list_value_);
        break;
  }

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.SessionOptionValue)
}
PROTOBUF_NDEBUG_INLINE SessionOptionValue::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : option_value_{},
        _cached_size_{0},
        _oneof_case_{} {}

inline void SessionOptionValue::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
SessionOptionValue::~SessionOptionValue() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.SessionOptionValue)
  SharedDtor(*this);
}
inline void SessionOptionValue::SharedDtor(MessageLite& self) {
  SessionOptionValue& this_ = static_cast<SessionOptionValue&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  if (this_.has_option_value()) {
    this_.clear_option_value();
  }
  this_._impl_.~Impl_();
}

void SessionOptionValue::clear_option_value() {
// @@protoc_insertion_point(one_of_clear_start:arrow.flight.protocol.SessionOptionValue)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  switch (option_value_case()) {
    case kStringValue: {
      _impl_.option_value_.string_value_.Destroy();
      break;
    }
    case kBoolValue: {
      // No need to clear
      break;
    }
    case kInt64Value: {
      // No need to clear
      break;
    }
    case kDoubleValue: {
      // No need to clear
      break;
    }
    case kStringListValue: {
      if (GetArena() == nullptr) {
        delete _impl_.option_value_.string_list_value_;
      } else if (::google::protobuf::internal::DebugHardenClearOneofMessageOnArena()) {
        ::google::protobuf::internal::MaybePoisonAfterClear(_impl_.option_value_.string_list_value_);
      }
      break;
    }
    case OPTION_VALUE_NOT_SET: {
      break;
    }
  }
  _impl_._oneof_case_[0] = OPTION_VALUE_NOT_SET;
}


inline void* PROTOBUF_NONNULL SessionOptionValue::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) SessionOptionValue(arena);
}
constexpr auto SessionOptionValue::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::ZeroInit(sizeof(SessionOptionValue),
                                            alignof(SessionOptionValue));
}
constexpr auto SessionOptionValue::InternalGenerateClassData_() {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &_SessionOptionValue_default_instance_._instance,
          &_table_.header,
          nullptr,  // IsInitialized
          &SessionOptionValue::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<SessionOptionValue>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &SessionOptionValue::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<SessionOptionValue>(), &SessionOptionValue::ByteSizeLong,
              &SessionOptionValue::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(SessionOptionValue, _impl_._cached_size_),
          false,
      },
      &SessionOptionValue::kDescriptorMethods,
      &descriptor_table_Flight_2eproto,
      nullptr,  // tracker
  };
}

PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull SessionOptionValue_class_data_ =
        SessionOptionValue::InternalGenerateClassData_();

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
SessionOptionValue::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&SessionOptionValue_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(SessionOptionValue_class_data_.tc_table);
  return SessionOptionValue_class_data_.base();
}
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 5, 1, 61, 2>
SessionOptionValue::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(SessionOptionValue,
                          _impl_._cached_size_),  // no hasbits
    0, // no _extensions_
    5, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967264,  // skipmap
    offsetof(decltype(_table_), field_entries),
    5,  // num_field_entries
    1,  // num_aux_entries
    offsetof(decltype(_table_), aux_entries),
    SessionOptionValue_class_data_.base(),
    nullptr,  // post_loop_handler
    ::_pbi::TcParser::GenericFallback,  // fallback
    #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
    ::_pbi::TcParser::GetTable<::arrow::flight::protocol::SessionOptionValue>(),  // to_prefetch
    #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
  }, {{
    {::_pbi::TcParser::MiniParse, {}},
  }}, {{
    65535, 65535
  }}, {{
    // string string_value = 1;
    {PROTOBUF_FIELD_OFFSET(SessionOptionValue, _impl_.option_value_.string_value_), _Internal::kOneofCaseOffset + 0, 0, (0 | ::_fl::kFcOneof | ::_fl::kUtf8String | ::_fl::kRepAString)},
    // bool bool_value = 2;
    {PROTOBUF_FIELD_OFFSET(SessionOptionValue, _impl_.option_value_.bool_value_), _Internal::kOneofCaseOffset + 0, 0, (0 | ::_fl::kFcOneof | ::_fl::kBool)},
    // sfixed64 int64_value = 3;
    {PROTOBUF_FIELD_OFFSET(SessionOptionValue, _impl_.option_value_.int64_value_), _Internal::kOneofCaseOffset + 0, 0, (0 | ::_fl::kFcOneof | ::_fl::kSFixed64)},
    // double double_value = 4;
    {PROTOBUF_FIELD_OFFSET(SessionOptionValue, _impl_.option_value_.double_value_), _Internal::kOneofCaseOffset + 0, 0, (0 | ::_fl::kFcOneof | ::_fl::kDouble)},
    // .arrow.flight.protocol.SessionOptionValue.StringListValue string_list_value = 5;
    {PROTOBUF_FIELD_OFFSET(SessionOptionValue, _impl_.option_value_.string_list_value_), _Internal::kOneofCaseOffset + 0, 0, (0 | ::_fl::kFcOneof | ::_fl::kMessage | ::_fl::kTvTable)},
  }},
  {{
      {::_pbi::TcParser::GetTable<::arrow::flight::protocol::SessionOptionValue_StringListValue>()},
  }},
  {{
    "\50\14\0\0\0\0\0\0"
    "arrow.flight.protocol.SessionOptionValue"
    "string_value"
  }},
};
PROTOBUF_NOINLINE void SessionOptionValue::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.SessionOptionValue)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  clear_option_value();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL SessionOptionValue::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const SessionOptionValue& this_ = static_cast<const SessionOptionValue&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL SessionOptionValue::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const SessionOptionValue& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.SessionOptionValue)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  switch (this_.option_value_case()) {
    case kStringValue: {
      const ::std::string& _s = this_._internal_string_value();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.SessionOptionValue.string_value");
      target = stream->WriteStringMaybeAliased(1, _s, target);
      break;
    }
    case kBoolValue: {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteBoolToArray(
          2, this_._internal_bool_value(), target);
      break;
    }
    case kInt64Value: {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteSFixed64ToArray(
          3, this_._internal_int64_value(), target);
      break;
    }
    case kDoubleValue: {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteDoubleToArray(
          4, this_._internal_double_value(), target);
      break;
    }
    case kStringListValue: {
      target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
          5, *this_._impl_.option_value_.string_list_value_, this_._impl_.option_value_.string_list_value_->GetCachedSize(), target,
          stream);
      break;
    }
    default:
      break;
  }
  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.SessionOptionValue)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t SessionOptionValue::ByteSizeLong(const MessageLite& base) {
  const SessionOptionValue& this_ = static_cast<const SessionOptionValue&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t SessionOptionValue::ByteSizeLong() const {
  const SessionOptionValue& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.SessionOptionValue)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  switch (this_.option_value_case()) {
    // string string_value = 1;
    case kStringValue: {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this_._internal_string_value());
      break;
    }
    // bool bool_value = 2;
    case kBoolValue: {
      total_size += 2;
      break;
    }
    // sfixed64 int64_value = 3;
    case kInt64Value: {
      total_size += 9;
      break;
    }
    // double double_value = 4;
    case kDoubleValue: {
      total_size += 9;
      break;
    }
    // .arrow.flight.protocol.SessionOptionValue.StringListValue string_list_value = 5;
    case kStringListValue: {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.option_value_.string_list_value_);
      break;
    }
    case OPTION_VALUE_NOT_SET: {
      break;
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void SessionOptionValue::MergeImpl(::google::protobuf::MessageLite& to_msg,
                            const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this =
      static_cast<SessionOptionValue*>(&to_msg);
  auto& from = static_cast<const SessionOptionValue&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.SessionOptionValue)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  if (const uint32_t oneof_from_case =
          from._impl_._oneof_case_[0]) {
    const uint32_t oneof_to_case = _this->_impl_._oneof_case_[0];
    const bool oneof_needs_init = oneof_to_case != oneof_from_case;
    if (oneof_needs_init) {
      if (oneof_to_case != 0) {
        _this->clear_option_value();
      }
      _this->_impl_._oneof_case_[0] = oneof_from_case;
    }

    switch (oneof_from_case) {
      case kStringValue: {
        if (oneof_needs_init) {
          _this->_impl_.option_value_.string_value_.InitDefault();
        }
        _this->_impl_.option_value_.string_value_.Set(from._internal_string_value(), arena);
        break;
      }
      case kBoolValue: {
        _this->_impl_.option_value_.bool_value_ = from._impl_.option_value_.bool_value_;
        break;
      }
      case kInt64Value: {
        _this->_impl_.option_value_.int64_value_ = from._impl_.option_value_.int64_value_;
        break;
      }
      case kDoubleValue: {
        _this->_impl_.option_value_.double_value_ = from._impl_.option_value_.double_value_;
        break;
      }
      case kStringListValue: {
        if (oneof_needs_init) {
          _this->_impl_.option_value_.string_list_value_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.option_value_.string_list_value_);
        } else {
          _this->_impl_.option_value_.string_list_value_->MergeFrom(*from._impl_.option_value_.string_list_value_);
        }
        break;
      }
      case OPTION_VALUE_NOT_SET:
        break;
    }
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void SessionOptionValue::CopyFrom(const SessionOptionValue& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.SessionOptionValue)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void SessionOptionValue::InternalSwap(SessionOptionValue* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_.option_value_, other->_impl_.option_value_);
  swap(_impl_._oneof_case_[0], other->_impl_._oneof_case_[0]);
}

::google::protobuf::Metadata SessionOptionValue::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

#if defined(PROTOBUF_CUSTOM_VTABLE)
SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse::SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse()
    : SuperType(SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse_class_data_.base()) {}
SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse::SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
    : SuperType(arena, SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse_class_data_.base()) {}
#else   // PROTOBUF_CUSTOM_VTABLE
SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse::SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse() : SuperType() {}
SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse::SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse(::google::protobuf::Arena* PROTOBUF_NULLABLE arena) : SuperType(arena) {}
#endif  // PROTOBUF_CUSTOM_VTABLE
inline void* PROTOBUF_NONNULL SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse(arena);
}
constexpr auto SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse),
                                            alignof(SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse));
}
constexpr auto SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse::InternalGenerateClassData_() {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &_SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse_default_instance_._instance,
          &_table_.header,
          nullptr,  // IsInitialized
          &SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse::SharedDtor,
          static_cast<void (::google::protobuf::MessageLite::*)()>(&SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse::ClearImpl),
              ::google::protobuf::Message::ByteSizeLongImpl, ::google::protobuf::Message::_InternalSerializeImpl
              ,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse, _impl_._cached_size_),
          false,
      },
      &SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse::kDescriptorMethods,
      &descriptor_table_Flight_2eproto,
      nullptr,  // tracker
  };
}

PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse_class_data_ =
        SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse::InternalGenerateClassData_();

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse_class_data_.tc_table);
  return SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse_class_data_.base();
}
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<1, 2, 1, 78, 2>
SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse, _impl_._has_bits_),
    0, // no _extensions_
    2, 8,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967292,  // skipmap
    offsetof(decltype(_table_), field_entries),
    2,  // num_field_entries
    1,  // num_aux_entries
    offsetof(decltype(_table_), aux_entries),
    SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse_class_data_.base(),
    nullptr,  // post_loop_handler
    ::_pbi::TcParser::DiscardEverythingFallback,  // fallback
    #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
    ::_pbi::TcParser::GetTable<::arrow::flight::protocol::SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse>(),  // to_prefetch
    #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
  }, {{
    // .arrow.flight.protocol.SessionOptionValue value = 2;
    {::_pbi::TcParser::FastMtS1,
     {18, 1, 0,
      PROTOBUF_FIELD_OFFSET(SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse, _impl_.value_)}},
    // string key = 1;
    {::_pbi::TcParser::FastUS1,
     {10, 0, 0,
      PROTOBUF_FIELD_OFFSET(SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse, _impl_.key_)}},
  }}, {{
    65535, 65535
  }}, {{
    // string key = 1;
    {PROTOBUF_FIELD_OFFSET(SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse, _impl_.key_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
    // .arrow.flight.protocol.SessionOptionValue value = 2;
    {PROTOBUF_FIELD_OFFSET(SetSessionOptionsRequest_SessionOptionsEntry_DoNotUse, _impl_.value_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
  }},
  {{
      {::_pbi::TcParser::GetTable<::arrow::flight::protocol::SessionOptionValue>()},
  }},
  {{
    "\102\3\0\0\0\0\0\0"
    "arrow.flight.protocol.SetSessionOptionsRequest.SessionOptionsEntry"
    "key"
  }},
};
// ===================================================================

class SetSessionOptionsRequest::_Internal {
 public:
  using HasBits =
      decltype(::std::declval<SetSessionOptionsRequest>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(SetSessionOptionsRequest, _impl_._has_bits_);
};

SetSessionOptionsRequest::SetSessionOptionsRequest(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, SetSessionOptionsRequest_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.SetSessionOptionsRequest)
}
PROTOBUF_NDEBUG_INLINE SetSessionOptionsRequest::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::SetSessionOptionsRequest& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        #ifdef PROTOBUF_INTERNAL_REMOVE_ARENA_PTRS_MAP_FIELD
        session_options_{visibility, ::_pbi::InternalMetadataOffset::Build<
            ::arrow::flight::protocol::SetSessionOptionsRequest,
            PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::SetSessionOptionsRequest, _impl_.session_options_)>()
        , from.session_options_}
        #else
        session_options_ { visibility, arena, from.session_options_ }
        #endif
     {}

SetSessionOptionsRequest::SetSessionOptionsRequest(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const SetSessionOptionsRequest& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, SetSessionOptionsRequest_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SetSessionOptionsRequest* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.SetSessionOptionsRequest)
}
PROTOBUF_NDEBUG_INLINE SetSessionOptionsRequest::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        #ifdef PROTOBUF_INTERNAL_REMOVE_ARENA_PTRS_MAP_FIELD
        session_options_{visibility, ::_pbi::InternalMetadataOffset::Build<
            ::arrow::flight::protocol::SetSessionOptionsRequest,
            PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::SetSessionOptionsRequest, _impl_.session_options_)>()
        }
        #else
        session_options_ { visibility, arena }
        #endif
     {}

inline void SetSessionOptionsRequest::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
SetSessionOptionsRequest::~SetSessionOptionsRequest() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.SetSessionOptionsRequest)
  SharedDtor(*this);
}
inline void SetSessionOptionsRequest::SharedDtor(MessageLite& self) {
  SetSessionOptionsRequest& this_ = static_cast<SetSessionOptionsRequest&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.~Impl_();
}

inline void* PROTOBUF_NONNULL SetSessionOptionsRequest::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) SetSessionOptionsRequest(arena);
}
#ifdef PROTOBUF_INTERNAL_REMOVE_ARENA_PTRS_REPEATED_PTR_FIELD
constexpr auto SetSessionOptionsRequest::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(SetSessionOptionsRequest),
                                            alignof(SetSessionOptionsRequest));
}
#else  // !PROTOBUF_INTERNAL_REMOVE_ARENA_PTRS_REPEATED_PTR_FIELD
constexpr auto SetSessionOptionsRequest::InternalNewImpl_() {
  constexpr auto arena_bits = ::google::protobuf::internal::EncodePlacementArenaOffsets({
      PROTOBUF_FIELD_OFFSET(SetSessionOptionsRequest, _impl_.session_options_) +
          decltype(SetSessionOptionsRequest::_impl_.session_options_)::
              InternalGetArenaOffset(
                  ::google::protobuf::Message::internal_visibility()),
  });
  if (arena_bits.has_value()) {
    return ::google::protobuf::internal::MessageCreator::CopyInit(
        sizeof(SetSessionOptionsRequest), alignof(SetSessionOptionsRequest), *arena_bits);
  } else {
    return ::google::protobuf::internal::MessageCreator(&SetSessionOptionsRequest::PlacementNew_,
                                 sizeof(SetSessionOptionsRequest),
                                 alignof(SetSessionOptionsRequest));
  }
}
#endif
constexpr auto SetSessionOptionsRequest::InternalGenerateClassData_() {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &_SetSessionOptionsRequest_default_instance_._instance,
          &_table_.header,
          nullptr,  // IsInitialized
          &SetSessionOptionsRequest::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<SetSessionOptionsRequest>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &SetSessionOptionsRequest::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<SetSessionOptionsRequest>(), &SetSessionOptionsRequest::ByteSizeLong,
              &SetSessionOptionsRequest::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(SetSessionOptionsRequest, _impl_._cached_size_),
          false,
      },
      &SetSessionOptionsRequest::kDescriptorMethods,
      &descriptor_table_Flight_2eproto,
      nullptr,  // tracker
  };
}

PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull SetSessionOptionsRequest_class_data_ =
        SetSessionOptionsRequest::InternalGenerateClassData_();

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
SetSessionOptionsRequest::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&SetSessionOptionsRequest_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(SetSessionOptionsRequest_class_data_.tc_table);
  return SetSessionOptionsRequest_class_data_.base();
}
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 2, 70, 2>
SetSessionOptionsRequest::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(SetSessionOptionsRequest, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    2,  // num_aux_entries
    offsetof(decltype(_table_), aux_entries),
    SetSessionOptionsRequest_class_data_.base(),
    nullptr,  // post_loop_handler
    ::_pbi::TcParser::GenericFallback,  // fallback
    #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
    ::_pbi::TcParser::GetTable<::arrow::flight::protocol::SetSessionOptionsRequest>(),  // to_prefetch
    #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
  }, {{
    {::_pbi::TcParser::MiniParse, {}},
  }}, {{
    65535, 65535
  }}, {{
    // map<string, .arrow.flight.protocol.SessionOptionValue> session_options = 1;
    {PROTOBUF_FIELD_OFFSET(SetSessionOptionsRequest, _impl_.session_options_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcRepeated | ::_fl::kMap)},
  }},
  {{
      {::_pbi::TcParser::GetMapAuxInfo(
          1, 0, 9, 11, 0)},
      {::_pbi::TcParser::GetTable<::arrow::flight::protocol::SessionOptionValue>()},
  }},
  {{
    "\56\17\0\0\0\0\0\0"
    "arrow.flight.protocol.SetSessionOptionsRequest"
    "session_options"
  }},
};
PROTOBUF_NOINLINE void SetSessionOptionsRequest::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.SetSessionOptionsRequest)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBitForRepeated(cached_has_bits, 0x00000001U)) {
    _impl_.session_options_.Clear();
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL SetSessionOptionsRequest::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const SetSessionOptionsRequest& this_ = static_cast<const SetSessionOptionsRequest&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL SetSessionOptionsRequest::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const SetSessionOptionsRequest& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.SetSessionOptionsRequest)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // map<string, .arrow.flight.protocol.SessionOptionValue> session_options = 1;
  if (CheckHasBitForRepeated(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_session_options().empty()) {
      using MapType = ::google::protobuf::Map<::std::string, ::arrow::flight::protocol::SessionOptionValue>;
      using WireHelper = _pbi::MapEntryFuncs<::std::string, ::arrow::flight::protocol::SessionOptionValue,
                                     _pbi::WireFormatLite::TYPE_STRING,
                                     _pbi::WireFormatLite::TYPE_MESSAGE>;
      const auto& field = this_._internal_session_options();

      if (stream->IsSerializationDeterministic() && field.size() > 1) {
        for (const auto& entry : ::google::protobuf::internal::MapSorterPtr<MapType>(field)) {
          target = WireHelper::InternalSerialize(
              1, entry.first, entry.second, target, stream);
          ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
              entry.first.data(), static_cast<int>(entry.first.length()),
 ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.SetSessionOptionsRequest.session_options");
        }
      } else {
        for (const auto& entry : field) {
          target = WireHelper::InternalSerialize(
              1, entry.first, entry.second, target, stream);
          ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
              entry.first.data(), static_cast<int>(entry.first.length()),
 ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.SetSessionOptionsRequest.session_options");
        }
      }
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.SetSessionOptionsRequest)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t SetSessionOptionsRequest::ByteSizeLong(const MessageLite& base) {
  const SetSessionOptionsRequest& this_ = static_cast<const SetSessionOptionsRequest&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t SetSessionOptionsRequest::ByteSizeLong() const {
  const SetSessionOptionsRequest& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.SetSessionOptionsRequest)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
   {
    // map<string, .arrow.flight.protocol.SessionOptionValue> session_options = 1;
    cached_has_bits = this_._impl_._has_bits_[0];
    if (CheckHasBitForRepeated(cached_has_bits, 0x00000001U)) {
      total_size +=
          1 * ::google::protobuf::internal::FromIntSize(this_._internal_session_options_size());
      for (const auto& entry : this_._internal_session_options()) {
        total_size += _pbi::MapEntryFuncs<::std::string, ::arrow::flight::protocol::SessionOptionValue,
                                       _pbi::WireFormatLite::TYPE_STRING,
                                       _pbi::WireFormatLite::TYPE_MESSAGE>::ByteSizeLong(entry.first, entry.second);
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void SetSessionOptionsRequest::MergeImpl(::google::protobuf::MessageLite& to_msg,
                            const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this =
      static_cast<SetSessionOptionsRequest*>(&to_msg);
  auto& from = static_cast<const SetSessionOptionsRequest&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.SetSessionOptionsRequest)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (CheckHasBitForRepeated(cached_has_bits, 0x00000001U)) {
    _this->_impl_.session_options_.MergeFrom(from._impl_.session_options_);
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void SetSessionOptionsRequest::CopyFrom(const SetSessionOptionsRequest& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.SetSessionOptionsRequest)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void SetSessionOptionsRequest::InternalSwap(SetSessionOptionsRequest* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.session_options_.InternalSwap(&other->_impl_.session_options_);
}

::google::protobuf::Metadata SetSessionOptionsRequest::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

class SetSessionOptionsResult_Error::_Internal {
 public:
  using HasBits =
      decltype(::std::declval<SetSessionOptionsResult_Error>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(SetSessionOptionsResult_Error, _impl_._has_bits_);
};

SetSessionOptionsResult_Error::SetSessionOptionsResult_Error(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, SetSessionOptionsResult_Error_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.SetSessionOptionsResult.Error)
}
SetSessionOptionsResult_Error::SetSessionOptionsResult_Error(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const SetSessionOptionsResult_Error& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, SetSessionOptionsResult_Error_class_data_.base()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(from._impl_) {
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}
PROTOBUF_NDEBUG_INLINE SetSessionOptionsResult_Error::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0} {}

inline void SetSessionOptionsResult_Error::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.value_ = {};
}
SetSessionOptionsResult_Error::~SetSessionOptionsResult_Error() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.SetSessionOptionsResult.Error)
  SharedDtor(*this);
}
inline void SetSessionOptionsResult_Error::SharedDtor(MessageLite& self) {
  SetSessionOptionsResult_Error& this_ = static_cast<SetSessionOptionsResult_Error&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.~Impl_();
}

inline void* PROTOBUF_NONNULL SetSessionOptionsResult_Error::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) SetSessionOptionsResult_Error(arena);
}
constexpr auto SetSessionOptionsResult_Error::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::ZeroInit(sizeof(SetSessionOptionsResult_Error),
                                            alignof(SetSessionOptionsResult_Error));
}
constexpr auto SetSessionOptionsResult_Error::InternalGenerateClassData_() {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &_SetSessionOptionsResult_Error_default_instance_._instance,
          &_table_.header,
          nullptr,  // IsInitialized
          &SetSessionOptionsResult_Error::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<SetSessionOptionsResult_Error>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &SetSessionOptionsResult_Error::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<SetSessionOptionsResult_Error>(), &SetSessionOptionsResult_Error::ByteSizeLong,
              &SetSessionOptionsResult_Error::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(SetSessionOptionsResult_Error, _impl_._cached_size_),
          false,
      },
      &SetSessionOptionsResult_Error::kDescriptorMethods,
      &descriptor_table_Flight_2eproto,
      nullptr,  // tracker
  };
}

PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull SetSessionOptionsResult_Error_class_data_ =
        SetSessionOptionsResult_Error::InternalGenerateClassData_();

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
SetSessionOptionsResult_Error::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&SetSessionOptionsResult_Error_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(SetSessionOptionsResult_Error_class_data_.tc_table);
  return SetSessionOptionsResult_Error_class_data_.base();
}
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2>
SetSessionOptionsResult_Error::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(SetSessionOptionsResult_Error, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    SetSessionOptionsResult_Error_class_data_.base(),
    nullptr,  // post_loop_handler
    ::_pbi::TcParser::GenericFallback,  // fallback
    #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
    ::_pbi::TcParser::GetTable<::arrow::flight::protocol::SetSessionOptionsResult_Error>(),  // to_prefetch
    #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
  }, {{
    // .arrow.flight.protocol.SetSessionOptionsResult.ErrorValue value = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(SetSessionOptionsResult_Error, _impl_.value_), 0>(),
     {8, 0, 0,
      PROTOBUF_FIELD_OFFSET(SetSessionOptionsResult_Error, _impl_.value_)}},
  }}, {{
    65535, 65535
  }}, {{
    // .arrow.flight.protocol.SetSessionOptionsResult.ErrorValue value = 1;
    {PROTOBUF_FIELD_OFFSET(SetSessionOptionsResult_Error, _impl_.value_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kOpenEnum)},
  }},
  // no aux_entries
  {{
  }},
};
PROTOBUF_NOINLINE void SetSessionOptionsResult_Error::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.SetSessionOptionsResult.Error)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.value_ = 0;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL SetSessionOptionsResult_Error::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const SetSessionOptionsResult_Error& this_ = static_cast<const SetSessionOptionsResult_Error&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL SetSessionOptionsResult_Error::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const SetSessionOptionsResult_Error& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.SetSessionOptionsResult.Error)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // .arrow.flight.protocol.SetSessionOptionsResult.ErrorValue value = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (this_._internal_value() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteEnumToArray(
          1, this_._internal_value(), target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.SetSessionOptionsResult.Error)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t SetSessionOptionsResult_Error::ByteSizeLong(const MessageLite& base) {
  const SetSessionOptionsResult_Error& this_ = static_cast<const SetSessionOptionsResult_Error&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t SetSessionOptionsResult_Error::ByteSizeLong() const {
  const SetSessionOptionsResult_Error& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.SetSessionOptionsResult.Error)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

   {
    // .arrow.flight.protocol.SetSessionOptionsResult.ErrorValue value = 1;
    cached_has_bits = this_._impl_._has_bits_[0];
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (this_._internal_value() != 0) {
        total_size += 1 +
                      ::_pbi::WireFormatLite::EnumSize(this_._internal_value());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void SetSessionOptionsResult_Error::MergeImpl(::google::protobuf::MessageLite& to_msg,
                            const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this =
      static_cast<SetSessionOptionsResult_Error*>(&to_msg);
  auto& from = static_cast<const SetSessionOptionsResult_Error&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.SetSessionOptionsResult.Error)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (from._internal_value() != 0) {
      _this->_impl_.value_ = from._impl_.value_;
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void SetSessionOptionsResult_Error::CopyFrom(const SetSessionOptionsResult_Error& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.SetSessionOptionsResult.Error)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void SetSessionOptionsResult_Error::InternalSwap(SetSessionOptionsResult_Error* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.value_, other->_impl_.value_);
}

::google::protobuf::Metadata SetSessionOptionsResult_Error::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

#if defined(PROTOBUF_CUSTOM_VTABLE)
SetSessionOptionsResult_ErrorsEntry_DoNotUse::SetSessionOptionsResult_ErrorsEntry_DoNotUse()
    : SuperType(SetSessionOptionsResult_ErrorsEntry_DoNotUse_class_data_.base()) {}
SetSessionOptionsResult_ErrorsEntry_DoNotUse::SetSessionOptionsResult_ErrorsEntry_DoNotUse(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
    : SuperType(arena, SetSessionOptionsResult_ErrorsEntry_DoNotUse_class_data_.base()) {}
#else   // PROTOBUF_CUSTOM_VTABLE
SetSessionOptionsResult_ErrorsEntry_DoNotUse::SetSessionOptionsResult_ErrorsEntry_DoNotUse() : SuperType() {}
SetSessionOptionsResult_ErrorsEntry_DoNotUse::SetSessionOptionsResult_ErrorsEntry_DoNotUse(::google::protobuf::Arena* PROTOBUF_NULLABLE arena) : SuperType(arena) {}
#endif  // PROTOBUF_CUSTOM_VTABLE
inline void* PROTOBUF_NONNULL SetSessionOptionsResult_ErrorsEntry_DoNotUse::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) SetSessionOptionsResult_ErrorsEntry_DoNotUse(arena);
}
constexpr auto SetSessionOptionsResult_ErrorsEntry_DoNotUse::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(SetSessionOptionsResult_ErrorsEntry_DoNotUse),
                                            alignof(SetSessionOptionsResult_ErrorsEntry_DoNotUse));
}
constexpr auto SetSessionOptionsResult_ErrorsEntry_DoNotUse::InternalGenerateClassData_() {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &_SetSessionOptionsResult_ErrorsEntry_DoNotUse_default_instance_._instance,
          &_table_.header,
          nullptr,  // IsInitialized
          &SetSessionOptionsResult_ErrorsEntry_DoNotUse::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<SetSessionOptionsResult_ErrorsEntry_DoNotUse>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &SetSessionOptionsResult_ErrorsEntry_DoNotUse::SharedDtor,
          static_cast<void (::google::protobuf::MessageLite::*)()>(&SetSessionOptionsResult_ErrorsEntry_DoNotUse::ClearImpl),
              ::google::protobuf::Message::ByteSizeLongImpl, ::google::protobuf::Message::_InternalSerializeImpl
              ,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(SetSessionOptionsResult_ErrorsEntry_DoNotUse, _impl_._cached_size_),
          false,
      },
      &SetSessionOptionsResult_ErrorsEntry_DoNotUse::kDescriptorMethods,
      &descriptor_table_Flight_2eproto,
      nullptr,  // tracker
  };
}

PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull SetSessionOptionsResult_ErrorsEntry_DoNotUse_class_data_ =
        SetSessionOptionsResult_ErrorsEntry_DoNotUse::InternalGenerateClassData_();

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
SetSessionOptionsResult_ErrorsEntry_DoNotUse::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&SetSessionOptionsResult_ErrorsEntry_DoNotUse_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(SetSessionOptionsResult_ErrorsEntry_DoNotUse_class_data_.tc_table);
  return SetSessionOptionsResult_ErrorsEntry_DoNotUse_class_data_.base();
}
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<1, 2, 1, 69, 2>
SetSessionOptionsResult_ErrorsEntry_DoNotUse::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(SetSessionOptionsResult_ErrorsEntry_DoNotUse, _impl_._has_bits_),
    0, // no _extensions_
    2, 8,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967292,  // skipmap
    offsetof(decltype(_table_), field_entries),
    2,  // num_field_entries
    1,  // num_aux_entries
    offsetof(decltype(_table_), aux_entries),
    SetSessionOptionsResult_ErrorsEntry_DoNotUse_class_data_.base(),
    nullptr,  // post_loop_handler
    ::_pbi::TcParser::DiscardEverythingFallback,  // fallback
    #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
    ::_pbi::TcParser::GetTable<::arrow::flight::protocol::SetSessionOptionsResult_ErrorsEntry_DoNotUse>(),  // to_prefetch
    #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
  }, {{
    // .arrow.flight.protocol.SetSessionOptionsResult.Error value = 2;
    {::_pbi::TcParser::FastMtS1,
     {18, 1, 0,
      PROTOBUF_FIELD_OFFSET(SetSessionOptionsResult_ErrorsEntry_DoNotUse, _impl_.value_)}},
    // string key = 1;
    {::_pbi::TcParser::FastUS1,
     {10, 0, 0,
      PROTOBUF_FIELD_OFFSET(SetSessionOptionsResult_ErrorsEntry_DoNotUse, _impl_.key_)}},
  }}, {{
    65535, 65535
  }}, {{
    // string key = 1;
    {PROTOBUF_FIELD_OFFSET(SetSessionOptionsResult_ErrorsEntry_DoNotUse, _impl_.key_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
    // .arrow.flight.protocol.SetSessionOptionsResult.Error value = 2;
    {PROTOBUF_FIELD_OFFSET(SetSessionOptionsResult_ErrorsEntry_DoNotUse, _impl_.value_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
  }},
  {{
      {::_pbi::TcParser::GetTable<::arrow::flight::protocol::SetSessionOptionsResult_Error>()},
  }},
  {{
    "\71\3\0\0\0\0\0\0"
    "arrow.flight.protocol.SetSessionOptionsResult.ErrorsEntry"
    "key"
  }},
};
// ===================================================================

class SetSessionOptionsResult::_Internal {
 public:
  using HasBits =
      decltype(::std::declval<SetSessionOptionsResult>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(SetSessionOptionsResult, _impl_._has_bits_);
};

SetSessionOptionsResult::SetSessionOptionsResult(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, SetSessionOptionsResult_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.SetSessionOptionsResult)
}
PROTOBUF_NDEBUG_INLINE SetSessionOptionsResult::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::SetSessionOptionsResult& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        #ifdef PROTOBUF_INTERNAL_REMOVE_ARENA_PTRS_MAP_FIELD
        errors_{visibility, ::_pbi::InternalMetadataOffset::Build<
            ::arrow::flight::protocol::SetSessionOptionsResult,
            PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::SetSessionOptionsResult, _impl_.errors_)>()
        , from.errors_}
        #else
        errors_ { visibility, arena, from.errors_ }
        #endif
     {}

SetSessionOptionsResult::SetSessionOptionsResult(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const SetSessionOptionsResult& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, SetSessionOptionsResult_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SetSessionOptionsResult* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.SetSessionOptionsResult)
}
PROTOBUF_NDEBUG_INLINE SetSessionOptionsResult::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        #ifdef PROTOBUF_INTERNAL_REMOVE_ARENA_PTRS_MAP_FIELD
        errors_{visibility, ::_pbi::InternalMetadataOffset::Build<
            ::arrow::flight::protocol::SetSessionOptionsResult,
            PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::SetSessionOptionsResult, _impl_.errors_)>()
        }
        #else
        errors_ { visibility, arena }
        #endif
     {}

inline void SetSessionOptionsResult::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
SetSessionOptionsResult::~SetSessionOptionsResult() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.SetSessionOptionsResult)
  SharedDtor(*this);
}
inline void SetSessionOptionsResult::SharedDtor(MessageLite& self) {
  SetSessionOptionsResult& this_ = static_cast<SetSessionOptionsResult&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.~Impl_();
}

inline void* PROTOBUF_NONNULL SetSessionOptionsResult::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) SetSessionOptionsResult(arena);
}
#ifdef PROTOBUF_INTERNAL_REMOVE_ARENA_PTRS_REPEATED_PTR_FIELD
constexpr auto SetSessionOptionsResult::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(SetSessionOptionsResult),
                                            alignof(SetSessionOptionsResult));
}
#else  // !PROTOBUF_INTERNAL_REMOVE_ARENA_PTRS_REPEATED_PTR_FIELD
constexpr auto SetSessionOptionsResult::InternalNewImpl_() {
  constexpr auto arena_bits = ::google::protobuf::internal::EncodePlacementArenaOffsets({
      PROTOBUF_FIELD_OFFSET(SetSessionOptionsResult, _impl_.errors_) +
          decltype(SetSessionOptionsResult::_impl_.errors_)::
              InternalGetArenaOffset(
                  ::google::protobuf::Message::internal_visibility()),
  });
  if (arena_bits.has_value()) {
    return ::google::protobuf::internal::MessageCreator::CopyInit(
        sizeof(SetSessionOptionsResult), alignof(SetSessionOptionsResult), *arena_bits);
  } else {
    return ::google::protobuf::internal::MessageCreator(&SetSessionOptionsResult::PlacementNew_,
                                 sizeof(SetSessionOptionsResult),
                                 alignof(SetSessionOptionsResult));
  }
}
#endif
constexpr auto SetSessionOptionsResult::InternalGenerateClassData_() {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &_SetSessionOptionsResult_default_instance_._instance,
          &_table_.header,
          nullptr,  // IsInitialized
          &SetSessionOptionsResult::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<SetSessionOptionsResult>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &SetSessionOptionsResult::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<SetSessionOptionsResult>(), &SetSessionOptionsResult::ByteSizeLong,
              &SetSessionOptionsResult::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(SetSessionOptionsResult, _impl_._cached_size_),
          false,
      },
      &SetSessionOptionsResult::kDescriptorMethods,
      &descriptor_table_Flight_2eproto,
      nullptr,  // tracker
  };
}

PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull SetSessionOptionsResult_class_data_ =
        SetSessionOptionsResult::InternalGenerateClassData_();

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
SetSessionOptionsResult::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&SetSessionOptionsResult_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(SetSessionOptionsResult_class_data_.tc_table);
  return SetSessionOptionsResult_class_data_.base();
}
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 2, 60, 2>
SetSessionOptionsResult::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(SetSessionOptionsResult, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    2,  // num_aux_entries
    offsetof(decltype(_table_), aux_entries),
    SetSessionOptionsResult_class_data_.base(),
    nullptr,  // post_loop_handler
    ::_pbi::TcParser::GenericFallback,  // fallback
    #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
    ::_pbi::TcParser::GetTable<::arrow::flight::protocol::SetSessionOptionsResult>(),  // to_prefetch
    #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
  }, {{
    {::_pbi::TcParser::MiniParse, {}},
  }}, {{
    65535, 65535
  }}, {{
    // map<string, .arrow.flight.protocol.SetSessionOptionsResult.Error> errors = 1;
    {PROTOBUF_FIELD_OFFSET(SetSessionOptionsResult, _impl_.errors_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcRepeated | ::_fl::kMap)},
  }},
  {{
      {::_pbi::TcParser::GetMapAuxInfo(
          1, 0, 9, 11, 0)},
      {::_pbi::TcParser::GetTable<::arrow::flight::protocol::SetSessionOptionsResult_Error>()},
  }},
  {{
    "\55\6\0\0\0\0\0\0"
    "arrow.flight.protocol.SetSessionOptionsResult"
    "errors"
  }},
};
PROTOBUF_NOINLINE void SetSessionOptionsResult::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.SetSessionOptionsResult)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBitForRepeated(cached_has_bits, 0x00000001U)) {
    _impl_.errors_.Clear();
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL SetSessionOptionsResult::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const SetSessionOptionsResult& this_ = static_cast<const SetSessionOptionsResult&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL SetSessionOptionsResult::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const SetSessionOptionsResult& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.SetSessionOptionsResult)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // map<string, .arrow.flight.protocol.SetSessionOptionsResult.Error> errors = 1;
  if (CheckHasBitForRepeated(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_errors().empty()) {
      using MapType = ::google::protobuf::Map<::std::string, ::arrow::flight::protocol::SetSessionOptionsResult_Error>;
      using WireHelper = _pbi::MapEntryFuncs<::std::string, ::arrow::flight::protocol::SetSessionOptionsResult_Error,
                                     _pbi::WireFormatLite::TYPE_STRING,
                                     _pbi::WireFormatLite::TYPE_MESSAGE>;
      const auto& field = this_._internal_errors();

      if (stream->IsSerializationDeterministic() && field.size() > 1) {
        for (const auto& entry : ::google::protobuf::internal::MapSorterPtr<MapType>(field)) {
          target = WireHelper::InternalSerialize(
              1, entry.first, entry.second, target, stream);
          ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
              entry.first.data(), static_cast<int>(entry.first.length()),
 ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.SetSessionOptionsResult.errors");
        }
      } else {
        for (const auto& entry : field) {
          target = WireHelper::InternalSerialize(
              1, entry.first, entry.second, target, stream);
          ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
              entry.first.data(), static_cast<int>(entry.first.length()),
 ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.SetSessionOptionsResult.errors");
        }
      }
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.SetSessionOptionsResult)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t SetSessionOptionsResult::ByteSizeLong(const MessageLite& base) {
  const SetSessionOptionsResult& this_ = static_cast<const SetSessionOptionsResult&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t SetSessionOptionsResult::ByteSizeLong() const {
  const SetSessionOptionsResult& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.SetSessionOptionsResult)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
   {
    // map<string, .arrow.flight.protocol.SetSessionOptionsResult.Error> errors = 1;
    cached_has_bits = this_._impl_._has_bits_[0];
    if (CheckHasBitForRepeated(cached_has_bits, 0x00000001U)) {
      total_size +=
          1 * ::google::protobuf::internal::FromIntSize(this_._internal_errors_size());
      for (const auto& entry : this_._internal_errors()) {
        total_size += _pbi::MapEntryFuncs<::std::string, ::arrow::flight::protocol::SetSessionOptionsResult_Error,
                                       _pbi::WireFormatLite::TYPE_STRING,
                                       _pbi::WireFormatLite::TYPE_MESSAGE>::ByteSizeLong(entry.first, entry.second);
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void SetSessionOptionsResult::MergeImpl(::google::protobuf::MessageLite& to_msg,
                            const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this =
      static_cast<SetSessionOptionsResult*>(&to_msg);
  auto& from = static_cast<const SetSessionOptionsResult&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.SetSessionOptionsResult)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (CheckHasBitForRepeated(cached_has_bits, 0x00000001U)) {
    _this->_impl_.errors_.MergeFrom(from._impl_.errors_);
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void SetSessionOptionsResult::CopyFrom(const SetSessionOptionsResult& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.SetSessionOptionsResult)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void SetSessionOptionsResult::InternalSwap(SetSessionOptionsResult* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.errors_.InternalSwap(&other->_impl_.errors_);
}

::google::protobuf::Metadata SetSessionOptionsResult::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

class GetSessionOptionsRequest::_Internal {
 public:
};

GetSessionOptionsRequest::GetSessionOptionsRequest(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::internal::ZeroFieldsBase(arena, GetSessionOptionsRequest_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::internal::ZeroFieldsBase(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.GetSessionOptionsRequest)
}
GetSessionOptionsRequest::GetSessionOptionsRequest(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const GetSessionOptionsRequest& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::internal::ZeroFieldsBase(arena, GetSessionOptionsRequest_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::internal::ZeroFieldsBase(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  GetSessionOptionsRequest* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.GetSessionOptionsRequest)
}

inline void* PROTOBUF_NONNULL GetSessionOptionsRequest::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) GetSessionOptionsRequest(arena);
}
constexpr auto GetSessionOptionsRequest::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::ZeroInit(sizeof(GetSessionOptionsRequest),
                                            alignof(GetSessionOptionsRequest));
}
constexpr auto GetSessionOptionsRequest::InternalGenerateClassData_() {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &_GetSessionOptionsRequest_default_instance_._instance,
          &_table_.header,
          nullptr,  // IsInitialized
          &GetSessionOptionsRequest::MergeImpl,
          ::google::protobuf::internal::ZeroFieldsBase::GetNewImpl<GetSessionOptionsRequest>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &GetSessionOptionsRequest::SharedDtor,
          ::google::protobuf::internal::ZeroFieldsBase::GetClearImpl<GetSessionOptionsRequest>(), &GetSessionOptionsRequest::ByteSizeLong,
              &GetSessionOptionsRequest::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(GetSessionOptionsRequest, _impl_._cached_size_),
          false,
      },
      &GetSessionOptionsRequest::kDescriptorMethods,
      &descriptor_table_Flight_2eproto,
      nullptr,  // tracker
  };
}

PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull GetSessionOptionsRequest_class_data_ =
        GetSessionOptionsRequest::InternalGenerateClassData_();

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
GetSessionOptionsRequest::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&GetSessionOptionsRequest_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(GetSessionOptionsRequest_class_data_.tc_table);
  return GetSessionOptionsRequest_class_data_.base();
}
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 0, 0, 0, 2>
GetSessionOptionsRequest::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(GetSessionOptionsRequest,
                          _impl_._cached_size_),  // no hasbits
    0, // no _extensions_
    0, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967295,  // skipmap
    offsetof(decltype(_table_), field_names),  // no field_entries
    0,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    GetSessionOptionsRequest_class_data_.base(),
    nullptr,  // post_loop_handler
    ::_pbi::TcParser::GenericFallback,  // fallback
    #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
    ::_pbi::TcParser::GetTable<::arrow::flight::protocol::GetSessionOptionsRequest>(),  // to_prefetch
    #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
  }, {{
    {::_pbi::TcParser::MiniParse, {}},
  }}, {{
    65535, 65535
  }}, // no field_entries, or aux_entries
  {{
  }},
};







::google::protobuf::Metadata GetSessionOptionsRequest::GetMetadata() const {
  return ::google::protobuf::internal::ZeroFieldsBase::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

#if defined(PROTOBUF_CUSTOM_VTABLE)
GetSessionOptionsResult_SessionOptionsEntry_DoNotUse::GetSessionOptionsResult_SessionOptionsEntry_DoNotUse()
    : SuperType(GetSessionOptionsResult_SessionOptionsEntry_DoNotUse_class_data_.base()) {}
GetSessionOptionsResult_SessionOptionsEntry_DoNotUse::GetSessionOptionsResult_SessionOptionsEntry_DoNotUse(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
    : SuperType(arena, GetSessionOptionsResult_SessionOptionsEntry_DoNotUse_class_data_.base()) {}
#else   // PROTOBUF_CUSTOM_VTABLE
GetSessionOptionsResult_SessionOptionsEntry_DoNotUse::GetSessionOptionsResult_SessionOptionsEntry_DoNotUse() : SuperType() {}
GetSessionOptionsResult_SessionOptionsEntry_DoNotUse::GetSessionOptionsResult_SessionOptionsEntry_DoNotUse(::google::protobuf::Arena* PROTOBUF_NULLABLE arena) : SuperType(arena) {}
#endif  // PROTOBUF_CUSTOM_VTABLE
inline void* PROTOBUF_NONNULL GetSessionOptionsResult_SessionOptionsEntry_DoNotUse::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) GetSessionOptionsResult_SessionOptionsEntry_DoNotUse(arena);
}
constexpr auto GetSessionOptionsResult_SessionOptionsEntry_DoNotUse::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(GetSessionOptionsResult_SessionOptionsEntry_DoNotUse),
                                            alignof(GetSessionOptionsResult_SessionOptionsEntry_DoNotUse));
}
constexpr auto GetSessionOptionsResult_SessionOptionsEntry_DoNotUse::InternalGenerateClassData_() {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &_GetSessionOptionsResult_SessionOptionsEntry_DoNotUse_default_instance_._instance,
          &_table_.header,
          nullptr,  // IsInitialized
          &GetSessionOptionsResult_SessionOptionsEntry_DoNotUse::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<GetSessionOptionsResult_SessionOptionsEntry_DoNotUse>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &GetSessionOptionsResult_SessionOptionsEntry_DoNotUse::SharedDtor,
          static_cast<void (::google::protobuf::MessageLite::*)()>(&GetSessionOptionsResult_SessionOptionsEntry_DoNotUse::ClearImpl),
              ::google::protobuf::Message::ByteSizeLongImpl, ::google::protobuf::Message::_InternalSerializeImpl
              ,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(GetSessionOptionsResult_SessionOptionsEntry_DoNotUse, _impl_._cached_size_),
          false,
      },
      &GetSessionOptionsResult_SessionOptionsEntry_DoNotUse::kDescriptorMethods,
      &descriptor_table_Flight_2eproto,
      nullptr,  // tracker
  };
}

PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull GetSessionOptionsResult_SessionOptionsEntry_DoNotUse_class_data_ =
        GetSessionOptionsResult_SessionOptionsEntry_DoNotUse::InternalGenerateClassData_();

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
GetSessionOptionsResult_SessionOptionsEntry_DoNotUse::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&GetSessionOptionsResult_SessionOptionsEntry_DoNotUse_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(GetSessionOptionsResult_SessionOptionsEntry_DoNotUse_class_data_.tc_table);
  return GetSessionOptionsResult_SessionOptionsEntry_DoNotUse_class_data_.base();
}
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<1, 2, 1, 77, 2>
GetSessionOptionsResult_SessionOptionsEntry_DoNotUse::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(GetSessionOptionsResult_SessionOptionsEntry_DoNotUse, _impl_._has_bits_),
    0, // no _extensions_
    2, 8,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967292,  // skipmap
    offsetof(decltype(_table_), field_entries),
    2,  // num_field_entries
    1,  // num_aux_entries
    offsetof(decltype(_table_), aux_entries),
    GetSessionOptionsResult_SessionOptionsEntry_DoNotUse_class_data_.base(),
    nullptr,  // post_loop_handler
    ::_pbi::TcParser::DiscardEverythingFallback,  // fallback
    #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
    ::_pbi::TcParser::GetTable<::arrow::flight::protocol::GetSessionOptionsResult_SessionOptionsEntry_DoNotUse>(),  // to_prefetch
    #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
  }, {{
    // .arrow.flight.protocol.SessionOptionValue value = 2;
    {::_pbi::TcParser::FastMtS1,
     {18, 1, 0,
      PROTOBUF_FIELD_OFFSET(GetSessionOptionsResult_SessionOptionsEntry_DoNotUse, _impl_.value_)}},
    // string key = 1;
    {::_pbi::TcParser::FastUS1,
     {10, 0, 0,
      PROTOBUF_FIELD_OFFSET(GetSessionOptionsResult_SessionOptionsEntry_DoNotUse, _impl_.key_)}},
  }}, {{
    65535, 65535
  }}, {{
    // string key = 1;
    {PROTOBUF_FIELD_OFFSET(GetSessionOptionsResult_SessionOptionsEntry_DoNotUse, _impl_.key_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
    // .arrow.flight.protocol.SessionOptionValue value = 2;
    {PROTOBUF_FIELD_OFFSET(GetSessionOptionsResult_SessionOptionsEntry_DoNotUse, _impl_.value_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
  }},
  {{
      {::_pbi::TcParser::GetTable<::arrow::flight::protocol::SessionOptionValue>()},
  }},
  {{
    "\101\3\0\0\0\0\0\0"
    "arrow.flight.protocol.GetSessionOptionsResult.SessionOptionsEntry"
    "key"
  }},
};
// ===================================================================

class GetSessionOptionsResult::_Internal {
 public:
  using HasBits =
      decltype(::std::declval<GetSessionOptionsResult>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(GetSessionOptionsResult, _impl_._has_bits_);
};

GetSessionOptionsResult::GetSessionOptionsResult(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, GetSessionOptionsResult_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.GetSessionOptionsResult)
}
PROTOBUF_NDEBUG_INLINE GetSessionOptionsResult::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::GetSessionOptionsResult& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        #ifdef PROTOBUF_INTERNAL_REMOVE_ARENA_PTRS_MAP_FIELD
        session_options_{visibility, ::_pbi::InternalMetadataOffset::Build<
            ::arrow::flight::protocol::GetSessionOptionsResult,
            PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::GetSessionOptionsResult, _impl_.session_options_)>()
        , from.session_options_}
        #else
        session_options_ { visibility, arena, from.session_options_ }
        #endif
     {}

GetSessionOptionsResult::GetSessionOptionsResult(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const GetSessionOptionsResult& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, GetSessionOptionsResult_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  GetSessionOptionsResult* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.GetSessionOptionsResult)
}
PROTOBUF_NDEBUG_INLINE GetSessionOptionsResult::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        #ifdef PROTOBUF_INTERNAL_REMOVE_ARENA_PTRS_MAP_FIELD
        session_options_{visibility, ::_pbi::InternalMetadataOffset::Build<
            ::arrow::flight::protocol::GetSessionOptionsResult,
            PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::GetSessionOptionsResult, _impl_.session_options_)>()
        }
        #else
        session_options_ { visibility, arena }
        #endif
     {}

inline void GetSessionOptionsResult::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
GetSessionOptionsResult::~GetSessionOptionsResult() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.GetSessionOptionsResult)
  SharedDtor(*this);
}
inline void GetSessionOptionsResult::SharedDtor(MessageLite& self) {
  GetSessionOptionsResult& this_ = static_cast<GetSessionOptionsResult&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.~Impl_();
}

inline void* PROTOBUF_NONNULL GetSessionOptionsResult::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) GetSessionOptionsResult(arena);
}
#ifdef PROTOBUF_INTERNAL_REMOVE_ARENA_PTRS_REPEATED_PTR_FIELD
constexpr auto GetSessionOptionsResult::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(GetSessionOptionsResult),
                                            alignof(GetSessionOptionsResult));
}
#else  // !PROTOBUF_INTERNAL_REMOVE_ARENA_PTRS_REPEATED_PTR_FIELD
constexpr auto GetSessionOptionsResult::InternalNewImpl_() {
  constexpr auto arena_bits = ::google::protobuf::internal::EncodePlacementArenaOffsets({
      PROTOBUF_FIELD_OFFSET(GetSessionOptionsResult, _impl_.session_options_) +
          decltype(GetSessionOptionsResult::_impl_.session_options_)::
              InternalGetArenaOffset(
                  ::google::protobuf::Message::internal_visibility()),
  });
  if (arena_bits.has_value()) {
    return ::google::protobuf::internal::MessageCreator::CopyInit(
        sizeof(GetSessionOptionsResult), alignof(GetSessionOptionsResult), *arena_bits);
  } else {
    return ::google::protobuf::internal::MessageCreator(&GetSessionOptionsResult::PlacementNew_,
                                 sizeof(GetSessionOptionsResult),
                                 alignof(GetSessionOptionsResult));
  }
}
#endif
constexpr auto GetSessionOptionsResult::InternalGenerateClassData_() {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &_GetSessionOptionsResult_default_instance_._instance,
          &_table_.header,
          nullptr,  // IsInitialized
          &GetSessionOptionsResult::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<GetSessionOptionsResult>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &GetSessionOptionsResult::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<GetSessionOptionsResult>(), &GetSessionOptionsResult::ByteSizeLong,
              &GetSessionOptionsResult::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(GetSessionOptionsResult, _impl_._cached_size_),
          false,
      },
      &GetSessionOptionsResult::kDescriptorMethods,
      &descriptor_table_Flight_2eproto,
      nullptr,  // tracker
  };
}

PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull GetSessionOptionsResult_class_data_ =
        GetSessionOptionsResult::InternalGenerateClassData_();

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
GetSessionOptionsResult::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&GetSessionOptionsResult_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(GetSessionOptionsResult_class_data_.tc_table);
  return GetSessionOptionsResult_class_data_.base();
}
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 2, 69, 2>
GetSessionOptionsResult::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(GetSessionOptionsResult, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    2,  // num_aux_entries
    offsetof(decltype(_table_), aux_entries),
    GetSessionOptionsResult_class_data_.base(),
    nullptr,  // post_loop_handler
    ::_pbi::TcParser::GenericFallback,  // fallback
    #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
    ::_pbi::TcParser::GetTable<::arrow::flight::protocol::GetSessionOptionsResult>(),  // to_prefetch
    #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
  }, {{
    {::_pbi::TcParser::MiniParse, {}},
  }}, {{
    65535, 65535
  }}, {{
    // map<string, .arrow.flight.protocol.SessionOptionValue> session_options = 1;
    {PROTOBUF_FIELD_OFFSET(GetSessionOptionsResult, _impl_.session_options_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcRepeated | ::_fl::kMap)},
  }},
  {{
      {::_pbi::TcParser::GetMapAuxInfo(
          1, 0, 9, 11, 0)},
      {::_pbi::TcParser::GetTable<::arrow::flight::protocol::SessionOptionValue>()},
  }},
  {{
    "\55\17\0\0\0\0\0\0"
    "arrow.flight.protocol.GetSessionOptionsResult"
    "session_options"
  }},
};
PROTOBUF_NOINLINE void GetSessionOptionsResult::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.GetSessionOptionsResult)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBitForRepeated(cached_has_bits, 0x00000001U)) {
    _impl_.session_options_.Clear();
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL GetSessionOptionsResult::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const GetSessionOptionsResult& this_ = static_cast<const GetSessionOptionsResult&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL GetSessionOptionsResult::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const GetSessionOptionsResult& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.GetSessionOptionsResult)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // map<string, .arrow.flight.protocol.SessionOptionValue> session_options = 1;
  if (CheckHasBitForRepeated(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_session_options().empty()) {
      using MapType = ::google::protobuf::Map<::std::string, ::arrow::flight::protocol::SessionOptionValue>;
      using WireHelper = _pbi::MapEntryFuncs<::std::string, ::arrow::flight::protocol::SessionOptionValue,
                                     _pbi::WireFormatLite::TYPE_STRING,
                                     _pbi::WireFormatLite::TYPE_MESSAGE>;
      const auto& field = this_._internal_session_options();

      if (stream->IsSerializationDeterministic() && field.size() > 1) {
        for (const auto& entry : ::google::protobuf::internal::MapSorterPtr<MapType>(field)) {
          target = WireHelper::InternalSerialize(
              1, entry.first, entry.second, target, stream);
          ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
              entry.first.data(), static_cast<int>(entry.first.length()),
 ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.GetSessionOptionsResult.session_options");
        }
      } else {
        for (const auto& entry : field) {
          target = WireHelper::InternalSerialize(
              1, entry.first, entry.second, target, stream);
          ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
              entry.first.data(), static_cast<int>(entry.first.length()),
 ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.GetSessionOptionsResult.session_options");
        }
      }
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.GetSessionOptionsResult)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t GetSessionOptionsResult::ByteSizeLong(const MessageLite& base) {
  const GetSessionOptionsResult& this_ = static_cast<const GetSessionOptionsResult&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t GetSessionOptionsResult::ByteSizeLong() const {
  const GetSessionOptionsResult& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.GetSessionOptionsResult)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
   {
    // map<string, .arrow.flight.protocol.SessionOptionValue> session_options = 1;
    cached_has_bits = this_._impl_._has_bits_[0];
    if (CheckHasBitForRepeated(cached_has_bits, 0x00000001U)) {
      total_size +=
          1 * ::google::protobuf::internal::FromIntSize(this_._internal_session_options_size());
      for (const auto& entry : this_._internal_session_options()) {
        total_size += _pbi::MapEntryFuncs<::std::string, ::arrow::flight::protocol::SessionOptionValue,
                                       _pbi::WireFormatLite::TYPE_STRING,
                                       _pbi::WireFormatLite::TYPE_MESSAGE>::ByteSizeLong(entry.first, entry.second);
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void GetSessionOptionsResult::MergeImpl(::google::protobuf::MessageLite& to_msg,
                            const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this =
      static_cast<GetSessionOptionsResult*>(&to_msg);
  auto& from = static_cast<const GetSessionOptionsResult&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.GetSessionOptionsResult)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (CheckHasBitForRepeated(cached_has_bits, 0x00000001U)) {
    _this->_impl_.session_options_.MergeFrom(from._impl_.session_options_);
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void GetSessionOptionsResult::CopyFrom(const GetSessionOptionsResult& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.GetSessionOptionsResult)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void GetSessionOptionsResult::InternalSwap(GetSessionOptionsResult* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.session_options_.InternalSwap(&other->_impl_.session_options_);
}

::google::protobuf::Metadata GetSessionOptionsResult::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

class CloseSessionRequest::_Internal {
 public:
};

CloseSessionRequest::CloseSessionRequest(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::internal::ZeroFieldsBase(arena, CloseSessionRequest_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::internal::ZeroFieldsBase(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.CloseSessionRequest)
}
CloseSessionRequest::CloseSessionRequest(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const CloseSessionRequest& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::internal::ZeroFieldsBase(arena, CloseSessionRequest_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::internal::ZeroFieldsBase(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  CloseSessionRequest* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.CloseSessionRequest)
}

inline void* PROTOBUF_NONNULL CloseSessionRequest::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) CloseSessionRequest(arena);
}
constexpr auto CloseSessionRequest::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::ZeroInit(sizeof(CloseSessionRequest),
                                            alignof(CloseSessionRequest));
}
constexpr auto CloseSessionRequest::InternalGenerateClassData_() {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &_CloseSessionRequest_default_instance_._instance,
          &_table_.header,
          nullptr,  // IsInitialized
          &CloseSessionRequest::MergeImpl,
          ::google::protobuf::internal::ZeroFieldsBase::GetNewImpl<CloseSessionRequest>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &CloseSessionRequest::SharedDtor,
          ::google::protobuf::internal::ZeroFieldsBase::GetClearImpl<CloseSessionRequest>(), &CloseSessionRequest::ByteSizeLong,
              &CloseSessionRequest::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(CloseSessionRequest, _impl_._cached_size_),
          false,
      },
      &CloseSessionRequest::kDescriptorMethods,
      &descriptor_table_Flight_2eproto,
      nullptr,  // tracker
  };
}

PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull CloseSessionRequest_class_data_ =
        CloseSessionRequest::InternalGenerateClassData_();

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CloseSessionRequest::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CloseSessionRequest_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(CloseSessionRequest_class_data_.tc_table);
  return CloseSessionRequest_class_data_.base();
}
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 0, 0, 0, 2>
CloseSessionRequest::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(CloseSessionRequest,
                          _impl_._cached_size_),  // no hasbits
    0, // no _extensions_
    0, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967295,  // skipmap
    offsetof(decltype(_table_), field_names),  // no field_entries
    0,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    CloseSessionRequest_class_data_.base(),
    nullptr,  // post_loop_handler
    ::_pbi::TcParser::GenericFallback,  // fallback
    #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
    ::_pbi::TcParser::GetTable<::arrow::flight::protocol::CloseSessionRequest>(),  // to_prefetch
    #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
  }, {{
    {::_pbi::TcParser::MiniParse, {}},
  }}, {{
    65535, 65535
  }}, // no field_entries, or aux_entries
  {{
  }},
};







::google::protobuf::Metadata CloseSessionRequest::GetMetadata() const {
  return ::google::protobuf::internal::ZeroFieldsBase::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

class CloseSessionResult::_Internal {
 public:
  using HasBits =
      decltype(::std::declval<CloseSessionResult>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(CloseSessionResult, _impl_._has_bits_);
};

CloseSessionResult::CloseSessionResult(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CloseSessionResult_class_data_.base()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.CloseSessionResult)
}
CloseSessionResult::CloseSessionResult(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const CloseSessionResult& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CloseSessionResult_class_data_.base()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(from._impl_) {
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}
PROTOBUF_NDEBUG_INLINE CloseSessionResult::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0} {}

inline void CloseSessionResult::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.status_ = {};
}
CloseSessionResult::~CloseSessionResult() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.CloseSessionResult)
  SharedDtor(*this);
}
inline void CloseSessionResult::SharedDtor(MessageLite& self) {
  CloseSessionResult& this_ = static_cast<CloseSessionResult&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.~Impl_();
}

inline void* PROTOBUF_NONNULL CloseSessionResult::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) CloseSessionResult(arena);
}
constexpr auto CloseSessionResult::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::ZeroInit(sizeof(CloseSessionResult),
                                            alignof(CloseSessionResult));
}
constexpr auto CloseSessionResult::InternalGenerateClassData_() {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &_CloseSessionResult_default_instance_._instance,
          &_table_.header,
          nullptr,  // IsInitialized
          &CloseSessionResult::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<CloseSessionResult>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &CloseSessionResult::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<CloseSessionResult>(), &CloseSessionResult::ByteSizeLong,
              &CloseSessionResult::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(CloseSessionResult, _impl_._cached_size_),
          false,
      },
      &CloseSessionResult::kDescriptorMethods,
      &descriptor_table_Flight_2eproto,
      nullptr,  // tracker
  };
}

PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull CloseSessionResult_class_data_ =
        CloseSessionResult::InternalGenerateClassData_();

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CloseSessionResult::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CloseSessionResult_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(CloseSessionResult_class_data_.tc_table);
  return CloseSessionResult_class_data_.base();
}
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2>
CloseSessionResult::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(CloseSessionResult, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    CloseSessionResult_class_data_.base(),
    nullptr,  // post_loop_handler
    ::_pbi::TcParser::GenericFallback,  // fallback
    #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
    ::_pbi::TcParser::GetTable<::arrow::flight::protocol::CloseSessionResult>(),  // to_prefetch
    #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
  }, {{
    // .arrow.flight.protocol.CloseSessionResult.Status status = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(CloseSessionResult, _impl_.status_), 0>(),
     {8, 0, 0,
      PROTOBUF_FIELD_OFFSET(CloseSessionResult, _impl_.status_)}},
  }}, {{
    65535, 65535
  }}, {{
    // .arrow.flight.protocol.CloseSessionResult.Status status = 1;
    {PROTOBUF_FIELD_OFFSET(CloseSessionResult, _impl_.status_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kOpenEnum)},
  }},
  // no aux_entries
  {{
  }},
};
PROTOBUF_NOINLINE void CloseSessionResult::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.CloseSessionResult)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.status_ = 0;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL CloseSessionResult::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const CloseSessionResult& this_ = static_cast<const CloseSessionResult&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL CloseSessionResult::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const CloseSessionResult& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.CloseSessionResult)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // .arrow.flight.protocol.CloseSessionResult.Status status = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (this_._internal_status() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteEnumToArray(
          1, this_._internal_status(), target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.CloseSessionResult)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t CloseSessionResult::ByteSizeLong(const MessageLite& base) {
  const CloseSessionResult& this_ = static_cast<const CloseSessionResult&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t CloseSessionResult::ByteSizeLong() const {
  const CloseSessionResult& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.CloseSessionResult)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

   {
    // .arrow.flight.protocol.CloseSessionResult.Status status = 1;
    cached_has_bits = this_._impl_._has_bits_[0];
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (this_._internal_status() != 0) {
        total_size += 1 +
                      ::_pbi::WireFormatLite::EnumSize(this_._internal_status());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void CloseSessionResult::MergeImpl(::google::protobuf::MessageLite& to_msg,
                            const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this =
      static_cast<CloseSessionResult*>(&to_msg);
  auto& from = static_cast<const CloseSessionResult&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.CloseSessionResult)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (from._internal_status() != 0) {
      _this->_impl_.status_ = from._impl_.status_;
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void CloseSessionResult::CopyFrom(const CloseSessionResult& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.CloseSessionResult)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void CloseSessionResult::InternalSwap(CloseSessionResult* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.status_, other->_impl_.status_);
}

::google::protobuf::Metadata CloseSessionResult::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// @@protoc_insertion_point(namespace_scope)
}  // namespace protocol
}  // namespace flight
}  // namespace arrow
namespace google {
namespace protobuf {
}  // namespace protobuf
}  // namespace google
// @@protoc_insertion_point(global_scope)
PROTOBUF_ATTRIBUTE_INIT_PRIORITY2 static ::std::false_type
    _static_init2_ [[maybe_unused]] =
        (::_pbi::AddDescriptors(&descriptor_table_Flight_2eproto),
         ::std::false_type{});
#include "google/protobuf/port_undef.inc"
